
// module drawxtl1.cxx - part of DRAWxtl V5.1 - the GUI version
// Coded using the FLTK 1.1.6 widget set
//
//     Larry W. Finger, Martin Kroeker and Brian Toby
//
//
// This module contains the following routines:
//
//  add_to_list - add atom position to unit-cell list
//  add_vert - add vertices (atom positions) to display lists
//  add_vert_nc - add vertices without checking
//  analyze_bonds - print the bond distance tables
//  axeqb - solve matrix equation AX = B for B (A is 3x3 only)
//  build_box_contents - find all atoms in the display parallelopiped
//  Calc_Rot - calculate rotation matrices
//  convert_ellipsoid - handles conversions from Bij or Uij to betaij
//  Conv_Sym_Mat - converts rotational part of symmetry matrix to a Cartesian rotation
//  dist - calculates distance between two vertices
//  dot0_3d - calculated dot product of two 3D vectors
//  eigen - calculates eigen values and vectors for ellipsoids
//  enormsq3d - calculates squared distance between two points
//  find_all_in_box - finds all atoms of a given type in the display box
//  generate_arrows - generates magnetic vector arrows
//  generate_bonds - Generate bond descriptions for the display lists
//  generate_cones - generate lone-pair cone descriptions for display lists
//  get_atom_id - reworks atom labels
//  get_input - calls actual routine that reads input, generates lattice metric and does preliminary ellipsoid processing
//  Locate_Triple - routine to place unit-cell vector triple
//  make_bmat - calculates lattice metric routines for conversion to cartesian coordinates
//  not_in_slab - check if a point is inside a given parallelepiped
//  Output_Spheres - adds sphere descriptions to output display lists
//  plot_vrml_poly - generates polyhedral descriptions to VRML output
//  print_sym - print symmetry operators

#include "drawxtl.h"
#include "DRAWxtlViewUI.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <FL/glut.H>
#if defined(__APPLE__)
#  include <openGL/glu.h>
#else
#  include <GL/glu.h>
#endif

// forward references for routines

#include "DRAWxtl_proto.h"

// global variables

#include "draw_ext.h"

/* ************************************************************** */
/* ************************************************************** */

void add_to_list (float xp[3], int no)

/* add atom position to unit-cell list if unique */
/* no - symmetry operator number */
{
  int i, ind;

  for (i = 0; i <= 2; ++i) {        	 /* get coordinates between 0 and 1 */
    if (xp[i] >= 1.0)
      xp[i] = xp[i] - 1.0f;
    if (xp[i] < 0.0)
      xp[i] = xp[i] + 1.0f;
  }
  ind = 0;
  for (i = 0; i < ncell; ++i) {
    if ((fabs (drvui->cell_xyz[i][0] - xp[0]) <= 0.00002) &&
    (fabs (drvui->cell_xyz[i][1] - xp[1]) <= 0.00002) &&
    (fabs (drvui->cell_xyz[i][2] - xp[2]) <= 0.00002))
      ind = 1;        			 /*set if same */
  }
  if (ind == 0) {        		 /* true for new position */
    for (i = 0; i <= 2; ++i)        	 /*copy to list */
      drvui->cell_xyz[ncell][i] = xp[i];
    drvui->sym_op_no[ncell++] = no;
  }
}

/* ************************************************************** */
/* ************************************************************** */

void add_vert (float vert[3], int no, int check)

/* procedure to add the vertex in 'vert' to the current list if the
   vertex is contained within the limits of the box, within the limits
   of a polyhedral center, or needed to complete a molecule */
{
  float c_vert[3];               /* vertex in cartesian coords */
  float r_vert[3];               /* vertex in rotated coordinates */
  int i, j;                      /* loop variables */
  float temp;                    /* temporary storage */
  int outside;                   /* temporary for molecular completion */
  int number;                    /* decoded atom number */
  int poly_center;               /* true if this atom is center of a polyhedron */
  int consider_extra;
  float d = 0.0f;

  for (i = 0; i < 3; ++i) {      /* convert vertex coordinates to Cartesian */
    c_vert[i] = 0.0f;
    for (j = 0; j < 3; ++j)
      c_vert[i] += (float) drvui->b_mat[i][j] * (vert[j] - origin[j]);
  }
  number = no / 100;
  if (number >= 0)
      poly_center = (((drvui->atom_n[number] >> 8) & 255) > 0);
  else
      poly_center = 0;
  consider_extra = (domolcomp != 0 || (drvui->npoly > 1 && !poly_center));
  outside = 0;
  if (check == 0) {
      for (i = 0; i <= 2; i++) {  /* check packing range */
        if ((vert[i] < drvui->cryst_lim[drvui->frame_no][i]) || (vert[i] > drvui->cryst_lim[drvui->frame_no][i + 3])) 
          outside = 1;
      }
      if (outside == 0 && slabmode == 1)
        outside = not_in_slab(c_vert[0], c_vert[1],c_vert[2]);
      if (((outside == 0) && (fabs (c_vert[0]) > boxlim[0])) ||
         (fabs (c_vert[1]) > boxlim[1]) || (fabs (c_vert[2]) > boxlim[2]))
        outside = 1;        		 /* check box limits */
      if (outside != 0 && !consider_extra)
        return;
  }
  d = 0.0f;
  if (outside == 1) {            /* molecule or polyhedral completion is in progress */
    float dmin, dmax;
    int polyno;        			 /* number of the polyhedron */

    for (i = 1; i < NvertM; i++) {
      dmax = drvui->mol_d * drvui->mol_d;
      dmin = 0.;
      d = 0.0f;
      for (j = 0; j < 3; j++) {
        temp = (s_vert[3 * i + j] - c_vert[j]);
        d += temp * temp;
      }
      j = drvui->atom_no[i] / 100;
      poly_center = (((drvui->atom_n[j] >> 8) & 255) > 0);
      if (poly_center) {
        polyno = (drvui->atom_n[j] >> 8 ) & 255;
        dmax = drvui->poly_size[polyno] * drvui->poly_size[polyno];
        dmin = drvui->poly_min[polyno] * drvui->poly_min[polyno];
      }
      if ((domolcomp != 0 || poly_center) && (d <= dmax && d > 0.02)) {
        outside = 2;
        if (poly_center && d < dmin)
          outside = 1;
        if ((fabs (c_vert[0]) > 10. * boxlim[0]) || (fabs (c_vert[1]) > 10. * boxlim[1])
                 || (fabs (c_vert[2]) > 10. * boxlim[2])) {
          fprintf (drvui->fcns, "mol'completing beyond 10*lattice constant - aborting run\n");
          outside = 1;        		 /* far out of check box limits - combinatorial explosion */
          break;
        }
      }
    }
  }
  if (outside == 1)
    return;
  i = (int) (no / 100);
  for (j = 1; j < NvertM; ++j) {
    if (fabs (s_vert[3 * j] - c_vert[0]) < 0.01 && fabs (s_vert[3 * j + 1] - c_vert[1]) < 0.01 &&
        fabs (s_vert[3 * j + 2] - c_vert[2]) < 0.01 && 
        check_atom_name(drvui->atom_l[i],drvui->atom_l[drvui->atom_no[j]/100])  && (number == i))
      return;        			 /* eliminate duplicates */
  }
  for (i = 0; i <= 2; ++i) {        	 /* add atom to Master List */
    xypos[3 * NvertM + i] = vert[i];
    s_vert[3 * NvertM + i] = c_vert[i];
//    if (offset[i] > c_vert[i])
//      offset[i] = c_vert[i];
  }
  i = no - 100*(no/100);
  drvui->atom_no[NvertM++] = no;
  if (NvertM > MAX_VERTS) {
    if (domolcomp != 0) {
      Error_Box ("Too many vertices for dimensions:\n"
          " Did you use a 'molcomp' command incorrectly?\n"
          " If not, increase MAX_VERTS.");
    } else {
      Error_Box("Too many vertices for dimensions - increase MAX_VERTS.");
    }
    return;
  }
  for (i = 0; i <= 2; ++i) {        	 /* calculate position of point after POV rotation */
    r_vert[i] = 0.0f;
    for (j = 0; j <= 2; ++j)
      r_vert[i] += (float) G_Rot[j][i] * c_vert[j];
  }
  for (i = 0; i <= 2; ++i) {        	 /* Update Min and Max of output */
    if (POV_Max[i] < r_vert[i])
      POV_Max[i] = r_vert[i];
    if (POV_Min[i] > r_vert[i])
      POV_Min[i] = r_vert[i];
  }
}

/* ************************************************************** */
/* ************************************************************** */

void add_vert_nc (float vert[3])

/* procedure to add the vertices in 'vert' to the current working list
   (NOT Master List ) without checking */
{

  int i, j;        			 /* loop variables */

  if (!s_vert) {
      if ((s_vert = (float *) malloc ((unsigned) (3 * (long) (MAX_VERTS 
                  * sizeof (float))))) == NULL) {
          Error_Box ("Unable to get initial s_vert allocation");
          return;
      }
  }
  for (i = 0; i <= 2; ++i) {        	 /* convert vertex coordinates to Cartesian */
      s_vert[3 * nvert + i] = 0.0f;
      for (j = 0; j <= 2; ++j)
          s_vert[3 * nvert + i] += (float) drvui->b_mat[i][j] * (vert[j] - origin[j]);
  }
  nvert++;
}

/* ************************************************************** */
/* ************************************************************** */

void analyze_bonds (void)

/* routine to print bond distances up to a maximum of 'printdist' Angstrom */
{
  int i, j, k, l, nvert1, *itype;
  float *d, dmin;

  if (!(itype = (int *) malloc ((unsigned) (MAX_VERTS * sizeof (int))))) {
      Error_Box("Unable to assign space for bond analysis.");
      return;
  }
  fprintf (drvui->flout, "\n\n Bond Distance Analysis\n");
/* Find all atoms of this type in master list */
  l = nvert = nvert1 = 1;
  for (j = 0; j < natom; ++j) {
    if (drvui->atom_fn[j] != drvui->frame_no) continue;
    l = nvert;
    find_all_in_box (j);
    for (k = l; k < nvert; ++k)
      itype[k] = j;        		 // itype[k] has atom number j
  }
  for (i = 0; i < natom; ++i) {        	 /* loop through atom types */
  if (drvui->atom_fn[i] != drvui->frame_no) continue;
/* add the base entry to the beginning of the atom list */
    for (j = 0; j < 3; ++j) {        	 /* convert vertex coordinates to Cartesian */
      s_vert[j] = 0.0f;
      for (k = 0; k <= 2; ++k)
        s_vert[j] += (float) drvui->b_mat[j][k] * (drvui->atom_xyz[i][k] - origin[k]);
    }
    if ((d = (float *) malloc ((unsigned) (nvert * sizeof (float)))) == NULL) {
      free (itype);
      Error_Box("Unable to get bond distance allocation");
      return;
    }
    for (j = 1; j < nvert; ++j)        	 /* calculate distances */
      d[j] = dist (0, j);
    fprintf (drvui->flout,
         "\nDistances from %c%c%c%c%3d at %8.5f%8.5f%8.5f    to\n\n",
         drvui->atom_l[i][0], drvui->atom_l[i][1], drvui->atom_l[i][2], drvui->atom_l[i][3],
         drvui->sv_atom_n[i]
         , drvui->atom_xyz[i][0], drvui->atom_xyz[i][1], drvui->atom_xyz[i][2]);
    for (k = nvert1; k < nvert; ++k) {
      dmin = 999999.9f;
      for (j = nvert1; j < nvert; ++j) {     /* loop through distances */
        if (d[j] < dmin) {
          l = j;
          dmin = d[j];
        }
      }
      if (dmin == 0.0 && itype[l] == i) {
        d[l] = 999999.9f;
      } else {
        if (dmin <= printdist) {
          fprintf (drvui->flout," %c%c%c%c%3d at %9.5f%9.5f%9.5f  %8.3f\n",
               drvui->atom_l[itype[l]][0], drvui->atom_l[itype[l]][1],drvui->atom_l[itype[l]][2]
               , drvui->atom_l[itype[l]][3], drvui->sv_atom_n[itype[l]],
               o_vert[3 * l], o_vert[3 * l + 1], o_vert[3 * l + 2], dmin);
          d[l] = 999999.9f;
        } else
          break;
      }
    }
    free (d);
  }
  free (itype);
  fprintf (drvui->flout, "\n\n");
}

/* ************************************************************** */
/* ************************************************************** */

void axeqb (double a1[3][3], double x[3], double b[3])

/* solve matrix equation ax = b for x, where a is 3x3 and x and b are
   vectors */
{
  int i, j;
  double a[3][3], det;

  det = determinant (a1);
  if (fabs (det) < 1.0e-6)
    det = 1.0e-6;
  a[0][0] = (a1[1][1] * a1[2][2] - a1[1][2] * a1[2][1]) / det;
  a[1][0] = -(a1[1][0] * a1[2][2] - a1[1][2] * a1[2][0]) / det;
  a[2][0] = (a1[1][0] * a1[2][1] - a1[1][1] * a1[2][0]) / det;
  a[0][1] = -(a1[0][1] * a1[2][2] - a1[0][2] * a1[2][1]) / det;
  a[1][1] = (a1[0][0] * a1[2][2] - a1[0][2] * a1[2][0]) / det;
  a[2][1] = -(a1[0][0] * a1[2][1] - a1[0][1] * a1[2][0]) / det;
  a[0][2] = (a1[0][1] * a1[1][2] - a1[0][2] * a1[1][1]) / det;
  a[1][2] = -(a1[0][0] * a1[1][2] - a1[0][2] * a1[1][0]) / det;
  a[2][2] = (a1[0][0] * a1[1][1] - a1[0][1] * a1[1][0]) / det;
  for (i = 0; i < 3; i++) {
    x[i] = 0.0;
    for (j = 0; j < 3; j++)
      x[i] += a[i][j] * b[j];
  }
}

/* ************************************************************** */
/* ************************************************************** */

void build_box_contents (void)

/* routine to locate all atoms in the display parallelopiped plus
   any atoms needed to complete molecules or polyhedra */
{
  float vert[3];        		 /* storage for center */
  int ix[3];        			 /* array for lattice offset */
  int j, k, saved_nvert;
  int atoms, loop;
  float saved_xyz[MAX_ATOM][3];

  loop = 0;
  while (1) {        		         // infinite loop here
    saved_nvert = NvertM;        	 // save number currently in list
    for (atoms = 0; atoms < natom; atoms++) {     // loop through atoms
      if (drvui->atom_fn[atoms] != drvui->frame_no) continue;
      for (j=0;j<3;j++) {            // put the base atom in the range 0-1
        saved_xyz[atoms][j] = drvui->atom_xyz[atoms][j];
        if (drvui->atom_xyz[atoms][j] < 0.0f) drvui->atom_xyz[atoms][j] += 1.0f;
        if (drvui->atom_xyz[atoms][j] > 1.0f) drvui->atom_xyz[atoms][j] -= 1.0f;
      }
      expand_atom (atoms);        	 // generate all atoms of this flavor in unit cell
      for (j = 0; j < ncell; ++j) {  // atoms in cell
        for (k = 0; k <= 2; ++k)
          vert[k] = drvui->cell_xyz[j][k];
        add_vert (vert, 100 * atoms + drvui->sym_op_no[j], 0);    // add point to cell if inside box
      }
    }
    if (saved_nvert == NvertM)
      break;
  }
  while (1) {        		         /* infinite loop here */
    saved_nvert = NvertM;        	 /* save number currently in list */
    for (atoms = 0; atoms < natom; atoms++) {     /* loop through atoms */
      if (drvui->atom_fn[atoms] != drvui->frame_no) continue;
      expand_atom (atoms);        	 /* generate all atoms of this flavor in unit cell */
      for (j=0;j<ncell;++j) { /*atoms in cell*/
        for (k=0;k<=2;++k) 
          vert[k] = drvui->cell_xyz[j][k];
        add_vert(vert,100*atoms+drvui->sym_op_no[j],0);  /* add point to cell if inside box*/
      }
      for (ix[0] = (int) drvui->cryst_lim[drvui->frame_no][0] - 1; ix[0] <= (int) drvui->cryst_lim[drvui->frame_no][3] + 1; ++ix[0]) {    /* step through cells parallel a */
        for (ix[1] = (int) drvui->cryst_lim[drvui->frame_no][1] - 1; ix[1] <= (int) drvui->cryst_lim[drvui->frame_no][4] + 1; ++ix[1]) {    /* parallel b */
          for (ix[2] = (int) drvui->cryst_lim[drvui->frame_no][2] - 1; ix[2] <= (int) drvui->cryst_lim[drvui->frame_no][5] + 1; ++ix[2]) {    /* parallel c */
            if (abs (ix[0]) + abs (ix[1]) + abs (ix[2]) != 0) {
              for (j = 0; j < ncell; ++j) {     /*atoms in cell */
                for (k = 0; k <= 2; ++k)
                  vert[k] = ix[k] + drvui->cell_xyz[j][k];
                add_vert (vert, 100 * atoms + drvui->sym_op_no[j], 0);    /* add point to cell if inside box */
              }
            }
          }
        }
      }
    }
    if ((domolcomp == 0 && drvui->npoly == 1) || saved_nvert == NvertM) {
      for (atoms = 0; atoms < natom; atoms++) {     /* loop through atoms */
          if (drvui->atom_fn[atoms] != drvui->frame_no) continue;
          for (j=0;j<3;j++) {            // restore base atom position
              drvui->atom_xyz[atoms][j] = saved_xyz[atoms][j];
          }
      }
      return;
    }
    if (loop == 0)
      fprintf (drvui->fcns,"\n");
    loop++;
    fprintf (drvui->fcns,"Recursive Atom Addition, Loop No. %3d, %4d Atoms in List.\n",loop, NvertM - 1);
  }
}        				 /* end of build_box_contents */

/* ************************************************************** */
/* ************************************************************** */

void Calc_Rot (float v1[3], float v2[3])

/* routine to calculate rotation axes needed to put v1 parallel to Z
   and the projection of v2 parallel to Y */
{
  float u1[3], u2[3], u3[3], temp;
  double U[3][3];
  int i, j;

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      U[i][j] = 0.0f;        	 /* clear U */
  temp = 0.0f;
  for (i = 0; i < 3; i++) {        	 /* convert v1 to Cartesian */
    u1[i] = 0.0f;
    for (j = 0; j < 3; j++)
      u1[i] += (float) drvui->b_mat[i][j] * v1[j];
    temp += u1[i] * u1[i];
  }
  if (temp <= 0.0) {
    u1[2] = 1.0f;        	 /* if null input, make u1 = 0,0,1 */
    U[2][2] = 1.0f;
  } else {
    temp = (float) sqrt (temp);
    for (i = 0; i < 3; i++) {        	 /* make a unit vector */
      u1[i] /= temp;
      U[i][2] = u1[i];        		 /* put column in U */
    }
  }
  for (i = 0; i < 3; i++) {        	 /* convert v2 to Cartesian */
    u2[i] = 0.0f;
    for (j = 0; j < 3; j++)
      u2[i] += (float) drvui->b_mat[i][j] * v2[j];
  }
  u3[0] = u2[1] * u1[2] - u1[1] * u2[2];     /* calculate u2 X u1 */
  u3[1] = u2[2] * u1[0] - u2[0] * u1[2];
  u3[2] = u2[0] * u1[1] - u1[0] * u2[1];
  temp = 0.0f;
  for (i = 0; i < 3; i++)
    temp += u3[i] * u3[i];
  if (temp <= 0.0) {
    u3[0] = 1.0f;        	 /* if null input, make u3 = 1,0,0 */
    U[0][0] = 1.0f;
  } else {
    temp = (float) sqrt (temp);
    for (i = 0; i < 3; i++) {        	 /* make a unit vector */
      u3[i] /= temp;
      U[i][0] = u3[i];        		 /* add column to U */
    }
  }
  u2[0] = u1[1] * u3[2] - u3[1] * u1[2];     /* calculate u2 = u1 X u3 */
  u2[1] = u1[2] * u3[0] - u1[0] * u3[2];
  u2[2] = u1[0] * u3[1] - u3[0] * u1[1];
  for (i = 0; i < 3; i++)
    U[i][1] = u2[i];
/* make determinant of U > 0 */
  if (determinant (U) < 0.0) {
    for (i = 0; i < 3; i++)
      U[i][0] = -U[i][0];
  }

/* The rotation required to get the desired orientation, which is described
    by G_Rot, has been set up so G_Rot * U = I. Thus G_Rot = U(inv),
    U is unitary and U(inv) = U(transpose). */

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      G_Rot[i][j] = (float) U[j][i];
/* calculate rotations */
  temp = (float) (G_Rot[1][2] * G_Rot[1][2] + G_Rot[2][2] * G_Rot[2][2]);
  if (temp > 0.001) {
    float temp2;

    yrot = (float) (atan (G_Rot[0][2] / sqrt (temp)) * RAD);
    temp2 = (float) (G_Rot[1][2] / cos (yrot / RAD));
    if (temp2 > 1.0)
      temp2 = 1.0f;
    if (temp2 < -1.0)
      temp2 = -1.0f;
    xrot = -(float) asin (temp2) * (float) RAD;
    temp2 = -(float) (G_Rot[0][1] / cos (yrot / RAD));
    if (temp2 > 1.0)
      temp2 = 1.0f;
    if (temp2 < -1.0)
      temp2 = -1.0f;
    zrot = (float) asin (temp2) * (float) RAD;
  }
  else {
    yrot = -(float) asin (G_Rot[0][2]) * (float) RAD;
    xrot = 0.0f;        		 /* XXXXXXXXXXX  - May not be right */
    zrot = 0.0f;        		 /* XXXXXXXXXXX  - May not be right */
  }
}


/* ************************************************************** */
/* ************************************************************** */

void convert_ellipsoid (void)

/* routine to convert ellipsoid coefficients Bij or Uij to betaij form */
{
  int i, j, save;

  for (i = 1; i < n_ellips; ++i) {
    save = drvui->ell_type[i]/1000;
    if (drvui->ell_type[i] % 1000 != 0) {     // multiply Uij and Bij by a*(i)*a*(j)/4
      for (j = 0; j < 3; ++j)
        drvui->ellips[i][j] *= drvui->rec_lat_con[j] * drvui->rec_lat_con[j] * 0.25f;
      drvui->ellips[i][3] *= drvui->rec_lat_con[0] * drvui->rec_lat_con[1] * 0.25f;
      drvui->ellips[i][4] *= drvui->rec_lat_con[0] * drvui->rec_lat_con[2] * 0.25f;
      drvui->ellips[i][5] *= drvui->rec_lat_con[1] * drvui->rec_lat_con[2] * 0.25f;
      if (drvui->ell_type[i] % 1000 == 1)        // multiply Uij by 8pi^2
        for (j = 0; j < 6; ++j)
          drvui->ellips[i][j] *= 78.9568f;
    }
    drvui->ell_type[i]= 1000 * save;
  }
}        				 /* end of convert_ellipsoid */

/* ************************************************************** */
/* ************************************************************** */

void Conv_Sym_Mat (void)

/* routine to convert the rotational part of a symmetry matrix to a Cartesian rotation */
{
  double sum;
  float angle;
  int j, k, l;
  double mat[3][3];

  for (l = 0; l < drvui->ng; l++) {
    for (j = 0; j < 3; j++) {
      for (k = 0; k < 3; k++) {
        mat[j][k] = drvui->rss[l][j][k];
      }
    }
    if (determinant (mat) < 0.0) {         /* if improper rotation, negate all elements */
      for (j = 0; j < 3; j++) {
        for (k = 0; k < 3; k++) {
          mat[j][k] *= -1.0f;
        }
      }
    }
    for (j = 0, sum = 0.0; j < 3; j++)         /* compute sum of squares of elements */
      for (k = 0; k < 3; k++)
        sum += mat[j][k] * mat[j][k];
    if (fabs (sum - 3.0) > 0.1) {         /* if sum = 3, rotation matrix OK */
/* at this point only 4 operators in trigonal or hexagonal system left */
/*    we need to find which we have */
      if (fabs (mat[0][0]) <= 0.1) {
        if (fabs (mat[0][1] - 1.0) < 0.0001) {
          angle = -1.0471976f;         /* -60 degrees (in radians) */
        } else {
          angle = 2.0943951f;         /* 120 degrees */
        }
      } else {
        if (fabs (mat[0][0] - 1.0) < 0.0001) {
          angle = 1.0471976f;         /* 60 degrees (in radians) */
        } else {
          angle = -2.0943951f;         /* -120 degrees */
        }
      }
      mat[0][0] = mat[1][1] = (float) cos (angle);
      mat[0][1] = (float) sin (angle);
      mat[1][0] = -mat[0][1];
    }
    if (determinant (mat) < 0.0) {         /* if improper rotation, negate all elements */
      for (j = 0; j < 3; j++) {
        for (k = 0; k < 3; k++) {
          mat[j][k] *= -1.0f;
        }
      }
    }
    for (j = 0; j < 3; j++) {        	 /* restore matrix */
      for (k = 0; k < 3; k++) {
        drvui->rss[l][j][k] = (float) mat[j][k];
      }
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

float dist (int j, int k)

/* calculates distance between vertices at s_vert[j] and s_vert[k] */
{
  float d, t;
  int i;

  d = 0.0f;
  for (i = 0; i <= 2; ++i) {
    t = s_vert[3 * j + i] - s_vert[3 * k + i];
    d += t * t;
  }
  return ((float) sqrt (d));
}

/* ************************************************************** */
/* ************************************************************** */

float dot0_3d (float x0, float y0, float z0, float x1, float y1, float z1,
	       float x2, float y2, float z2)
{
// 3D dot product of (P1-P0) and (P2-P0), John Burkardt

  return (x1 - x0) * (x2 - x0) + (y1 - y0) * (y2 - y0) +
          (z1 - z0) * (z2 - z0);
}

/* ************************************************************** */
/* ************************************************************** */

int eigen (float *biso, float beta[6], float valu[3], float vect[3][3])

/* routine to calculate eigenvalues 'valu' and vectors 'v' for
   array of anisotropic thermal parameters 'beta', 'biso' is the equivalent
   isotropic B  */
/* routine copied from ORTEP */
{
  static double errnd = 0.000007;
  double b0, b1, b2, c0, c1;
  double a[3][3], b[3][3], w[3][3], u[3], v[3], z[3];
  double x, y, tem, sigma, vnew, vold, smax;
  double Vect[3][3];
  int i, j, l, ii, iii, i1, imax = 0;
  static double tpi2 = 19.7392088;         /* 1/2 pi^2 */

#define phif(z) ((b2 - z) * z + b1) * z + b0

/* put beta into square symmetric matrix b (called M in ORTEP) */

  for (i = 0; i < 3; i++)
    b[i][i] = beta[i];
  b[0][1] = b[1][0] = beta[3];
  b[0][2] = b[2][0] = beta[4];
  b[1][2] = b[2][1] = beta[5];
/* multiply b * ginv to get w */
  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++) {
      w[i][j] = 0.0;
      for (l = 0; l < 3; l++) {
        w[i][j] += b[l][i] * drvui->ginv[j][l];
      }
    }
  }
  sigma = 0.0;
  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++) {
      a[i][j] = w[i][j];
      sigma += a[i][j] * a[i][j];
    }
  }
  if (sigma <= 0.0)
    return (0);        			 /* error on null 'w' */
  sigma = sqrt (sigma);
/* get coefficients of third-order characteristic equation 
        Equation is -L^3 + b2 L^2 + b1 L + b0 = 0
*/
  b2 = a[0][0] + a[1][1] + a[2][2];
  b1 = -a[0][0] * a[1][1] - a[0][0] * a[2][2] - a[1][1] * a[2][2]
    + a[0][2] * a[2][0] + a[0][1] * a[1][0] + a[1][2] * a[2][1];
  b0 = a[0][0] * a[1][1] * a[2][2] + a[1][0] * a[2][1] * a[0][2]
    + a[2][0] * a[1][2] * a[0][1] - a[2][0] * a[0][2] * a[1][1]
    - a[0][0] * a[2][1] * a[1][2] - a[1][0] * a[0][1] * a[2][2];

  *biso = (float) (4.0 * b2 / 3.0);

/* first root by bisection */

  x = 0.0;
  y = sigma;
  tem = phif (sigma);
  vnew = 0.0;
  if (b0 != 0.0) {
    if ((b0 < 0.0) && (tem <= 0.0))
      y = -y;
    if (b0 > 0.0) {
      y = 0.0;
      x = sigma;
      if (tem > 0.0)
        x = -x;
    }

/* phif(y) > 0 and phif(x) < 0  */

    for (j = 0; j < 80; ++j) {
      if ((tem = phif (vnew)) == 0.0)
        break;
      if (tem < 0.0) {
        x = vnew;
      } else
    y = vnew;
      vold = vnew;
      vnew = (x + y) * 0.5;
      tem = fabs (vold - vnew);
      if (tem <= errnd)
        break;
      if (vold != 0.0) {
        if (fabs (tem / vold) <= errnd)
          break;
      }
    }
    if (j == 80) {
      fprintf (drvui->flout, "No convergence.\n");
      return (0);
    }
  }
  u[2] = vnew;
  c1 = b2 - vnew;
  c0 = b1 + c1 * vnew;
  tem = c1 * c1 + 4.0 * c0;
  if (tem < -0.0001) {
    fprintf (drvui->flout, "Complex Roots! tem = %f\n", tem);
    return (0);        			 /* error if complex roots */
  }
  if (tem < 0.0)
    tem = 0.0;
  tem = sqrt (tem);
  u[0] = 0.5 * (c1 - tem);
  u[1] = 0.5 * (c1 + tem);
  for (j = 0; j < 2; j++) {        	 /* sort roots in increasing order */
    if (u[j] > u[2]) {
      tem = u[2];
      u[2] = u[j];
      u[j] = tem;
    }
  }
  if (u[0] < 0.0) {
    fprintf (drvui->flout, "Imaginary Root, squares are: %8.5f %8.5f %8.5f\n",u[0], u[1], u[2]);
    return (0);
  }

/* count multiple roots */
  tem = 100.0 * errnd;
  j = 0;
  if ((u[1] - u[0] - tem) < 0)
    j++;
  if ((u[2] - u[1] - tem) < 0)
    j += 2;
  if (j == 3) {
/* 3 equal roots */
    for (ii = 0; ii < 3; ii++) {
      for (i = 0; i < 3; i++)
        vect[i][ii] = 0.0;
      vect[ii][ii] = 1.0;
      valu[ii] = (float) sqrt (u[ii] / tpi2);
    }
    return (1);
  }
  for (ii = 0; ii < 3; ii++) {
    for (iii = 0; iii < 3; iii++) {
      for (i = 0; i < 3; i++)
        a[i][iii] = w[i][iii];
      a[iii][iii] = w[iii][iii] - u[ii];
    }
    smax = 0.0;
    for (i = 0; i < 3; i++) {
      i1 = (i < 2) ? i + 1 : 0;
      b[0][i] = a[1][i] * a[2][i1] - a[2][i] * a[1][i1];
      b[1][i] = a[2][i] * a[0][i1] - a[0][i] * a[2][i1];
      b[2][i] = a[0][i] * a[1][i1] - a[1][i] * a[0][i1];
      tem = b[0][i] * b[0][i] + b[1][i] * b[1][i] + b[2][i] * b[2][i];
      if (tem > smax) {
        smax = tem;
        imax = i;
      }
    }
    if (smax <= 0.0)
      return (0);
    smax = (float) sqrt (smax);
    for (i = 0; i < 3; i++)
      v[i] = b[i][imax] / smax;
    axeqb (a, z, v);
/* convert direct-space vector z into Cartesian system */
    for (i = 0; i < 3; i++) {
      v[i] = 0.0;
      for (j = 0; j < 3; j++)
        v[i] += drvui->b_mat[i][j] * z[j];
    }
    tem = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
    tem = (float) sqrt (tem);
    for (i = 0; i < 3; i++)
      Vect[ii][i] = v[i] / tem;        	 /* unit vector */
  }
  for (i = 0; i < 3; i++)
    valu[i] = (float) sqrt (u[i] / tpi2);
/* look for degenerate vectors */
  if (fabs (Vect[0][0] * Vect[1][0] + Vect[0][1] * Vect[1][1]
        + Vect[0][2] * Vect[1][2]) > 0.001) {    /* set #1 = 2 x 3 */
    Vect[0][0] = Vect[1][1] * Vect[2][2] - Vect[2][1] * Vect[1][2];
    Vect[0][1] = Vect[2][0] * Vect[1][2] - Vect[1][0] * Vect[2][2];
    Vect[0][2] = Vect[1][0] * Vect[2][1] - Vect[2][0] * Vect[1][1];
  }
  if (fabs (Vect[2][0] * Vect[1][0] + Vect[2][1] * Vect[1][1]
        + Vect[2][2] * Vect[1][2]) > 0.001) {    /* set #3 = 1 x 2 */
    Vect[2][0] = Vect[0][1] * Vect[1][2] - Vect[1][1] * Vect[0][2];
    Vect[2][1] = Vect[1][0] * Vect[0][2] - Vect[0][0] * Vect[1][2];
    Vect[2][2] = Vect[0][0] * Vect[1][1] - Vect[1][0] * Vect[0][1];
  }
  if (determinant (Vect) < 0.0) {
    for (i = 0; i < 3; i++)
      for (j = 0; j < 3; j++)
        Vect[i][j] *= -1.0;
  }
  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      vect[i][j] = (float) Vect[i][j];
  return (1);
}

/* ************************************************************** */
/* ************************************************************** */

float enormsq0_3d (float x0, float y0, float z0, float x1, float y1, float z1)
{
// square of the Euclidean norm of (P1-P0) in 3D, John Burkardt

  return (x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0) +
          (z1 - z0) * (z1 - z0);
}

/* ************************************************************** */
/* ************************************************************** */

void find_all_in_box (int i)

/* routine to locate all atoms of type 'i' in the display box */
/* i - atom number */
{
  int n, j, k, l;        		 /* loop indices */
  int sym_no;        			 /* symmetry number */

  for (n = 1; n < NvertM; n++) {         /* loop through the Master list */
    j = (int) drvui->atom_no[n] / 100;        	 /* extract atom number */
    if (j == i) {
      drvui->orig_atom_no[nvert] = i;
      sym_no = drvui->atom_no[n] - 100 * j;         /* symmetry operator number */
      for (k = 0; k < 3; k++) {
        o_vert[3 * nvert + k] = xypos[3 * n + k];
        s_vert[3 * nvert + k] = 0.0f;
        for (l = 0; l < 3; l++) {         /* calculate cartesian coordinates */
          s_vert[3 * nvert + k] +=
        (float) drvui->b_mat[k][l] * (xypos[3 * n + l] - origin[l]);
        }
      }
      vert_sym_no[nvert++] = sym_no;
      if (nvert > MAX_VERTS) {
        Error_Box("Too many vertices for dimensions - increase MAX_VERTS.");
        return;
      }
      if (nvert > 2 * NvertM) {
#ifdef WIN32
        Error_Box("Overrun of vert_sym_no. Please send 'str' file\n"
                   "to Larry.Finger@@lwfinger.net.");
        exit(0);
#else
        vert_sym_no = (int *)realloc(vert_sym_no,  (nvert + 2) * sizeof(int));
        drvui->orig_atom_no = (int *)realloc(drvui->orig_atom_no, (nvert+2) * sizeof(int));
        o_vert = (float *)realloc (o_vert, (6 * (nvert+2) * sizeof (float)));
#endif
      }
    }
  }
}        				 /* end of find_all_in_box */

/* ************************************************************** */
/* ************************************************************** */

void generate_arrows (void)

// routine to generate arrow objects for magnetic moments
{
  int Arrow_Count;				 // Counter for number output
  int i, k, l;				     // loop counters
  float glr,glb,glg;
  GLUquadricObj *glu_quadObj;
  float base;
  float x[3], xp[3], yp[3];
  int xi,xj,xk;
  float phi, chi;
  int savedNvertM = NvertM;                // save the end of the vertex list
     
  Arrow_Count = 0;
  for (i = 0; i < drvui->nmag; ++i) {	   // loop through magnetic arrows
    if (drvui->arrow_fn[i] != drvui->frame_no) continue;
    for (k=0;k<3;k++) {
        x[k] = drvui->mag_xp[i][k] - drvui->xyzoff[k];
    }
    for (xi = -3; xi<=3; xi++) {
        xp[0] = (float)xi;
        for (xj = -3; xj<=3; xj++) {
            xp[1] = (float)xj;
            for (xk=-3; xk<=3; xk++) {
                xp[2] = (float)xk;         // xp is a magnetic cell translation
                for (k=0;k<3;k++) {
                    yp[k]=x[k];            // yp will be position in nuclear cell
                    for (l=0;l<3;l++) {
                        yp[k] += drvui->mag_matrix[k][l] * xp[l];
                    }
                }
                add_vert(yp,100*(i+MAX_ATOM),0); // add vector to vertex list
            }
        }
    }
  }                    // end of loop through magnetic arrows
  for (k=savedNvertM; k<NvertM; k++) {
    float tmp, tmp1;
    i = drvui->atom_no[k]/100-MAX_ATOM;     // get number of magnetic item
    for (l = 0; l<3; l++) {
        x[l] = drvui->mag_xc[i][l];
    }
    double t = sqrt(x[0] * x[0] + x[1] * x[1] + x[2] * x[2]);
    if (t < 0.002) continue;      // skip this arrow if 0 length
    for (l=0; l<3; l++) {
        x[l] /= (float)t;
    }
    t = sqrt(x[0] * x[0] + x[1] * x[1]);
    chi = (float) atan2(t, (double) x[2]);
    if (t > 0.001)
        phi = (float) atan2(x[0]/t,x[1]/t);
    else
        phi = 0.0f;
    glLoadName(200000+i);
    glPushName(k);
    glPushMatrix ();
    (void) sscanf (drvui->col_arrow_v[i], "%f %f %f", &glr, &glg, &glb);
    glColor3f (glr, glg, glb);
    glTranslatef (s_vert[3*k], s_vert[3*k+1], 
            s_vert[3*k+2]);
    glRotatef (-phi* (float) RAD, 0.0, 0.0, 1.0f);
    glRotatef (-chi* (float) RAD, 1.0f, 0., 0.);
    glTranslatef(0.0,0.0,-drvui->arrow_length[i]/2.0f);
    glu_quadObj = gluNewQuadric ();
    gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
    gluCylinder (glu_quadObj, drvui->arrow_diam[i],
              drvui->arrow_diam[i], 0.9 * drvui->arrow_length[i], 10, 1);
#if !defined (WIN32) && !defined (__APPLE__)
    free(glu_quadObj);
#endif
    glPopMatrix();
    glPushMatrix();
    glTranslatef (s_vert[3*k], s_vert[3*k+1], 
              s_vert[3*k+2]);
    glRotatef (-phi* (float) RAD, 0.0, 0.0, 1.0f);
    glRotatef (-chi* (float) RAD, 1.0f, 0., 0.);
    glTranslatef(0.0f,0.0f, drvui->arrow_length[i] * 0.27f);
	
    glu_quadObj = gluNewQuadric ();
    gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
    gluCylinder (glu_quadObj, drvui->arrow_diam[i] * 2.0,
   	   	       0.01, 0.4*drvui->arrow_length[i], 10, 1);
#if !defined (WIN32) && !defined (__APPLE__)
    free(glu_quadObj);
#endif
    glPopMatrix ();
    glPopName();

// arrow in POV

    tmp = 0.5f * drvui->arrow_length[i];
    base = 2.0f * drvui->arrow_diam[i];
    fprintf (drvui->fpoutp, " object{\n");
    fprintf (drvui->fpoutp, "   union{\n");
    fprintf (drvui->fpoutp,"     object{\n");
    fprintf (drvui->fpoutp,"      cylinder{ < 0,0,%8.5f > , < 0,0,%8.5f >,%8.5f\n",
         -0.5f * drvui->arrow_length[i], 0.27f * drvui->arrow_length[i],drvui->arrow_diam[i]);
    fprintf (drvui->fpoutp,
         "         texture{pigment{color %s}}\n", drvui->col_arrow_p[i]);
    fprintf (drvui->fpoutp, "     finish{phong %5.2f phong_size %5.2f}}\n     }\n",
         drvui->Phong_Value, drvui->Phong_Size);
    tmp1 = 0.27f * drvui->arrow_length[i];
    tmp = 0.77f * drvui->arrow_length[i];
    fprintf (drvui->fpoutp, "     object{\n");
    fprintf (drvui->fpoutp, "      cone{<0, 0, %8.5f>, %8.5f, <0, 0, %8.5f>,  0 \n",
             tmp1, base, tmp); 
    fprintf (drvui->fpoutp,
		     "         texture{pigment{color %s}}\n", drvui->col_arrow_p[i]);
    fprintf (drvui->fpoutp, "    finish{phong %5.2f phong_size %5.2f}}\n     }\n",
         drvui->Phong_Value, drvui->Phong_Size);
    fprintf (drvui->fpoutp,"   rotate <%8.5f,0,0>\n", -chi * RAD);
    fprintf (drvui->fpoutp,"   rotate <0,0,%8.5f>\n", -phi * RAD);
    fprintf (drvui->fpoutp,"   translate <%8.5f, %8.5f, %8.5f>\n }}\n",
             s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);

  	// arrow in VRML
    float d1, d2, d3, alpha, temp;
    float cosalp;
    d1 = x[1];
    d2 = -x[0];
    d3 = 0.0f;
    temp = (float) sqrt(d1*d1 + d2 * d2 + d3*d3 );
    if (temp < 0.001f) {
        d1 = 1.0;
        temp = 1.0;
    }
    d1 /= temp;
    d2 /= temp;
    d3 /= temp;
    cosalp = x[2]/(float)sqrt(x[0] * x[0] + x[1] * x[1] + x[2] * x[2]);;
    alpha = -(float) acos (cosalp);
    if (Vrml2) {
// cylinder for shaft of arrow
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "    rotation %6.3f %6.3f %6.3f %7.3f\n",
                d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "    translation %5.3f %5.3f %5.3f\n",
              s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "    children [\n");
      fprintf (drvui->fpoutv, "      Transform{\n");
      fprintf (drvui->fpoutv, "        rotation 1 0 0 1.5708\n");
      fprintf (drvui->fpoutv, "        translation 0 0 %5.3f\n",-0.05f * drvui->arrow_length[i]);
      fprintf (drvui->fpoutv, "        children [\n");
      fprintf (drvui->fpoutv, "          Shape {\n");
      fprintf (drvui->fpoutv, "            geometry Cylinder {\n");
      fprintf (drvui->fpoutv, "              radius %f\n", drvui->arrow_diam[i]);
      fprintf (drvui->fpoutv, "              height %f\n", 0.9f * drvui->arrow_length[i]);
      fprintf (drvui->fpoutv, "            }\n");
      fprintf (drvui->fpoutv, "            appearance Appearance { material Material "
                                               "{ diffuseColor %s}}\n",
                                               drvui->col_arrow_v[i]);
      fprintf (drvui->fpoutv, "          }\n");
      fprintf (drvui->fpoutv, "        ]\n");
      fprintf (drvui->fpoutv, "      }\n");
      fprintf (drvui->fpoutv, "    ]\n");
      fprintf (drvui->fpoutv, "  }\n");
// Cone at end of arrow
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "    rotation %6.3f %6.3f %6.3f %7.3f\n",
                d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "    translation %5.3f %5.3f %5.3f\n",
              s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "    children Transform {\n");
      fprintf (drvui->fpoutv, "      children [\n");
      fprintf (drvui->fpoutv, "        Transform{\n");
      fprintf (drvui->fpoutv, "          rotation 1 0 0 1.5708\n");
      fprintf (drvui->fpoutv, "          translation 0 0 %5.3f\n",
              0.5f * drvui->arrow_length[i]);
      fprintf (drvui->fpoutv, "          children [\n");
      fprintf (drvui->fpoutv, "            Shape {\n");
      fprintf (drvui->fpoutv, "              geometry Cone {\n");
      fprintf (drvui->fpoutv, "                bottomRadius %f\n", 2.0 * drvui->arrow_diam[i]);
      fprintf (drvui->fpoutv, "                height %f\n", drvui->arrow_length[i]*0.4);
      fprintf (drvui->fpoutv, "              }\n");
      fprintf (drvui->fpoutv, "              appearance Appearance { material Material "
                                               "{   diffuseColor %s}}\n",drvui->col_arrow_v[i]);
      fprintf (drvui->fpoutv, "            }\n");
      fprintf (drvui->fpoutv, "          ]\n");
      fprintf (drvui->fpoutv, "        }\n");
      fprintf (drvui->fpoutv, "      ]\n");
      fprintf (drvui->fpoutv, "    }\n");
      fprintf (drvui->fpoutv, "  }\n");
    } else { // VRML 1
      fprintf (drvui->fpoutv, " Separator{\n");
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "   translation %5.3f %5.3f %5.3f\n",
              s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "   rotation %6.3f %6.3f %6.3f %7.3f\n  }\n",
                d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "  Material {diffuseColor %s}\n", drvui->col_arrow_v[i]);
      fprintf (drvui->fpoutv, "  Rotation{rotation 1 0 0 1.5708}\n");
      fprintf (drvui->fpoutv, "  Cylinder {\n");
      fprintf (drvui->fpoutv, "   parts SIDES\n");
      fprintf (drvui->fpoutv, "   radius %f\n", drvui->arrow_diam[i]);
      fprintf (drvui->fpoutv, "   height %f\n", 0.9f * drvui->arrow_length[i]);
      fprintf (drvui->fpoutv, "  }\n }\n");
      fprintf (drvui->fpoutv, " Separator{\n");
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "   translation %5.3f %5.3f %5.3f\n",
              s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "   rotation %6.3f %6.3f %6.3f %7.3f\n  }\n",
                d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "  Material {diffuseColor %s}\n", drvui->col_arrow_v[i]);
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "   rotation 1 0 0 1.5708\n");
      fprintf (drvui->fpoutv, "   translation 0 0 %5.3f\n  }\n",
              0.5f * drvui->arrow_length[i]);
      fprintf (drvui->fpoutv, "  Cone {\n");
      fprintf (drvui->fpoutv, "   parts SIDES\n");
      fprintf (drvui->fpoutv, "   bottomRadius %f\n",2.0 * drvui->arrow_diam[i]);
      fprintf (drvui->fpoutv, "   height %f\n", drvui->arrow_length[i]*0.4);
      fprintf (drvui->fpoutv, "  }\n }\n");
    }
    Arrow_Count++;
  }


  if (Arrow_Count > 0) {
    fprintf (drvui->fcns,"%4d magnetic moment arrows.\n", Arrow_Count);
    fprintf (drvui->flout, "Generated %4d arrows.\n", Arrow_Count);
  }
  NvertM = savedNvertM;  // restore vertex count
}						 // end of generate_arrows

/* ************************************************************** */
/* ************************************************************** */

void generate_bonds (void)

/* routine to generate bond descriptions and add then to edit list */
{
  int start, same_type;
  int i, j, k;        			 /* loop counters */
  int N_Bond;        			 /* Counter for number of bonds of a type */
  int l1;        			 /* loop counter for matrix operations */
  int nvert1;        			 /* vertex counter */
  int Bond_Count;        		 /* Counter for bonds output */
  char bnd1[4], bnd2[4];        	 /* temporary storage of bond labels */
  float d;        			 /* bond distance */
  float df[3];        			 /* vector difference */
  float beta;        			 /* rotation angle */
  float gamma;        			 /* another rotation angle */
  float at1[3], at2[3];        		 /*initial coordinates of bond endpoints */
  float clip0, clip1;        		 /* clipping scale factors */
  int m, l2, l3, out = 0;
  int ell[2];        			 /* save ellipsoid numbers */

  int is_ellipsoid[2];        		 /* ellipsoid markers - for moving bond endpoint */
  int o1, p, p1, q, r;        		 /* loop variables for rotation matrix calculation */
  float elrot0[3][3];        		 /* rotation matrices of ellipsoids */
  float Z0[3], Z1[3];        		 /* components of rms displacement along bond */
  int *ellips_id;
  float factor, rms0, rms1;
  float glr, glg, glb;
  GLUquadricObj *glu_quadObj;
  int omit;

  if (drvui->nbond == 1)
    return;        			 /* exit if no work to be done */

  Bond_Count = 0;
  if (!(ellips_id = (int *) malloc ((unsigned) ((long) (MAX_VERTS * sizeof (int)))))) {
    Error_Box("Unable to get ellips_id allocation!");
    return;
  }
  for (i = 1; i < drvui->nbond; ++i) {        	 /* loop through bond types */
    if (drvui->bond_fn[i] != drvui->frame_no) continue;  // skip if not in this frame
    nvert = 0;        			         /* empty the vertex list */
    for (j = 0; j < natom; j++) {        /* loop through atom types */
      if (drvui->atom_fn[j] != drvui->frame_no) continue;
      int o = drvui->atom_n[j];
      if (((o & 255) > 0) || (((o >> 24) & 255) > 0)) { // sphere or ellipsoid

        o = (drvui->atom_n[j] >> 24) & 255;
        if (o > 0) {                      // ellipsoid
          if (check_atom_name(drvui->bond_l1[i],drvui->ellips_l[o])) {
            if (do_ellipsoids == 1 && El_Cutout == 1) {    /* skip if B iso */
              is_ellipsoid[0] = 1;         /* or no cutout */
              ell[0] = o;
            }
            o = nvert;
            find_all_in_box (j);          // get all atoms of type j
            for (p = o; p < nvert; p++)
              ellips_id[p] = ell[0];      // save ellipsoid id for these entries
          }
        } else {                          // sphere
          o = drvui->atom_n[j] & 255;     // number of sphere
          is_ellipsoid[0] = 0;
          if (check_atom_name(drvui->bond_l1[i], drvui->sphere_l[o])) {
            find_all_in_box (j);         /* get all atoms of type j */
          }
        }        			 /* o > 0 */
      }
    }        				 /* for (j=0;j<natom;++j) */

    bnd1[0] = drvui->bond_l1[i][0];
    bnd1[1] = drvui->bond_l1[i][1];
    bnd1[2] = drvui->bond_l1[i][2];
    bnd1[3] = drvui->bond_l1[i][3];
    nvert1 = nvert;
    if (check_atom_name(drvui->bond_l1[i], drvui->bond_l2[i])) {
      bnd2[0] = drvui->bond_l1[i][0];        	 /* atoms are same type */
      bnd2[1] = drvui->bond_l1[i][1];
      bnd2[2] = drvui->bond_l1[i][2];
      bnd2[3] = drvui->bond_l1[i][3];
      is_ellipsoid[1] = is_ellipsoid[0];
      ell[1] = ell[0];
      same_type = 1;
    } else {
      bnd2[0] = drvui->bond_l2[i][0];
      bnd2[1] = drvui->bond_l2[i][1];
      bnd2[2] = drvui->bond_l2[i][2];
      bnd2[3] = drvui->bond_l2[i][3];
      same_type = 0;
/* atoms i and j are not the same type */
      for (j = 0; j < natom; ++j) {         /* loop through atom types */
        if (drvui->atom_fn[j] != drvui->frame_no) continue;
        if (((drvui->atom_n[j] & 255) > 0) || (((drvui->atom_n[j] >> 24) & 255) > 0)) { // sphere or ellipsoid
          int o;

          o = (drvui->atom_n[j] >> 24) & 255;
          if (o > 0) {                      // ellipsoid
            if (check_atom_name(drvui->bond_l2[i], drvui->ellips_l[o])) {
              if (do_ellipsoids == 1 && El_Cutout == 1) {    /* skip if B iso */
                is_ellipsoid[1] = 1;         /* or no cutout */
                ell[1] = o;
              }
              o = nvert;
              find_all_in_box (j);         /* get all atoms of type j */
              for (p = o; p < nvert; p++)
                ellips_id[p] = ell[1];         /* save ellipsoid type */
            }
          } else {
            o = drvui->atom_n[j] & 255;               // sphere number
            if (check_atom_name(drvui->bond_l2[i], drvui->sphere_l[o])) {
              find_all_in_box (j);         /* get all atoms of type j */
              is_ellipsoid[1] = 0;
            }
          }        			 // o > 0
        }
      }        				 /* for (j=0;j<natom;++j) */
    }        				 /* if atoms are same type */

    if (((nvert1 > 0) && (nvert > nvert1)) || (same_type == 1)) {
      N_Bond = 0;
      for (j = 0; j < nvert1; ++j) {         /* loop through atoms of type 1 */
        start = nvert1;
        if (same_type == 1)
          start = j;
        for (k = start; k < nvert; ++k) {     /* loop on type 2 atoms  */
          d = dist (j, k);        	 /* get atomic distance */
          if ((d >= drvui->bond_min[i]) && (d <= drvui->bond_max[i])) {    /* here if distance within range */
            df[0] = df[1] = drvui->Bond_Mult * drvui->bond_size[i];    /* scaling */
            df[2] = 0.5f *d;

            Bond_Count++;        	 /* update count */
            for (p = 0; p < 3; p++) {         /* initialize endpoints of bond */
              Z0[p] = s_vert[3 * j + p];
              Z1[p] = s_vert[3 * k + p];
            }

            if (do_ellipsoids == 1 && El_Cutout == 1) {

/* Elipsoid has octant cutout, calculate rms displacement along bond  */

              factor = (float) (Ellipsoid_Scale * 0.9f / (PI * d * 1.414));    /* calculate multiplier */

/* Mean-square displacement of a scaled thermal ellipsoid along the direction of a bond is given by: */
/*                                                                                      */
/*      D = factor^2 * (X'G'R'BRGX)                                                     */
/*                                                                                      */
/*  where X is the interatomic vector given in triclinic fractional coordinates,        */
/*        G is the lattice metric with the ij'th element given by a[i].a[j],            */
/*        R is the rotational part of the symmetry operation that transform atom i,     */
/*        B is a square matrix of the betaij thermal factors,                           */
/*    and a prime indicates the transpose of the matrix.                                */
/*                                                                                      */
              if (is_ellipsoid[0] == 1) {     /* at one end of bond */
                float B[3][3], BP[3][3];

                o1 = vert_sym_no[j];
                p1 = ellips_id[j];         /* have original ellipsoid number */
                for (p = 0; p < 3; p++)
                  B[p][p] = drvui->ellips[p1][p];   // extract betaij's into square matrix
                B[0][1] = B[1][0] = drvui->ellips[p1][3];
                B[0][2] = B[2][0] = drvui->ellips[p1][4];
                B[1][2] = B[2][1] = drvui->ellips[p1][5];
                for (p = 0; p < 3; p++) {
                  df[p] = 0.0;
                  for (q = 0; q < 3; ++q) {
                    elrot0[p][q] = 0.0;
                    df[p] += drvui->ginv[q][p] * (o_vert[3 * k + q] - o_vert[3 * j + q]);
                    for (r = 0; r < 3; r++)
                      elrot0[p][q] += drvui->rss[o1][r][p] * B[r][q];
                  }
                }
                for (p = 0; p < 3; p++) {
                  for (q = 0; q < 3; q++) {
                    BP[p][q] = 0.0;
                    for (r = 0; r < 3; r++)
                      BP[p][q] += (float) elrot0[p][r] * drvui->rss[o1][r][q];    /* BP is rotated betij matrix */
                  }
                }
                for (p = 0; p < 3; p++) {
                  Z0[p] = 0.0;
                  for (q = 0; q < 3; q++)
                    Z0[p] += df[q] * BP[q][p];
                }
/* rms0 is the scaled rms displacement value of the ellipsoid in the direction of the bond */
                rms0 = factor * (float) sqrt (Z0[0] * df[0] +
                          Z0[1] * df[1] + Z0[2] * df[2]);
/* calculate components of end point at intersection of ellipsoid and bond */
                for (q = 0; q < 3; q++)
                  Z0[q] = s_vert[3 * j + q] + (s_vert[3 * k + q] - s_vert[3 * j + q]) * rms0 / d;
              }        			 /* if is_ellipsoid[0] */

              if (is_ellipsoid[1] == 1) {     /* ellipsoid at other end of bond */
                float B[3][3], BP[3][3];

                o1 = vert_sym_no[k];
                p1 = ellips_id[k];         /* have original ellipsoid number */
                for (p = 0; p < 3; p++)
                  B[p][p] = drvui->ellips[p1][p];     /* extract betaij's into square matrix */
                B[0][1] = B[1][0] = drvui->ellips[p1][3];
                B[0][2] = B[2][0] = drvui->ellips[p1][4];
                B[1][2] = B[2][1] = drvui->ellips[p1][5];
                for (p = 0; p < 3; p++) {
                  df[p] = 0.0;
                  for (q = 0; q < 3; q++) {
                    elrot0[p][q] = 0.0;
                    df[p] += drvui->ginv[q][p] * (o_vert[3 * k + q] - o_vert[3 * j + q]);
                    for (r = 0; r < 3; ++r)
                      elrot0[p][q] += drvui->rss[o1][r][p] * B[r][q];
                  }
                }
                for (p = 0; p < 3; p++) {
                  for (q = 0; q < 3; q++) {
                    BP[p][q] = 0.0;
                    for (r = 0; r < 3; r++)
                      BP[p][q] += (float) elrot0[p][r] * drvui->rss[o1][r][q];    /* BP is rotated betij matrix */
                  }
                }

                for (p = 0; p < 3; p++) {
                  Z1[p] = 0.0;
                  for (q = 0; q < 3; q++)
                    Z1[p] += df[q] * BP[q][p];
                }
/* rms1 is the scaled rms displacement value of the ellipsoid in the direction of the bond */

                rms1 = factor * (float) sqrt (Z1[0] * df[0] +
        			      Z1[1] * df[1] + Z1[2] * df[2]);

/* calculate components of end point at intersection of ellipsoid and bond */
                for (q = 0; q < 3; q++)
                  Z1[q] = s_vert[3 * k + q] - (s_vert[3 * k + q] - s_vert[3 * j + q]) * rms1 / d;
              }        			 /* if is_ellipsoid[1] */
            }        			 /* if any segmented ellipsoids present */

            for (l1 = 0; l1 < 3; l1++) {
              at1[l1] = (float) Z0[l1];
              at2[l1] = (float) Z1[l1];
              df[l1] = (float) (Z1[l1] - Z0[l1]);
            }

            if (clipflag != 0) {
              out = 0;
              l2 = 0;
              l3 = 0;
              clip0 = 0.0f;
              clip1 = 1.0f;
              for (l1 = 0; l1 < 3; l1++) {
                if (o_vert[3 * j + l1] < drvui->clip_lim[drvui->frame_no][l1] - 0.01
                      || o_vert[3 * j + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3] + 0.01)
                  l2 = 1;
                if (o_vert[3 * k + l1] < drvui->clip_lim[drvui->frame_no][l1] - 0.01
                      || o_vert[3 * k + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3] + 0.01)
                  l3 = 1;
              }
              if (l2 && l3)
                out = 1;        	 /* both atoms outside */
              if (l3)
                clip1 = 0.5f;
              if (l2)
                clip0 = 0.5f;

              at2[0] = at1[0] + clip1 * df[0];
              at2[1] = at1[1] + clip1 * df[1];
              at2[2] = at1[2] + clip1 * df[2];

              at1[0] = at1[0] + clip0 * df[0];
              at1[1] = at1[1] + clip0 * df[1];
              at1[2] = at1[2] + clip0 * df[2];
            }        			 /* if clipflag */

            omit = 0;
            for (l1 = 0; l1 < Omit->nomits; l1++) {
              if (Omit->omit1[l1]== i*1000 && Omit->omit2[l1] == Bond_Count) omit = 1;
            }

            if ((clipflag == 0 || out == 0) && omit == 0) {
              fprintf (drvui->fpoutp," /* Bond: %c%c %8.5f %8.5f %8.5f TO %c%c %8.5f %8.5f %8.5f*/\n",
                 bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2],
                 bnd2[0], bnd2[1], o_vert[3 * k],o_vert[3 * k + 1], o_vert[3 * k + 2]);
              if (drvui->bond_style[i] == 0) {
                fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                   at1[0], at1[1], at1[2], at2[0], at2[1],at2[2], drvui->Bond_Mult * drvui->bond_size[i]);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", drvui->col_bond_p[i]);
                fprintf(drvui->fpoutp," }\n");
              } else {
                for (m = 0; m < 3; m++)
                  df[m] = (float) (at2[m] - at1[m]);
                fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                   at1[0], at1[1], at1[2], at1[0]+df[0]/5.,at1[1]+df[1]/5.,at1[2]+df[2]/5., drvui->Bond_Mult * drvui->bond_size[i]);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", drvui->col_bond_p[i]);
                fprintf(drvui->fpoutp," }\n");
                fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                   at1[0]+df[0]*2./5., at1[1]+df[1]*2./5, at1[2]+df[2]*2./5., 
                   at1[0]+df[0]*3./5.,at1[1]+df[1]*3./5.,at1[2]+df[2]*3/5., 
                   drvui->Bond_Mult * drvui->bond_size[i]);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", drvui->col_bond_p[i]);
                fprintf(drvui->fpoutp," }\n");
                fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                   at1[0]+df[0]*4./5., at1[1]+df[1]*4./5, at1[2]+df[2]*4./5., 
                   at2[0],at2[1],at2[2], drvui->Bond_Mult * drvui->bond_size[i]);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", drvui->col_bond_p[i]);
                fprintf(drvui->fpoutp," }\n");
              }
              N_Bond++;
              if (no_comment == 0)
                fprintf (drvui->fpoutv,"# Bond: %c%c %8.5f %8.5f %8.5f TO %c%c %8.5f %8.5f %8.5f\n",
        	  bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2],
        	  bnd2[0], bnd2[1], o_vert[3 * k],o_vert[3 * k + 1], o_vert[3 * k + 2]);
              for (m = 0; m < 3; m++)
                df[m] = (float) (at2[m] - at1[m]);
              d = (float) sqrt (df[0] * df[0] + df[1] * df[1] +    df[2] * df[2]);
/*
   calculate rotation matrices needed to rotate bond about Z onto Y-Z plane and then about X so that
   bond is parallel to Z.  Finally, apply inverse transformation to cylinder, which started parallel
   to X.  Resulting cylinder will be parallel to vector between atoms.
*/
              beta = (float) atan2 (df[0], df[1] + 0.0000001f);    /* rotation angle about Z (in radians) */
              gamma = (float) sqrt (df[0] * df[0] + df[1] * df[1]);
              gamma = (float) atan2 (gamma, df[2]);    /* Rotation angle about X */
              glu_quadObj = gluNewQuadric ();
              glPushMatrix ();
              glLoadName(i*1000);
              glPushName(Bond_Count);
              (void)sscanf (drvui->col_bond_v[i], "%f %f %f", &glr, &glg, &glb);
              glColor3f (glr, glg, glb);
              if (drvui->bond_style[i] == 0) {
                glPushMatrix ();
                glTranslatef (at1[0], at1[1], at1[2]);
                glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.);
                glRotatef (-gamma * (float) RAD, 1., 0.0f, 0.0f);
                gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
                gluCylinder (glu_quadObj,drvui->Bond_Mult * drvui->bond_size[i],
                         drvui->Bond_Mult * drvui->bond_size[i], d, 10, 1);
                glPopMatrix ();
              } else {
                glPushMatrix ();
                glTranslatef (at1[0], at1[1], at1[2]);
                glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.);
                glRotatef (-gamma * (float) RAD, 1., 0.0f, 0.0f);
                gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
                gluCylinder (glu_quadObj,drvui->Bond_Mult * drvui->bond_size[i],
                           drvui->Bond_Mult * drvui->bond_size[i], d/5., 10, 1);
                glPopMatrix ();
                glPushMatrix ();
                glTranslatef (at1[0]+df[0]*0.4f, at1[1]+df[1]*0.4f,at1[2]+df[2]*0.4f);
                glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.);
                glRotatef (-gamma * (float) RAD, 1., 0.0f, 0.0f);
                gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
                gluCylinder (glu_quadObj,drvui->Bond_Mult * drvui->bond_size[i],
                           drvui->Bond_Mult * drvui->bond_size[i], d/5., 10, 1);
                glPopMatrix ();
                glPushMatrix ();
                glTranslatef (at1[0]+df[0]*0.8f,at1[1]+df[1]*0.8f,at1[2]+df[2]*0.8f);
                glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.);
                glRotatef (-gamma * (float) RAD, 1., 0.0f, 0.0f);
                gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
                gluCylinder (glu_quadObj,drvui->Bond_Mult * drvui->bond_size[i],
                           drvui->Bond_Mult * drvui->bond_size[i], d/5., 10, 1);
                glPopMatrix ();
              }
              glPopName();
              glPopMatrix();
#if !defined (WIN32) && !defined (__APPLE__)
              free(glu_quadObj);
#endif
              if (Vrml2) {
                if (drvui->bond_style[i] == 0) {
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                    (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
                  fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 - gamma);
                  fprintf (drvui->fpoutv, " children [ Shape {\n");
                  fprintf (drvui->fpoutv," appearance Appearance { material Material {   diffuseColor %s} }\n",
          	     drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  geometry Cylinder {");
                  fprintf (drvui->fpoutv, "radius %f", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, " height %f", d);
                  fprintf (drvui->fpoutv, "}}]}}\n");
              } else {
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                      at1[0] + df[0] / 5.,at1[1] + df[1] / 5., at1[2] + df[2] / 5.);
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
                  fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 - gamma);
                  fprintf (drvui->fpoutv, " children [ Shape {\n");
                  fprintf (drvui->fpoutv," appearance Appearance { material Material {   diffuseColor %s} }\n",
                      drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  geometry Cylinder {");
                  fprintf (drvui->fpoutv, "radius %f", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, " height %f", d/5.);
                  fprintf (drvui->fpoutv, "}}]}}\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                      (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
                  fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 - gamma);
                  fprintf (drvui->fpoutv, " children [ Shape {\n");
                  fprintf (drvui->fpoutv," appearance Appearance { material Material {   diffuseColor %s} }\n",
                      drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  geometry Cylinder {");
                  fprintf (drvui->fpoutv, "radius %f", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, " height %f", d/5.);
                  fprintf (drvui->fpoutv, "}}]}}\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                      at1[0] + df[0] *4. / 5., at1[1] + df[1] * 4. / 5., at1[2] + df[2] *4. / 5.);
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
                  fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 - gamma);
                  fprintf (drvui->fpoutv, " children [ Shape {\n");
                  fprintf (drvui->fpoutv," appearance Appearance { material Material {   diffuseColor %s} }\n",
                      drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  geometry Cylinder {");
                  fprintf (drvui->fpoutv, "radius %f", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, " height %f", d/5.);
                  fprintf (drvui->fpoutv, "}}]}}\n");
                }
              } else {
                if (drvui->bond_style[i] == 0) {
                  fprintf (drvui->fpoutv, " Separator{\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",
        	     (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 - gamma);
                  fprintf (drvui->fpoutv,"  Material {\n   diffuseColor %s\n", drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  }\n");
                  fprintf (drvui->fpoutv, "  Cylinder {\n");
                  fprintf (drvui->fpoutv, "   parts SIDES\n");
                  fprintf (drvui->fpoutv, "   radius %f\n", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, "   height %f\n", d);
                  fprintf (drvui->fpoutv, "  }\n }\n");
                } else {
                  fprintf (drvui->fpoutv, " Separator{\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",
                       at1[0] + df[0] / 5., at1[1] + df[1] / 5., at1[2] + df[2] / 5.);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 - gamma);
                  fprintf (drvui->fpoutv,"  Material {\n   diffuseColor %s\n", drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  }\n");
                  fprintf (drvui->fpoutv, "  Cylinder {\n");
                  fprintf (drvui->fpoutv, "   parts SIDES\n");
                  fprintf (drvui->fpoutv, "   radius %f\n", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, "   height %f\n", d/5.);
                  fprintf (drvui->fpoutv, "  }\n }\n");
                  fprintf (drvui->fpoutv, " Separator{\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",
                       (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 - gamma);
                  fprintf (drvui->fpoutv,"  Material {\n   diffuseColor %s\n", drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  }\n");
                  fprintf (drvui->fpoutv, "  Cylinder {\n");
                  fprintf (drvui->fpoutv, "   parts SIDES\n");
                  fprintf (drvui->fpoutv, "   radius %f\n", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, "   height %f\n", d/5.);
                  fprintf (drvui->fpoutv, "  }\n }\n");
                  fprintf (drvui->fpoutv, " Separator{\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",
                      at1[0] + df[0] * 4./ 5., at1[1] + df[1] * 4./ 5., at1[2] + df[2] *4. / 5.);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 - gamma);
                  fprintf (drvui->fpoutv,"  Material {\n   diffuseColor %s\n", drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  }\n");
                  fprintf (drvui->fpoutv, "  Cylinder {\n");
                  fprintf (drvui->fpoutv, "   parts SIDES\n");
                  fprintf (drvui->fpoutv, "   radius %f\n", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, "   height %f\n", d/5.);
                  fprintf (drvui->fpoutv, "  }\n }\n");
                }
              }
            }
          }        			 /* if distance OK */
        }        			 /* for k */
      }        				 /* for j */
    }        				 /* if nvert */
  }        				 /* for i (bond types) */
  free (ellips_id);
  fprintf (drvui->fcns,"%4d bonds.\n", Bond_Count);
  fprintf (drvui->flout, "\nGenerated %4d bonds.\n\n", Bond_Count);
}        				 /* end of generate_bonds */

/* ************************************************************** */
/* ************************************************************** */

void generate_cones (void)

/* routine to generate lone pair cone descriptions and add them to edit list */
{
  int start, same_type;
  int i, j, k, l;        		 /* loop counters */
  int N_Bond;        			 /* Counter for number of bonds of a type */
  int l1;        			 /* loop counter for matrix operations */
  int nvert1;        			 /* vertex counter */
  int Bond_Count;        		 /* Counter for bonds output */
  char bnd1[4];        			 /* temporary storage of bond labels */
  float d = 0;        			 /* bond distance */
  float df[3];        			 /* vector difference */
  float beta;        			 /* rotation angle */
  float gamma;        			 /* another rotation angle */
  float at1[3], at2[3];        		 /*initial coordinates of bond endpoints */
  float clip0, clip1;        		 /* clipping scale factors */
  int m, l2, l3, out = 0;

  int p;        			 /* loop variables for rotation matrix calculation */
  float Z0[3], Z1[3];        		 /* components of rms displacement along bond */
  float mybonds[20][20];        	 /*temporary storage for atoms bonded to lonepair site */
  int nbnds = 0;
  float vecsum[3], vecpro[3];
  float glr, glg, glb;
  GLUquadricObj *glu_quadObj;

  if (drvui->ncone == 1)
    return;        			 /* exit if no work to be done */

  Bond_Count = 0;
  for (i = 1; i < drvui->ncone; ++i) {        	 // loop through cone types
    if (drvui->cone_fn[i] != drvui->frame_no) continue;  // skip if wrong frame
    nvert = 0;        			         // empty the vertex list
    for (j = 0; j < natom; ++j) {        // loop through atom types
      if (drvui->atom_fn[j] != drvui->frame_no) continue;
      if (((drvui->atom_n[j] & 255) > 0) || (((drvui->atom_n[j] >> 24) & 255) > 0)) { // sphere or ellipsoid
        int o;

        o = drvui->atom_n[j] >> 24;
        if (o > 0) {                     // ellipsoid
          if (check_atom_name(drvui->cone_l1[i], drvui->ellips_l[o]))
          find_all_in_box (j);         /* get all atoms of type j */
        } else {
          o = drvui->atom_n[j] & 255;
          if (check_atom_name(drvui->cone_l1[i], drvui->sphere_l[o])) {
            find_all_in_box (j);         /* get all atoms of type j */
          }
        }        			 // o > 0
      }
    }        				 /* for (j=0;j<natom;++j) */

    bnd1[0] = drvui->cone_l1[i][0];
    bnd1[1] = drvui->cone_l1[i][1];
    bnd1[2] = drvui->cone_l1[i][2];
    bnd1[3] = drvui->cone_l1[i][3];
    nvert1 = nvert;
    same_type = 0;

/* atoms i and j are not the same type */
    for (j = 0; j < natom; ++j) {         /* loop through atom types */
      if (drvui->atom_fn[j] != drvui->frame_no) continue;
      if (((drvui->atom_n[j] & 255) > 0) || (((drvui->atom_n[j] >> 24) & 255) > 0)) { // sphere or ellipsoid
        find_all_in_box (j);        	 /* get all atoms of type j */
      }
    }        				 /* for (j=0;j<natom;++j) */

    if (((nvert1 > 0) && (nvert > nvert1)) || (same_type == 1)) {
      N_Bond = 0;
      nbnds = 0;
      for (j = 0; j < nvert1; ++j) {         /* loop through atoms of type 1 */
        nbnds = 0;
        start = nvert1;
        if (same_type)
          start = j;
        start = 0;
        for (k = start; k < nvert; ++k) {     /* loop on type 2 atoms  */
          d = dist (j, k);        	 /* get atomic distance */
          if (d > 0.2 && d <= drvui->cone_height[i]) {     /* here if distance within range */
            mybonds[0][nbnds] = s_vert[3 * k];
            mybonds[1][nbnds] = s_vert[3 * k + 1];
            mybonds[2][nbnds] = s_vert[3 * k + 2];
            nbnds++;
          }        			 /* if distance ok */
        }        			 /* for k */
    /*# */
        if (nbnds < 4 - drvui->numlonepairs[i]) {
          continue;
    }
        df[0] = df[1] = drvui->Bond_Mult * drvui->bond_size[i];    /* scaling */
        df[2] = 0.5f *d;

        Bond_Count++;        		 /* update count */
        for (p = 0; p < 3; p++) {         /* initialize endpoints of bond */
          Z0[p] = s_vert[3 * j + p];
        }

        for (l = 0; l < drvui->numlonepairs[i]; l++) {
          for (p = 0; p < 3; p++) {         /* initialize endpoints of bond */
            Z1[p] = 0;

            if (drvui->numlonepairs[i] == 1) {
              for (k = 0; k < nbnds; k++)
                Z1[p] -= mybonds[p][k] - s_vert[3 * j + p];
              Z1[p] = Z0[p] + Z1[p];
            } else {
              if (drvui->numlonepairs[i] == 2) {
                vecsum[p] = 0.;
                vecpro[p] = 0.;
                for (k = 0; k < nbnds; k++) {
                  vecsum[p] += mybonds[p][k] - s_vert[3 * j + p];
                }

                vecpro[0] = (s_vert[3 * j + 1] - mybonds[1][0])
                     * (s_vert[3 * j + 2] - mybonds[2][1])
                     - (s_vert[3 * j + 1] - mybonds[1][1])
                     * (s_vert[3 * j + 2] - mybonds[2][0]);
                vecpro[1] = (s_vert[3 * j + 2] - mybonds[2][0])
                     * (s_vert[3 * j] - mybonds[0][1])
                     - (s_vert[3 * j + 2] - mybonds[2][1])
                     * (s_vert[3 * j] - mybonds[0][0]);
                vecpro[2] = (s_vert[3 * j] - mybonds[0][0])
                     * (s_vert[3 * j + 1] - mybonds[1][1])
                     - (s_vert[3 * j] - mybonds[0][1])
                     * (s_vert[3 * j + 1] - mybonds[1][0]);
                switch (l) {
                case 0:
                  Z1[p] = -vecsum[p] + vecpro[p];
                  break;
                case 1:
                  Z1[p] = -vecsum[p] - vecpro[p];
                  break;
                }
                Z1[p] = Z0[p] + Z1[p];
              }
            }
          }        			 /* for p */

          d = (float) sqrt ((Z0[0] - Z1[0]) * (Z0[0] - Z1[0])
                          + (Z0[1] - Z1[1]) * (Z0[1] - Z1[1])
                          + (Z0[2] - Z1[2]) * (Z0[2] - Z1[2]));
          d = drvui->cone_height[i] / d;
          Z1[0] = Z0[0] + d * (Z1[0] - Z0[0]);
          Z1[1] = Z0[1] + d * (Z1[1] - Z0[1]);
          Z1[2] = Z0[2] + d * (Z1[2] - Z0[2]);

          for (l1 = 0; l1 < 3; l1++) {
            at1[l1] = (float) Z0[l1];
            at2[l1] = (float) Z1[l1];
            df[l1] = (float) (Z1[l1] - Z0[l1]);
          }

          if (clipflag != 0) {
            out = 0;
            l2 = 0;
            l3 = 0;
            clip0 = 0.0f;
            clip1 = 1.0f;
            for (l1 = 0; l1 < 3; l1++) {
              if (o_vert[3 * j + l1] < drvui->clip_lim[drvui->frame_no][l1] - 0.01
                 || o_vert[3 * j + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3] + 0.01)
                l2 = 1;
              if (o_vert[3 * k + l1] < drvui->clip_lim[drvui->frame_no][l1] - 0.01
                 || o_vert[3 * k + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3] + 0.01)
                l3 = 1;
            }
            if (l2 && l3)
              out = 1;        		 /* both atoms outside */
            if (l2)
              out = 1;
            if (l3)
              clip1 = 0.5f;
            if (l2)
              clip0 = 0.5f;

            at1[0] = at1[0] + clip0 * df[0];
            at1[1] = at1[1] + clip0 * df[1];
            at1[2] = at1[2] + clip0 * df[2];

            at2[0] = at1[0] + clip1 * df[0];
            at2[1] = at1[1] + clip1 * df[1];
            at2[2] = at1[2] + clip1 * df[2];

          }        			 /* if clipflag */

          if (clipflag == 0 || out == 0) {
            fprintf (drvui->fpoutp," /* LonePair on %c%c %8.5f %8.5f %8.5f */\n",
               bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2]);
            fprintf (drvui->fpoutp," cone{<%8.5f, %8.5f, %8.5f>, %8.5f, <%8.5f, %8.5f, %8.5f>,  %8.5f \n",
               at1[0], at1[1], at1[2],drvui->Bond_Mult * drvui->cone_min[i], at2[0], at2[1],
               at2[2], drvui->Bond_Mult * drvui->cone_max[i]);
            fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", drvui->col_cone_p[i]);
            fprintf(drvui->fpoutp," }\n");
            fprintf (drvui->fpoutp," sphere{<%8.5f, %8.5f, %8.5f> %8.5f \n",
               at2[0], at2[1], at2[2], drvui->Bond_Mult * drvui->cone_max[i]);
            fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", drvui->col_cone_p[i]);
            fprintf(drvui->fpoutp," }\n");
            N_Bond++;
            if (no_comment == 0)
              fprintf (drvui->fpoutv,"# LonePair on: %c%c %8.5f %8.5f %8.5f\n",
                 bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2]);
            for (m = 0; m < 3; m++)
              df[m] = (float) (at2[m] - at1[m]);
            d = (float) sqrt (df[0] * df[0] + df[1] * df[1] + df[2] * df[2]);
/*
   calculate rotation matrices needed to rotate bond about Z onto Y-Z plane and then about X so that
   bond is parallel to Z.  Finally, apply inverse transformation to cylinder, which started parallel
   to X.  Resulting cylinder will be parallel to vector between atoms.
*/
            beta = (float) atan2 (df[0], df[1] + 0.0000001f);    /* rotation angle about Z (in radians) */
            gamma = (float) sqrt (df[0] * df[0] + df[1] * df[1]);
            gamma = (float) atan2 (gamma, df[2]);    /* Rotation angle about X */
            glPushMatrix ();
            sscanf (drvui->col_cone_v[i], "%f %f %f", &glr, &glg, &glb);
            glColor3f (glr, glg, glb);
            glTranslatef (at1[0], at1[1], at1[2]);
            glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.0f);
            glRotatef (-gamma * (float) RAD, 1.0f, 0.0f, 0.0f);
            glu_quadObj = gluNewQuadric ();
            gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
            gluCylinder (glu_quadObj, drvui->Bond_Mult * drvui->cone_min[i],
                drvui->Bond_Mult * drvui->cone_max[i], d, 10, 1);
#if !defined (WIN32) && !defined (__APPLE__)
            free(glu_quadObj);
#endif
            glPopMatrix ();
            glPushMatrix ();
            glTranslatef (at2[0], at2[1], at2[2]);
            glutSolidSphere (drvui->cone_max[i], 10, 10);
            glPopMatrix ();
            if (Vrml2) {
              fprintf (drvui->fpoutv, "  Transform{\n ");
              fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                 (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
              fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
              fprintf (drvui->fpoutv,"   children Transform { rotation 0 1 0 %f\n", 3.14);
              fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 + gamma);
              fprintf (drvui->fpoutv, " children [ Shape {\n");
              fprintf (drvui->fpoutv, " appearance Appearance { material Material {   diffuseColor %s} }\n",
                 drvui->col_cone_v[i]);
              fprintf (drvui->fpoutv, "  geometry Cone {");
              fprintf (drvui->fpoutv, "bottomRadius %f", drvui->Bond_Mult * drvui->cone_max[i]);
              fprintf (drvui->fpoutv, " height %f", d);
              fprintf (drvui->fpoutv, "}}]}}}\n");
              fprintf (drvui->fpoutv, "  Transform {");
              fprintf (drvui->fpoutv," translation %8.5f %8.5f %8.5f\n",at2[0], at2[1], at2[2]);
              fprintf (drvui->fpoutv, " children [\n Shape {");
              fprintf (drvui->fpoutv,"appearance Appearance {material Material { diffuseColor %s} }\n",
                 drvui->col_cone_v[i]);
              fprintf (drvui->fpoutv,"  geometry Sphere { radius %8.5f}\n", drvui->cone_max[i]);
              fprintf (drvui->fpoutv, "}]}\n");
            } else {
              fprintf (drvui->fpoutv, " Separator{\n");
              fprintf (drvui->fpoutv, "  Transform{\n ");
              fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",(at1[0] + at2[0]) / 2.,
                 (at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
              fprintf (drvui->fpoutv, "  Rotation {\n");
              fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
              fprintf (drvui->fpoutv, "  Rotation {\n");
              fprintf (drvui->fpoutv, "   rotation 0 1 0  %f\n  }\n", 3.14);
              fprintf (drvui->fpoutv, "  Rotation {\n");
              fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 + gamma);
              fprintf (drvui->fpoutv, "  Material {\n   diffuseColor %s\n", drvui->col_cone_v[i]);
              fprintf (drvui->fpoutv, "  }\n");
              fprintf (drvui->fpoutv, "  Cone {\n");
              fprintf (drvui->fpoutv, "   parts SIDES\n");
              fprintf (drvui->fpoutv, "   bottomRadius %f\n",drvui->Bond_Mult * drvui->cone_max[i]);
              fprintf (drvui->fpoutv, "   height %f\n", d);
              fprintf (drvui->fpoutv, "  }\n }\n");
              fprintf (drvui->fpoutv, " Separator {");
              fprintf (drvui->fpoutv, "  Transform {");
              fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f}\n", at2[0], at2[1], at2[2]);
              fprintf (drvui->fpoutv, "  Material {");
              fprintf (drvui->fpoutv, " shininess 0.3\n");
              fprintf (drvui->fpoutv, " diffuseColor %s}\n", drvui->col_cone_v[i]);
              fprintf (drvui->fpoutv, "  Sphere { radius %8.5f}\n", drvui->cone_max[i]);
              fprintf (drvui->fpoutv, " }\n");
            }        			 /* if (Vrml2) */
          }        			 /* if not clipped */

        }        			 /* for l */
      }        				 /* for j */
    }        				 /* if nvert */
  }        				 /* for i */
  fprintf (drvui->fcns, "%4d cones.\n", Bond_Count);
  fprintf (drvui->flout, "\nGenerated %4d cones.\n\n", Bond_Count);
}        				 /* end of generate_bonds */

/* ************************************************************** */
/* ************************************************************** */

void get_atom_id (void)

/* routine to rework atom numbers to point at polyhedron, sphere
   plane, or ellipsoid characteristics list */
{
  int i, j;        		 /* local variables */

  for (i = 0; i < natom; ++i) {
    if (drvui->atom_fn[i] != drvui->frame_no) continue;
    drvui->sv_atom_n[i] = drvui->atom_n[i];        	 /* save original atom number */
    drvui->atom_n[i] = 0;                       // clear encoded location
    for (j = 1; j < drvui->nsphere; ++j) {
      if (check_atom_name(drvui->atom_l[i], drvui->sphere_l[j])
            && (drvui->sv_atom_n[i] == drvui->sphere_n[j] || drvui->sphere_n[j] == -1)) {
        drvui->atom_n[i] = j;        	 // atom is in sphere list
      }
    }
    for (j = 1; j < drvui->npoly; ++j) {
      if (check_atom_name(drvui->atom_l[i], drvui->poly_l[j])) {
        drvui->atom_n[i] += (j << 8);        	 // atom is in poly. list
      }
    }
    for (j = 1; j < drvui->nplane; ++j) {
      if (check_atom_name(drvui->atom_l[i], drvui->plane_l[j])) {
        drvui->atom_n[i] += (j << 16);         // atom is in plane list
      }
    }
    if (do_ellipsoids) {
      for (j = 1; j < n_ellips; ++j) {
        if (check_atom_name(drvui->atom_l[i], drvui->ellips_l[j])
            && (drvui->sv_atom_n[i] == drvui->ellips_n[j])) {
          drvui->atom_n[i] += (j << 24);         // atom is in ellipsoid list
        }
      }
    }
  }
}        				 // end of get_atom_id

/* ************************************************************** */
/* ************************************************************** */

void get_input (int Quick)

/* Routine to process input file, make lattice metric and do preliminary
   ellipsoid processing */
{
  int i, j, k;
  float biso;

  if (!Quick)
    fprintf (drvui->flout,"\n********************* Input: ************************\n");
  read_inp (Quick);
  if (Quick)
    return;
  fprintf (drvui->flout,"*****************************************************\n\n");
  if (!Labels)
    Display_axes = 0;        		 /* If no labels, no vectors */
  if (drvui->frame_no == 1)
    if (Vrml2)
      fprintf (drvui->fcns,"\nGenerating VRML97 Output.\n");
    else
      fprintf (drvui->fcns,"\nGenerating VRML1.0 Output.\n");
  print_sym ();        			 /* print symmetry information */
  if (drvui->xyzoff_read)
      drvui->origin1_flag = 0;
  (void)fclose (drvui->fpin);        		 /* close file */
  make_bmat (drvui->sys, drvui->lat_con, drvui->b_mat,
      drvui->ginv, drvui->rec_lat_con);  /* create the lattice metric */
/* set limits of search regions */
  if (packflag == 0 && boxflag == 0) {         /* neither given */
    drvui->cryst_lim[drvui->frame_no][0] = origin[0] - 0.5f;
    drvui->cryst_lim[drvui->frame_no][1] = origin[1] - 0.5f;
    drvui->cryst_lim[drvui->frame_no][2] = origin[2] - 0.5f;     /* set to generate one cell */
    drvui->cryst_lim[drvui->frame_no][3] = origin[0] + 0.5f;
    drvui->cryst_lim[drvui->frame_no][4] = origin[1] + 0.5f;
    drvui->cryst_lim[drvui->frame_no][5] = origin[2] + 0.5f;
    packflag = 1;
  }
  if (packflag != 0) {        		 /* pack command given - calculate approximate bounds */
    float temp, temp1;

    for (i = 0; i < 3; i++) {
      temp = origin[i] - drvui->cryst_lim[drvui->frame_no][i];
      temp1 = drvui->cryst_lim[drvui->frame_no][i + 3] - origin[i];
      if (temp1 > temp)
        temp = temp1;
      boxlim[i] = (temp + (float) 0.2) * drvui->lat_con[i] / 0.5f;
    }
  } else {        			 /* bounds command given - calculate approximate pack limits */

    for (i = 0; i < 3; i++) {
      drvui->cryst_lim[drvui->frame_no][i] = origin[i] - boxlim[i] / drvui->lat_con[i] - 0.2f;
      drvui->cryst_lim[drvui->frame_no][i + 3] = boxlim[i] / drvui->lat_con[i] - origin[i] + 0.2f;
    }
  }
//  if ((Options & L_OPT) != 0)
//    Calc_Rot (drvui->lookat_v1, drvui->lookat_v2); // Get rotation angles for "lookat"
  if (n_ellips > 1) {
    convert_ellipsoid ();        	 /* change Bij and Uij forms to betaij */
    fprintf (drvui->flout, "\nAnalysis of Thermal Ellipsoids:\n");
    for (i = 1; i < n_ellips; i++) {
      if (eigen (&biso, drvui->ellips[i], drvui->ellips_RMS[i], drvui->ellips_EV[i]) == 0) {
        fprintf (drvui->flout,"Ellipsoid for %c%c%c%c %d is non-positive definite and has "
           "been removed from the active list.\n", drvui->ellips_l[i][0], drvui->ellips_l[i][1],
           drvui->ellips_l[i][2], drvui->ellips_l[i][3], drvui->ellips_n[i]);
        drvui->ell_type[i] = -100;         // disable this one
        drvui->ellips_RMS[i][0] = (float) sqrt (biso / 78.957);    /* convert Biso to mu */
      } else {
        drvui->ell_type[i] += 1;
        fprintf (drvui->flout,"\n%c%c%c%c %3d Equivalent Isotropic B: %6.2f, U: %6.3f\n",
           drvui->ellips_l[i][0], drvui->ellips_l[i][1], drvui->ellips_l[i][2], 
           drvui->ellips_l[i][3], drvui->ellips_n[i], biso, biso / 78.957);
        fprintf (drvui->flout, "      RMS Amplitudes: %6.3f %6.3f %6.3f\n",
           drvui->ellips_RMS[i][0], drvui->ellips_RMS[i][1], drvui->ellips_RMS[i][2]);
        fprintf (drvui->flout, "      Eigen Vectors (by row):\n");
        for (j = 0; j < 3; j++) {
          fprintf (drvui->flout, "     ");
          for (k = 0; k < 3; k++)
            fprintf (drvui->flout, "     %8.5f", drvui->ellips_EV[i][j][k]);
          fprintf (drvui->flout, "\n");
        }
      }
    }
    Ellipsoid_Scale = P_to_C (Ellipsoid_Prob);
  }
}        				 /* end of get_input */

/* ************************************************************** */
/* ************************************************************** */

void Locate_Triple (void)

/* routine to try to place the vector triple where it will be seen in the
POV diagram */
{
  int i, j;
  float temp, pos[3], cpos[3];

  for (i = 0; i < 3; ++i) {        	 // convert origin to Cartesian
    pos[i] = 0.0f;
    for (j = 0; j < 3; ++j)
      pos[i] -= (float) drvui->b_mat[i][j] * origin[j];
  }
  for (i = 0; i <= 2; ++i) {        	 // calculate position of point after POV rotation
    cpos[i] = 0.0f;
    for (j = 0; j <= 2; ++j)
      cpos[i] += (float) G_Rot[j][i] * pos[j];
  }
  for (i = 0; i < 3; i++) {
    temp = 0.04f *(POV_Max[i] - POV_Min[i]);

    if (cpos[i] > 0)
      pos[i] = POV_Max[i] + temp;
    else
      pos[i] = POV_Min[i] - temp;
  }
// calculate coordinates of pos before rotation
  for (i = 0; i < 3; i++) {
    offset[i] = 0.0f;
    for (j = 0; j < 3; j++) {
      offset[i] += (float) G_Rot[i][j] * pos[j];
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

void make_bmat (int sys, float lat_con[6], double b_mat[3][3], float ginv[3][3],
                float rec_lat_con[6])
{
  float snal, snbe, snga, csal, csbe, csga, vol;
  int i, j;

  for (i = 3; i <= 5; ++i)
    if (lat_con[i] <= 0)
      lat_con[i] = 90.0f;
  if (sys == 5)
    lat_con[5] = 120.0f;
  for (i = 0; i <= 2; ++i)
    for (j = 0; j <= 2; ++j)
      b_mat[j][i] = 0.0f;         /* initialize matrix */
  b_mat[0][0] = 1.0f;
  csal = (float) cos (lat_con[3] / RAD);
  snal = (float) sin (lat_con[3] / RAD);
  csbe = (float) cos (lat_con[4] / RAD);
  snbe = (float) sin (lat_con[4] / RAD);
  csga = (float) cos (lat_con[5] / RAD);
  snga = (float) sin (lat_con[5] / RAD);
  for (i = 0; i < 3; i++)
    ginv[i][i] = lat_con[i] * lat_con[i];
  ginv[0][1] = ginv[1][0] = lat_con[0] * lat_con[1] * csga;
  ginv[0][2] = ginv[2][0] = lat_con[0] * lat_con[2] * csbe;
  ginv[1][2] = ginv[2][1] = lat_con[1] * lat_con[2] * csal;
  b_mat[0][1] = csga;
  b_mat[1][1] = snga;
  b_mat[0][2] = csbe;
  b_mat[1][2] = (csal - csbe * csga) / snga;
  b_mat[2][2] =    (float) sqrt (1.0 - b_mat[0][2] * b_mat[0][2] -
              b_mat[1][2] * b_mat[1][2]);
  for (i = 0; i <= 2; ++i)
    for (j = 0; j <= 2; ++j)
      b_mat[j][i] *= lat_con[i];
/* calculate cell volume and reciprocal lengths */
  vol = (float) (b_mat[0][0] * (b_mat[1][1] * b_mat[2][2] - b_mat[1][2] * b_mat[2][1])
         - b_mat[0][1] * (b_mat[1][0] * b_mat[2][2] - b_mat[1][2] * b_mat[2][0])
         + b_mat[0][2] * (b_mat[1][0] * b_mat[2][1] - b_mat[1][1] * b_mat[2][0]));
  rec_lat_con[0] = lat_con[1] * lat_con[2] * snal / vol;
  rec_lat_con[1] = lat_con[0] * lat_con[2] * snbe / vol;
  rec_lat_con[2] = lat_con[1] * lat_con[0] * snga / vol;

}        				 /* end of make_bmat */

/* ************************************************************** */
/* ************************************************************** */

int not_in_slab(float x, float y, float z)
{
// BOX_CONTAINS_POINT_3D from John Burkardt's geometry.c
  float dot;
  float normsq;

  dot = dot0_3d (slabx1,slaby1,slabz1,slabx2,slaby2,slabz2,x,y,z);
  if (dot <0.0) return 1;
	
  normsq = enormsq0_3d(slabx1,slaby1,slabz1,slabx2,slaby2,slabz2);
  if (normsq <dot) return 1;
	
  dot = dot0_3d (slabx1,slaby1,slabz1,slabx3,slaby3,slabz3,x,y,z);
  if (dot <0.0) return 1;

  normsq = enormsq0_3d(slabx1,slaby1,slabz1,slabx3,slaby3,slabz3);
  if (normsq <dot) return 1;

  dot = dot0_3d (slabx1,slaby1,slabz1,slabx4,slaby4,slabz4,x,y,z);
  if (dot <0.0) return 1;

  normsq = enormsq0_3d(slabx1,slaby1,slabz1,slabx4,slaby4,slabz4);
  if (normsq <dot) return 1;

  return 0;
}

/* ************************************************************** */
/* ************************************************************** */

void Output_Spheres (float *radii, int i)

/* routine to output the spheres with vertices defined in s_vert and o_vert
   The radius of the sphere is in radii */
{
  float glr, glg, glb;
  int k;
  int outside = 0;
  int l1;
  int omit;

  for (k = 0; k < nvert; ++k) {
    if (clipflag != 0) {

      outside = 0;
      for (l1 = 0; l1 < 3; l1++)
        if (o_vert[3 * k + l1] < drvui->clip_lim[drvui->frame_no][l1] - 0.01 ||
            o_vert[3 * k + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3] + 0.01)
          outside = 1;
    }

      omit = 0;
      for (l1 = 0; l1 < Omit->nomits; l1++) {
    if (Omit->omit1[l1]==i && Omit->omit2[l1]==k) omit = 1;
      }

    if (omit == 0 && (clipflag == 0 || outside == 0)) {
      fprintf (drvui->fpoutp," /* Sphere for %c%c%c%c at %8.5f %8.5f %8.5f */ \n",
           drvui->sphere_l[i][0], drvui->sphere_l[i][1], drvui->sphere_l[i][2],
           drvui->sphere_l[i][3], o_vert[3 * k], o_vert[3 * k + 1], o_vert[3 * k + 2]);
      fprintf (drvui->fpoutp, " sphere{<%8.5f, %8.5f, %8.5f>, %8.5f\n",
           s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2], radii[k]);
      fprintf (drvui->fpoutp, "  texture{pigment{color %s  }}\n", drvui->sphere_col_p[i]);
      fprintf (drvui->fpoutp, "   finish{phong %5.2f phong_size %5.2f}\n",
          drvui->Phong_Value, drvui->Phong_Size);
      fprintf (drvui->fpoutp, " }\n");
      glPushMatrix ();
      glLoadName(i);
      glPushName(k);
      glTranslatef (s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
      (void)sscanf (drvui->sphere_col_v[i], "%f %f %f", &glr, &glg, &glb);
      glColor3f (glr, glg, glb);
      glutSolidSphere (radii[k], 10, 10);
      glPopName();
      glPopMatrix ();
      if (no_comment == 0)
        fprintf (drvui->fpoutv,"# Sphere for %c%c%c%c at %8.5f %8.5f %8.5f \n",
          drvui->sphere_l[i][0], drvui->sphere_l[i][1], drvui->sphere_l[i][2],
          drvui->sphere_l[i][3], o_vert[3 * k], o_vert[3 * k + 1], o_vert[3 * k + 2]);
      if (Vrml2) {
        fprintf (drvui->fpoutv, "  Transform {\n");
        fprintf (drvui->fpoutv, "   translation %8.5f %8.5f %8.5f\n",
          s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
        fprintf (drvui->fpoutv, "   children [\n    Shape {\n");
        fprintf (drvui->fpoutv, "     appearance Appearance {material Material { diffuseColor %s} }\n",
          drvui->sphere_col_v[i]);
        fprintf (drvui->fpoutv, "     geometry Sphere { radius %8.5f}\n", radii[k]);
        fprintf (drvui->fpoutv, "    }\n   ]\n  }\n");
      } else {    // VRML1
        fprintf (drvui->fpoutv, " Separator {\n");
        fprintf (drvui->fpoutv, "  Transform {");
        fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f}\n",s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
        fprintf (drvui->fpoutv, "  Material {\n");
        fprintf (drvui->fpoutv, "   shininess 0.3\n");
        fprintf (drvui->fpoutv, "   diffuseColor %s\n  }\n", drvui->sphere_col_v[i]);
        fprintf (drvui->fpoutv, "  Sphere { radius %8.5f}\n }\n", radii[k]);
      }
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

void plot_vrml_poly (int polyno)
{
  int i, j, k, l, m, n;
  float *out;
  int *point, equal;
  char edgecolor[40];

/* get space for unique vertices and a list to point to them */

  if (!(out = (float *) malloc ((unsigned) ((draw_list + 10) * 3 * sizeof (float))))) {
    Error_Box("Unable to allocate space in plot_vrml_poly.");
    return;
  }
  if (!(point = (int *) malloc ((unsigned) ((draw_list + 10) * sizeof (int))))) {
    free (out);
    Error_Box("Unable to allocate space in plot_vrml_poly.");
    return;
  }
  if (!Vrml2)
    fprintf (drvui->fpoutv, "Coordinate3 { point [\n");
  n = 0;
  for (i = 0; i < draw_list; i++) {         /* build list of unique points */
    j = poly_list[i];
/*    fprintf(fpoutv,"# %d-sided polygon:\n",j); */
    for (k = i + 1; k <= i + j; k++) {
      l = poly_list[k] - 1;
      for (m = 0; m < 3; m++)
        out[3 * n + m] = s_vert[3 * l + m];
      equal = 0;
      for (m = 0; m < n; m++) {        	 /* check if new */
        if ((fabs (out[3 * m] - out[3 * n]) < 0.0002)
            && (fabs (out[3 * m + 1] - out[3 * n + 1]) < 0.0002)
            && (fabs (out[3 * m + 2] - out[3 * n + 2]) < 0.0002)) {
          point[k] = m;        		 /* point previously found */
          equal = 1;
          break;
        }
      }
      if (equal == 0)
        point[k] = n++;        		 /* new point - save it by incrementing n */
    }
    i = i + j;
  }
  --n;
  if (Vrml2)
    fprintf (drvui->fpoutv,"}\n geometry IndexedFaceSet { coord Coordinate{ point [\n");
  for (l = 0; l < n; l++)        	 /* output the unique set of coordinates */
    fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f,\n", out[3 * l], out[3 * l + 1],out[3 * l + 2]);
  fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f]\n", out[3 * n], out[3 * n + 1],out[3 * n + 2]);
  if (Vrml2) {
    fprintf (drvui->fpoutv, " }\n");
    fprintf (drvui->fpoutv, " coordIndex [\n");
  } else {
    fprintf (drvui->fpoutv, "}\n IndexedFaceSet { coordIndex [\n");
  }
  l = 0;
  for (i = 0; i < draw_list - 1; i++) {
    j = poly_list[i];
    l++;
    for (k = 0; k < j; k++)
      fprintf (drvui->fpoutv, "%d,", point[l++]);
    if (i + j < draw_list - 2) {
      fprintf (drvui->fpoutv, "-1,\n");
    } else {
      if (Vrml2) {
        fprintf (drvui->fpoutv, "-1]\n");
      } else {
        fprintf (drvui->fpoutv, "-1]\n  }\n");
      }
    }
    i = i + j;
  }
  if (Vrml2) {
    fprintf (drvui->fpoutv, " solid FALSE\n");
    fprintf (drvui->fpoutv, " convex TRUE\n");
    fprintf (drvui->fpoutv, " creaseAngle 1.5708\n");
    fprintf (drvui->fpoutv, "  }\n }\n");
  }
  if (edges > 0) {
    if (drvui->p_rad_edge[polyno] != 0.)
      strncpy (edgecolor, drvui->p_col_edge_v[polyno],39);
    else
      strncpy (edgecolor, col_edge_v,39);
    edgecolor[39] = 0;
    if (Vrml2) {
      fprintf (drvui->fpoutv, " Shape {\n");
      fprintf (drvui->fpoutv, "  geometry IndexedLineSet {\n   coord Coordinate{\n    point [\n");
      for (l = 0; l < n; l++)        	 /* output the unique set of coordinates */
        fprintf (drvui->fpoutv, "     %5.3f %5.3f %5.3f,\n", out[3 * l],out[3 * l + 1], out[3 * l + 2]);
      fprintf (drvui->fpoutv, "     %5.3f %5.3f %5.3f]\n", out[3 * n],out[3 * n + 1], out[3 * n + 2]);
      fprintf (drvui->fpoutv, "  }\n   coordIndex [\n    ");
    } else {
      fprintf (drvui->fpoutv, " Material {\n   diffuseColor %s \n }\n", edgecolor);
      fprintf (drvui->fpoutv, "IndexedLineSet { coordIndex [\n    ");
    }
    l = 0;
    for (i = 0; i < draw_list - 1; i++) {
      j = poly_list[i];
      l++;
      for (k = 0; k < j; k++)
        fprintf (drvui->fpoutv, "%d,", point[l++]);
      if (i + j < draw_list - 2) {
        fprintf (drvui->fpoutv, "%d,-1,\n    ", point[l - j]);
      } else {
        if (Vrml2) {
          fprintf (drvui->fpoutv, "%d,-1]\n  ", point[l - j]);
        } else {
          fprintf (drvui->fpoutv, "%d,-1]\n  }\n }\n", point[l - j]);
        }
      }
      i = i + j;
    }
    if (Vrml2) {
      fprintf (drvui->fpoutv, "  color Color { color [%s]}\n colorIndex[0]\n colorPerVertex FALSE}\n",edgecolor);
      fprintf (drvui->fpoutv, "  }\n");
    }
  } else {
    if (!Vrml2)
      fprintf (drvui->fpoutv, " }\n");
  }
  free (out);        			 /* release the space */
  free (point);
  return;
}

/* ************************************************************** */
/* ************************************************************** */

void print_sym (void)

/* print symmetry information */
{
  int i, j, k;

  switch (drvui->sys) {
  case 1:
    fprintf (drvui->flout, "Triclinic - ");
    break;
  case 2:
    fprintf (drvui->flout, "Monoclinic - ");
    break;
  case 3:
    fprintf (drvui->flout, "Orthorhombic - ");
    break;
  case 4:
    fprintf (drvui->flout, "Tetragonal - ");
    break;
  case 5:
    fprintf (drvui->flout, "Hexagonal - ");
    break;
  case 6:
    fprintf (drvui->flout, "Cubic - ");
  }
  if (!drvui->acentric) {
    fprintf (drvui->flout, "Centric - ");
  } else {
    fprintf (drvui->flout, "Acentric - ");
  }
  switch (drvui->nbr) {
  case 1:
    fprintf (drvui->flout, "P lattice");
    break;
  case 2:
    fprintf (drvui->flout, "A-centered");
    break;
  case 3:
    fprintf (drvui->flout, "B-centered");
    break;
  case 4:
    fprintf (drvui->flout, "C-centered");
    break;
  case 5:
    fprintf (drvui->flout, "F-centered");
    break;
  case 6:
    fprintf (drvui->flout, "I-centered");
    break;
  case 7:
    fprintf (drvui->flout, "R-centered");
    break;
  default:
    fprintf (drvui->flout, "**** Unknown Lattice ***");
  }
  fprintf (drvui->flout, " with %d Symmetry Operators\n\n", drvui->ng);
  for (i = 1; i <= drvui->ng; ++i) {

    for (j = 1; j <= 3; ++j) {
      for (k = 1; k <= 3; ++k) {
        drvui->rss[i - 1][j - 1][k - 1] = (float) drvui->ss[i - 1][j - 1][k - 1];
        if (drvui->ss[i - 1][j - 1][k - 1] != 0) {
          if (drvui->ss[i - 1][j - 1][k - 1] > 0) {
            fprintf (drvui->flout, "+");
          } else {
            fprintf (drvui->flout, "-");
          }
          switch (k) {
          case 1:
            fprintf (drvui->flout, "x");
            break;
          case 2:
            fprintf (drvui->flout, "y");
            break;
          case 3:
            fprintf (drvui->flout, "z");
          }
        }
      }
      if (drvui->ts[i - 1][j - 1] != 0.0) {
        if (drvui->ts[i - 1][j - 1] == 0.5)
          fprintf (drvui->flout, "+1/2");
        if (drvui->ts[i - 1][j - 1] == 0.25)
          fprintf (drvui->flout, "+1/4");
        if (drvui->ts[i - 1][j - 1] == 0.75)
          fprintf (drvui->flout, "+3/4");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.33333333) < 0.00001)
          fprintf (drvui->flout, "+1/3");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.66666667) < 0.00001)
          fprintf (drvui->flout, "+2/3");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.16666667) < 0.00001)
          fprintf (drvui->flout, "+1/6");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.83333333) < 0.00001)
          fprintf (drvui->flout, "+5/6");
      }
      if (j != 3)
        fprintf (drvui->flout, ",");
    }
    if (i != drvui->ng) {
      if (i % 4 != 0) {
        fprintf (drvui->flout, " ; ");
      } else {
        fprintf (drvui->flout, "\n");
      }
    }
  }
  Conv_Sym_Mat ();
  fprintf (drvui->flout, "\n");
}        				 /* end of print_sym */
