
// module drawxtl2.cxx - part of DRAWxtl V5.1
// Coded using the FLTK 1.1.6 widget set
//
//     Larry W. Finger, Martin Kroeker and Brian Toby
//
//
// This module contains the following routines:
//
//  determinant - only for 3x3 matrices
//  generate_ellipsoids - builds the ellipsoid descriptions for the output lists
//  generate_gl_texts - adds text labels to openGL rendering
//  generate_planes - builds the plane descriptions for the output lists
//  generate_poly - builds the polyhedra descriptions for the output lists
//  generate_texts - adds text labels to the output lists (not openGL)
//
#include "drawxtl.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <FL/glut.h>

#include "draw_ext.h"

#include "DRAWxtl_proto.h"

/* ************************************************************** */
/* ************************************************************** */

float determinant (double rot[3][3])
{
  double temp;

  temp = rot[0][0] * (rot[1][1] * rot[2][2] - rot[1][2] * rot[2][1])
  - rot[1][0] * (rot[0][1] * rot[2][2] - rot[0][2] * rot[2][1])
  + rot[2][0] * (rot[0][1] * rot[1][2] - rot[0][2] * rot[1][1]);
  return (float) temp;
}

/* ************************************************************** */
/* ************************************************************** */

void generate_ellipsoids (void)

/* routine to generate ellipsoid objects and add then to edit list */
{
  float ev[3];                   /* scaled eigenvalue */
  double rot[3][3];              /* rotation matrix for ellipsoid */
  int i, j, k, l, m, n, o, oo;   /* loop counters */
  int Vo[3];
  int n_ellipsoids;
  float EV[3][3];                /* array for eigenvectors */
  float alpha1, alpha2, alpha3;     /* terms needed to calculate rotations */
  float d1, d2, d3, alpha, sinalp, radius, temp;
  float glr, glg, glb;
  GLdouble clipx[4] = { 0., 0., 0., 0. };
  GLdouble clipy[4] = { 0., 0., 0., 0. };
  GLdouble clipz[4] = { 0., 0., 0., 0. };

  int outside = 0;
  int ntest[8][3] =
  { {1, 1, 1}, {-1, 1, 1}, {1, -1, 1}, {1, 1, -1}, {-1, -1, 1}, {1, -1, -1},
    {-1, 1, -1}, {-1, -1, -1} };                  /* list of octant coordinates to check */
  double eye[3], vmin, dist;                      /* used to find octant for cutout */
  int best;           /* octant list pointer */
  int omit,l1;
  char string[40];
  float sinal1, sinal2, sinal3, cosal1, cosal2, cosal3;
  float rm[16];

  n_ellipsoids = 0;
  radius = 0.00015f *Scale;

  if (Ellipaxis_width != 0.)
    radius = Ellipaxis_width;
  fprintf (drvui->flout, "\nGenerated %6.1f percent probability ellipsoids for:\n\n",
     100. * Ellipsoid_Prob);

  if (El_Cutout == 1) {           // if cutouts, update G_Rot
    glPushMatrix();
    glLoadIdentity();
    crystal->calculate(rm);       // get rotation matrix
    glPopMatrix();
    for (m = 0; m < 3; m++) {     //  and copy to G_Rot
      G_Rot[0][m] = rm[m];
      G_Rot[1][m] = rm[m + 4];
      G_Rot[2][m] = rm[m + 8];
    }
  }
  for (i = 1; i < n_ellips; ++i) {           // loop through ellipsoids
    if (drvui->ell_type[i] < -1) continue;   // This one is non-positive definite
    if (drvui->ell_type[i] < 1000) continue; // ellipsoid OK but not displayed
    for (j = 0; j < 3; j++) {
      ev[j] = drvui->ellips_RMS[i][j] * Ellipsoid_Scale;
      for (k = 0; k < 3; k++)
        EV[j][k] = drvui->ellips_EV[i][j][k];
    }
    for (j = 0; j < natom; ++j) {      /* loop through atoms */
      if (drvui->atom_fn[j] != drvui->frame_no) continue;
      nvert = 0;                       /* clear the vertex list */
      if (((drvui->atom_n[j] >> 24) & 255) == i)/* this ellipsoid */
        find_all_in_box (j);           /* generate all locations for this atom */
      if (nvert > 0) {                 /* if some present */
        if (do_ellipsoids == 0) {      /* Biso sphere */
          float *radii;                /* allocate storage for radii of spheres */

          radii = (float *)malloc ((unsigned) (nvert * sizeof (float)));
          for (k = 0; k < nvert; ++k)  /* copy radii info to storage */
            radii[k] = drvui->ellips_RMS[i][0] * Ellipsoid_Scale;
          Output_Spheres (radii, i);
          free (radii);
        } else {
          for (k = 0; k < nvert; ++k) {     /* loop through the atom positions */
            if (clipflag != 0) {

              outside = 0;
              for (l1 = 0; l1 < 3; l1++)
                if (o_vert[3 * k + l1] < drvui->clip_lim[drvui->frame_no][l1] || 
                    o_vert[3 * k + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3])
                  outside = 1;
            }

            omit = 0;
            for (l1=0;l1<Omit->nomits;l1++){
              if (Omit->omit1[l1]==i*100000 && Omit->omit2[l1] == k) omit = 1;
            }  
            if ((!clipflag || !outside) && !omit) {
              o = vert_sym_no[k];
              fprintf (drvui->flout,"  %c%c%c%c %1d at %8.5f %8.5f %8.5f with sym op %d\n",
                 drvui->atom_l[j][0], drvui->atom_l[j][1], drvui->atom_l[j][2],drvui->atom_l[j][3], drvui->sv_atom_n[j]
                 , o_vert[3 * k], o_vert[3 * k + 1],o_vert[3 * k + 2], o);
              if (no_comment == 0)
                fprintf (drvui->fpoutv,"# Ellipsoid around %c%c%c%c %1d at %8.5f %8.5f %8.5f \n",
                    drvui->atom_l[j][0], drvui->atom_l[j][1],drvui->atom_l[j][2], drvui->atom_l[j][3],
                    drvui->sv_atom_n[j], o_vert[3 * k], o_vert[3 * k + 1], o_vert[3 * k + 2]);
              fprintf (drvui->fpoutp,"/* Ellipsoid around %c%c%c%c %1d at %8.5f %8.5f %8.5f  */\n",
                    drvui->atom_l[j][0], drvui->atom_l[j][1], drvui->atom_l[j][2],drvui->atom_l[j][3], drvui->sv_atom_n[j], o_vert[3 * k]
                    , o_vert[3 * k + 1], o_vert[3 * k + 2]);
	             for (l = 0; l < 3; ++l) {
		           for (m = 0; m < 3; ++m) {
		             rot[l][m] = 0.0;
		             for (n = 0; n < 3; ++n) {
		               rot[l][m] += EV[l][n] * drvui->rss[o][n][m];
                  }
                }           // for on m
              }             // for on l

// make rotation right-handed (det > 0)

              if (determinant (rot) < 0.0) {
                for (l = 0; l < 3; l++) {   // make rot right handed
                  for (m = 0; m < 3; m++) {
                    rot[l][m] *= -1.0;
                  }
                }
              }
              if (fabs ((temp = determinant (rot)) - 1.0f) > 1.0e-3)
                fprintf (drvui->flout, "Determinant of rot = %8.5f\n", temp);

// calculate ellipsoid rotation angles for POV and openGL

              alpha2 = (float) asin (rot[0][2]);
              cosal2 = (float) cos (alpha2);
              if (fabs (cosal2) > 0.00001) {
                alpha1 = (float) atan2 (-rot[1][2] / cosal2, rot[2][2] / cosal2);
                alpha3 = (float) atan2 (-rot[0][1] / cosal2, rot[0][0] / cosal2);
              } else {
                alpha3 = (float) atan2 (rot[1][0], -rot[2][0]);
                alpha1 = 0.0f;
              }

// recalculate rot using alpha1,2,3 from above
              
              sinal2 = (float) sin(alpha2);
              sinal1 = (float) sin(alpha1);
              cosal1 = (float) cos(alpha1);
              sinal3 = (float) sin(alpha3);
              cosal3 = (float) cos(alpha3);
              rot[0][0] =  cosal2 * cosal3;
              rot[1][0] =  -cosal2 * sinal3;
              rot[2][0] = sinal2;
              rot[0][1] = cosal1 * sinal3 + sinal1 * sinal2 * cosal3;
              rot[1][1] = cosal1 * cosal3 - sinal1 * sinal2 * sinal3;
              rot[2][1] = -sinal1 * cosal2;
              rot[0][2] = sinal1 * sinal3 - cosal1 * sinal2 * cosal3;
              rot[1][2] = sinal1 * cosal3 + cosal1 * sinal2 * sinal3;
              rot[2][2] = cosal1 * cosal2;

              if (El_Cutout == 1) {           // select octant for cutout
                double fullrot[3][3];
                vmin =  -1000000.;            // initialize to impossible value
                best = 0;
                for (m = 0; m < 3; m++) {     // calculate combined rotation matrix
                    for (n = 0; n < 3; n++) {
                        fullrot[m][n] = 0.0;
                        for (oo = 0; oo < 3; oo++) {
                            fullrot[m][n] += rot[oo][m] * G_Rot[oo][n];
                        }
                    }
                }
                if (fabs ((temp = determinant (fullrot)) - 1.0f) > 1.0e-3)
                    fprintf (drvui->flout, "Determinant of fullrot = %8.5f\n", temp);

                eye[0] = fullrot[0][2];       // get the rotated position of the viewer (z axis)
                eye[1] = fullrot[1][2];
                eye[2] = fullrot[2][2];

                for (n = 0; n < 8; n++) {
                  for (m = 0; m < 3; m++)     // take next triple from list
                    Vo[m] = ntest[n][m];

                  dist = eye[0] * Vo[0] + eye[1] * Vo[1] + eye[2] * Vo[2];
                  if (dist > vmin) {          // maximum is the one we want
                    vmin = dist;
                    best = n;
                  }
                }

                for (m = 0; m < 3; m++)       // set the octant to display
                  Vo[m] = ntest[best][m];

              }           // ... if cutout

// output this ellipsoid to the POV file

              fprintf (drvui->fpoutp," object{\n");
              fprintf (drvui->fpoutp,"   union{\n");
              fprintf (drvui->fpoutp,"     object{\n");
              fprintf (drvui->fpoutp,"      cylinder{ < 0,0,-0.5 > , < 0,0,0.5 >, 1. open\n");  // first principal ellipse
              fprintf (drvui->fpoutp,"       texture{pigment{color %s}}\n",Ellipaxis_color);
              fprintf (drvui->fpoutp,"       scale<%8.5f,%8.5f,%8.5f>\n      }\n     }\n",
                     ev[0], ev[1], radius);
              fprintf (drvui->fpoutp,"     object{\n      cylinder{ < 0,-0.5,0 > , < 0,0.5,0 >, 1. open\n");  // second principal ellipse
              fprintf (drvui->fpoutp,"       texture{pigment{color %s}}\n",Ellipaxis_color);
              fprintf (drvui->fpoutp,"       scale<%8.5f,%8.5f,%8.5f>\n      }\n     }\n",
                     ev[0], radius, ev[2]);
              fprintf (drvui->fpoutp,"     object{\n      cylinder{ < -0.5,0,0 > , < 0.5,0,0 >, 1. open\n");  // third principal ellipse
              fprintf (drvui->fpoutp,"       texture{pigment{color %s}}\n",Ellipaxis_color);
              fprintf (drvui->fpoutp,"       scale<%8.5f,%8.5f,%8.5f>\n      }\n     }\n",
                     radius, ev[1], ev[2]);
              fprintf (drvui->fpoutp,"     object{\n");
              if (El_Cutout)
                fprintf (drvui->fpoutp,"      difference {\n");
              fprintf (drvui->fpoutp,"       sphere{< 0,0,0 > 1}\n");
              if (El_Cutout) {
                fprintf (drvui->fpoutp,"       box{<0,0,0>,<%d,%d,%d>}\n",
                    Vo[0], Vo[1], Vo[2]);
              }
              fprintf (drvui->fpoutp,"       texture{pigment{color %s  }}\n",
                       drvui->ellips_col_p[i]);
              fprintf (drvui->fpoutp,"       finish{phong %5.2f phong_size %5.2f reflection 0.1}\n",
                       drvui->Phong_Value, drvui->Phong_Size);
              fprintf (drvui->fpoutp,"       scale<%8.5f,%8.5f,%8.5f>\n",
                       ev[0], ev[1], ev[2]);
              if (El_Cutout) {
                fprintf (drvui->fpoutp,"      }\n");
                fprintf (drvui->fpoutp,"     }\n");
                fprintf (drvui->fpoutp,"     object{\n");
                fprintf (drvui->fpoutp,"      cylinder{ < 0,0,-0.5 > , < 0,0,0.5 >, 0.99\n");  // first principal ellipse
                fprintf (drvui->fpoutp,"       texture{pigment{color %s}}\n",drvui->Cutout_color);
                fprintf (drvui->fpoutp,"       scale<%8.5f,%8.5f,%8.5f>\n      }\n     }\n",
                     ev[0], ev[1], radius);
                fprintf (drvui->fpoutp,"     object{\n      cylinder{ < 0,-0.5,0 > , < 0,0.5,0 >, 0.99\n");  // second principal ellipse
                fprintf (drvui->fpoutp,"       texture{pigment{color %s}}\n",drvui->Cutout_color);
                fprintf (drvui->fpoutp,"       scale<%8.5f,%8.5f,%8.5f>\n      }\n     }\n",
                     ev[0], radius, ev[2]);
                fprintf (drvui->fpoutp,"     object{\n      cylinder{ < -0.5,0,0 > , < 0.5,0,0 >, 0.99\n");  // third principal ellipse
                fprintf (drvui->fpoutp,"       texture{pigment{color %s}}\n",drvui->Cutout_color);
                fprintf (drvui->fpoutp,"       scale<%8.5f,%8.5f,%8.5f>\n      }\n     }\n",
                      radius, ev[1], ev[2]);
              }
              fprintf (drvui->fpoutp,"      }\n");
              if (!El_Cutout) fprintf (drvui->fpoutp,"     }\n");
              fprintf (drvui->fpoutp,"   rotate <%8.5f,0,0>\n", -alpha1 * RAD);
              fprintf (drvui->fpoutp,"   rotate <0,%8.5f,0>\n", -alpha2 * RAD);
              fprintf (drvui->fpoutp,"   rotate <0,0,%8.5f>\n", -alpha3 * RAD);
              fprintf (drvui->fpoutp,"   translate <%8.5f, %8.5f, %8.5f>\n }\n",
                 s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);

// Output the ellipsoid to the VRML file

              d1 = (float) (0.5 * (rot[2][1] - rot[1][2]));
              d2 = (float) (0.5 * (rot[0][2] - rot[2][0]));
              d3 = (float) (0.5 * (rot[1][0] - rot[0][1]));

// d's contain direction cosines * sin(alpha)

              alpha = (float)(0.5 * (rot[0][0] + rot[1][1] +
                    rot[2][2] - 1.0f));
              alpha = (float) acos (alpha);
              sinalp = (float) sin (alpha);
              if (fabs (sinalp) > 0.0001) {
                d1 = d1 / sinalp;
                d2 = d2 / sinalp;
                d3 = d3 / sinalp;

// Make sure alpha is in correct quadrant

                if (fabs (d1) > 0.0001) {
                  temp = (float) (1.0 - cos (alpha)) * d2 * d3 -
                        d1 * (float) sin (alpha);
                  if (fabs (rot[1][2] - temp) > 0.0001)
                    alpha = -alpha;
                } else {
                  if (fabs (d2) > 0.0001) {
                    temp = (float) (1.0 - cos (alpha)) * d1 *
                          d3 - d2 * (float) sin (alpha);
                    if (fabs (rot[2][0] - temp) > 0.0001)alpha = -alpha;
                  } else {
                    if (fabs (d3) > 0.0001) {
                      temp = (float) (1.0 - cos (alpha)) *
                           d1 * d2 - d3 * (float) sin (alpha);
                      if (fabs (rot[0][1] - temp) > 0.0001)
                        alpha = -alpha;
                    }
                  }
                }
              } else {
                d1 = (float) sqrt (fabs(rot[0][0]));
                d2 = (float) sqrt (fabs(rot[1][1]));
                d3 = (float) sqrt (fabs(rot[2][2]));
              }      //sinalp != 0

              temp = (float) sqrt (d1 * d1 + d2 * d2 + d3 * d3);
              d1 = d1 / temp;
              d2 = d2 / temp;
              d3 = d3 / temp;

              if (!Vrml2)
                fprintf (drvui->fpoutv," Separator {\n");
              fprintf (drvui->fpoutv,"  Transform {\n");
              fprintf (drvui->fpoutv,
                "   rotation %6.3f %6.3f %6.3f %7.3f\n",
                d1, d2, d3, -alpha);
              fprintf (drvui->fpoutv,
                "   translation %8.5f %8.5f %8.5f\n",
                s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
              if (strlen (drvui->ellips_col_v[i]) == 0)
                strcpy (drvui->ellips_col_v[i], "1 0 0");
              if (Vrml2) {
                fprintf (drvui->fpoutv,"   scale %8.5f %8.5f %8.5f\n", ev[0], ev[1], ev[2]);
                fprintf (drvui->fpoutv, "  children [ Shape {");
                fprintf (drvui->fpoutv, " appearance Appearance {\n");
                fprintf (drvui->fpoutv,"  material Material {diffuseColor  %s}\n",
                    drvui->ellips_col_v[i]);
                fprintf (drvui->fpoutv, "  }\n");
                fprintf (drvui->fpoutv, "  geometry Sphere {");
                fprintf (drvui->fpoutv, "   radius     1.}}]}\n");
              } else {
                fprintf (drvui->fpoutv,"   scaleFactor %8.5f %8.5f %8.5f\n",ev[0], ev[1], ev[2]);
                fprintf (drvui->fpoutv, "  }\n");
                fprintf (drvui->fpoutv, "  Material {\n");
                fprintf (drvui->fpoutv, "   shininess 0.3\n");
                fprintf (drvui->fpoutv, "   diffuseColor  %s\n", drvui->ellips_col_v[i]);
                fprintf (drvui->fpoutv, "  }\n");
                fprintf (drvui->fpoutv, "  Sphere {\n");
                fprintf (drvui->fpoutv, "   radius     1\n  }\n");
                fprintf (drvui->fpoutv, " }\n");
              }
// output ellipsoid to screen with openGL
              glPushMatrix ();
              glLoadName (i*100000);
              glPushName (k);
              (void)sscanf (drvui->ellips_col_v[i], "%f %f %f", &glr, &glg, &glb);
              glColor3f (glr, glg, glb);
              glTranslatef (s_vert[3 * k],s_vert[3 * k + 1], s_vert[3 * k + 2]);
              glRotatef (-alpha3 * (float) RAD, 0.0f, 0.0f, 1.0f);
              glRotatef (-alpha2 * (float) RAD, 0.0f, 1.0f, 0.0f);
              glRotatef (-alpha1 * (float) RAD, 1.0f, 0.0f, 0.0f);
              glScalef (ev[0], ev[1], ev[2]);
              if (El_Cutout) {
                clipx[0] = -(GLdouble)Vo[0];
                clipy[1] = -(GLdouble)Vo[1];
                clipz[2] = -(GLdouble)Vo[2];
                glClipPlane (GL_CLIP_PLANE0, clipx);
                glClipPlane (GL_CLIP_PLANE1, clipy);
                glClipPlane (GL_CLIP_PLANE2, clipz);
                glEnable (GL_CLIP_PLANE0);
                glutSolidSphere (1., 10, 10);
                glDisable (GL_CLIP_PLANE0);
                glEnable (GL_CLIP_PLANE1);
                glutSolidSphere (1., 10, 10);
                glDisable (GL_CLIP_PLANE1);
                glEnable (GL_CLIP_PLANE2);
                glutSolidSphere (1., 10, 10);
                glDisable (GL_CLIP_PLANE2);
                strncpy(string,Ellipaxis_color,39);
                trim_string(string,40);
                check_color(string);
                (void)sscanf(string,"%f %f %f",&glr,&glg,&glb);
                glColor3f (glr, glg, glb);
                glDisable (GL_LIGHTING);
                glBegin (GL_LINES);
                glVertex3f (0.0f, 0.0f, 0.0f);
                glVertex3f ((float) Vo[0], 0.0f, 0.0f);
                glVertex3f (0.0f, 0.0f, 0.0f);
                glVertex3f (0.0f, (float) Vo[1], 0.0f);
                glVertex3f (0.0f, 0.0f, 0.0f);
                glVertex3f (0.0f, 0.0f, (float) Vo[2]);
                glEnd ();
                glEnable (GL_LIGHTING);
              } else {
                glutSolidSphere (1., 10, 10);
              }                // if El_Cutout
              glPopName();
              glPopMatrix ();
              n_ellipsoids++;
            }           // if clipflag == 0
          }             // for on k
        }               // if ellipsoid 
      }                 // if nvert > 0
    }		            // for all atoms j of this type 
  }                     // for on i - loop through ellipsoids
  fprintf (drvui->fcns,"%4d ellipsoids\n", n_ellipsoids);
  fprintf (drvui->flout, "%4d ellipsoids generated\n", n_ellipsoids);
}             // end of generate_ellipsoids

/* ************************************************************** */
/* ************************************************************** */

void generate_planes (void)

/* procedure to generate corners of planes */
{
  int i, j, k;             /* loop counters, etc */
  int nvert_start;         /* starting number for cations */
  int planeno;             /* number of current polyhedron */
  int Plane_Count;         /* number of planes output */
  int ii, jj, ll, l;
  int *no, *ns;
  float xm, ym, zm, vx, vy, vz, bm, vxs, vys, vzs, bms, phi, phi0, cosphi,
  radius;
  int outside;
  float glr, glg, glb;
  static int warn = 0;

  radius = rad_edge * (float) (0.01 * Scale);
  if (drvui->nplane == 1)
    return;           /* exit NOW if no work */
  Plane_Count = 0;
  for (i = 0; i < 3; ++i)
    boxlim[i] += 3.0f;          /* expand box for cation generation */
  nvert = 0;                    /* make list empty */
  for (i = 0; i < natom; ++i) { /* loop through atoms */
      if (drvui->atom_fn[i] != drvui->frame_no) continue;
      find_all_in_box (i);      /* generate all potential cations */
  }
  for (i = 0; i < 3; ++i)
    boxlim[i] -= 3.0f;          /* shrink box to original size */
  if (nvert == 0)
    return;                     /* if no anions, no polyhedra */
  nvert_start = nvert;          /* save start of cation list */
  for (i = 0; i < natom; ++i) { /* start loop for cations */
    if (drvui->atom_fn[i] != drvui->frame_no) continue;
    nvert = nvert_start;        /* get rid of previous cations */
    planeno = (drvui->atom_n[i] >> 16) & 255;     /* save plane number */
    if (planeno > 0 && drvui->plane_fn[planeno] == drvui->frame_no) { /* greater than zero for polyhedral centers */
      find_all_in_box (i);      /* add all polyhedral centers in box */
      if (nvert_start < nvert) {/* check for some found */
        draw_list = 0;          /* initialize the draw list */
        for (j = nvert_start; j < nvert; ++j) {   /* loop through cations */
          outside = 0;
          if (clipflag != 0) {
            for (ii = 0; ii < 3; ii++) {
              if (o_vert[3 * j + ii] < drvui->clip_lim[drvui->frame_no][ii] - 0.01 || 
                  o_vert[3 * j + ii] > drvui->clip_lim[drvui->frame_no][ii + 3] + 0.01)
                outside = 1;
            }
          }
          if (domolcomp && packflag) {
            for (ii = 0; ii < 3; ii++) {
              if (o_vert[3 * j + ii] < drvui->cryst_lim[drvui->frame_no][ii] - 0.01 || 
                  o_vert[3 * j + ii] > drvui->cryst_lim[drvui->frame_no][ii + 3] + 0.01)
                outside = 1;
            }
          }
          numb_list = 0;            /* list clear */
          for (k = 0; k < nvert_start; ++k) {   /* loop through anions */
            if (dist (j, k) <= 0.) continue;
            if (dist (j, k) <= drvui->plane_size[planeno] + 0.0001 && !outside)  /* add vertex k to poly list */
              vertex_list[numb_list++] = k;
          }

          if (numb_list > 2) {       /* only do if at least a triangle */
            fprintf (drvui->flout,"Polygon with %d vertices around %c%c%c%c at %8.5f %8.5f %8.5f\n",
                   numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1],drvui->atom_l[i][2], drvui->atom_l[i][3]
                   , s_vert[3 * j], s_vert[3 * j + 1], s_vert[3 * j + 2]);
            if (no_comment == 0)
              fprintf (drvui->fpoutv,"# Polygon (%d) around %c%c%c%c at %8.5f %8.5f %8.5f \n",
                   numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1],drvui->atom_l[i][2], drvui->atom_l[i][3]
                   , o_vert[3 * j], o_vert[3 * j + 1], o_vert[3 * j + 2]);
            if (Vrml2) {
              fprintf (drvui->fpoutv, " Shape {");
              fprintf (drvui->fpoutv, "appearance Appearance {\n");
              fprintf (drvui->fpoutv,"   material Material {diffuseColor %s} \n \n",
                  drvui->plane_col_v[planeno]);
            } else {
              fprintf (drvui->fpoutv, " Separator {\n        ");
              fprintf (drvui->fpoutv, "  Material {\n");
              fprintf (drvui->fpoutv, "   diffuseColor %s \n  }\n",drvui->plane_col_v[planeno]);
            }
            Plane_Count++;       /* increment number output */
            poly_list[0] = numb_list;
/* Make vertices consecutive - find center of polygon */
            xm = ym = zm = 0.0f;
            if (!(no = (int *)malloc ((unsigned)((4 * numb_list) * sizeof (int))))) {
              Error_Box("Unable to get no allocation");
              return;
            }
            if (!(ns = (int *)malloc ((unsigned)((4 * numb_list) * sizeof (int))))) {
              Error_Box("Unable to get ns allocation");
              free (no);
              return;
            }
            for (ii = 0; ii < numb_list; ii++) {
              no[ii] = vertex_list[ii];
              xm = xm + s_vert[3 * no[ii]];
              ym = ym + s_vert[3 * no[ii] + 1];
              zm = zm + s_vert[3 * no[ii] + 2];
            }
            xm = xm / (float) numb_list;
            ym = ym / (float) numb_list;
            zm = zm / (float) numb_list;

/* sort vertices in consecutive order based on minimal angles between vectors */

	    for (ii = 0; ii < numb_list; ii++)
              ns[ii] = 0;
            ns[0] = no[0];
            for (ii = 0; ii < numb_list - 1; ii++) {
              phi0 = 1000.0f;
              vxs = s_vert[3 * ns[ii]] - xm;
              vys = s_vert[3 * ns[ii] + 1] - ym;
              vzs = s_vert[3 * ns[ii] + 2] - zm;
              bms = (float) sqrt (vxs * vxs + vys * vys + vzs * vzs);
              for (jj = 0; jj < numb_list; jj++) {
                for (ll = 0; ll < ii; ll++)
                  if (ns[ll] == no[jj])
                    jj++;
                if (ns[ii] != no[jj] && jj < numb_list) {
                  vx = s_vert[3 * no[jj]] - xm;
                  vy = s_vert[3 * no[jj] + 1] - ym;
                  vz = s_vert[3 * no[jj] + 2] - zm;
                  bm = (float) sqrt (vx * vx + vy * vy + vz * vz);
                  cosphi = (vxs * vx + vys * vy + vzs * vz) / (bms * bm);
                  if (cosphi < -1.)
                    cosphi = -1.0f;
                  if (cosphi > 1.)
                    cosphi = 1.0f;
                  phi = (float) acos (cosphi) * (float) RAD;
                  if (phi < phi0) {
                    phi0 = phi;
                    ns[ii + 1] = no[jj];
                  }
                }
              }
            }
            draw_list = 1;
            poly_list[0] = numb_list;
            for (l = 0; l < numb_list; ++l)
              poly_list[draw_list++] = ns[l] + 1;
            poly_list[draw_list++] = 0;     /* terminate polygon list */
            if (draw_list > 4 * NvertM) {
              poly_list = (int *)malloc ((unsigned) (4 * (nvert+2) * sizeof (int)));
              if (!warn) {
                fprintf (drvui->flout,"Overrun of poly_list.\n");
                fprintf (drvui->flout, "Overrun of poly_list.\n");
                Error_Box("Overrun of poly_list. Please send 'str' file\n"
                   "to Larry.Finger@@lwfinger.net.");
                warn = 1;
              }
          }
            plot_vrml_poly (planeno);
            fprintf (drvui->fpoutp," /* Polygon (%d) around %c%c%c%c at %8.5f %8.5f %8.5f */ \n",
                  numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1],drvui->atom_l[i][2], drvui->atom_l[i][3]
                  , o_vert[3 * j], o_vert[3 * j + 1], o_vert[3 * j + 2]);
            glPushMatrix ();
            (void)sscanf (drvui->plane_col_v[planeno], "%f %f %f", &glr, &glb, &glg);
            glColor3f (glr, glb, glg);
            glBegin (GL_TRIANGLES);
            for (l = 0; l < numb_list - 2; ++l) {
              fprintf (drvui->fpoutp," triangle {< %8.5f, %8.5f, %8.5f>,\n",
                   s_vert[3 * ns[0]], s_vert[3 * ns[0] + 1],s_vert[3 * ns[0] + 2]);
              fprintf (drvui->fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
                   s_vert[3 * ns[l + 1]],s_vert[3 * ns[l + 1] + 1],
                   s_vert[3 * ns[l + 1] + 2],s_vert[3 * ns[l + 2]],
                   s_vert[3 * ns[l + 2] + 1], s_vert[3 * ns[l + 2] + 2]);
              fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n",drvui->plane_col_p[planeno]);
              fprintf(drvui->fpoutp," }\n");
              glVertex3f (s_vert[3 * ns[0]],s_vert[3 * ns[0] + 1], s_vert[3 * ns[0] + 2]);
              glVertex3f (s_vert[3 * ns[l + 1]],s_vert[3 * ns[l + 1] + 1],s_vert[3 * ns[l + 1] + 2]);
              glVertex3f (s_vert[3 * ns[l + 2]],s_vert[3 * ns[l + 2] + 1],s_vert[3 * ns[l + 2] + 2]);
            }
            glEnd ();
            glPopMatrix ();
            if (edges) {
              float df[3], at[3];
              int m;
              for (l = 0; l < numb_list - 1; ++l) {
                fprintf (drvui->fpoutp,
                     " cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                     s_vert[3 * ns[l]], s_vert[3 * ns[l] + 1], s_vert[3 * ns[l] + 2],
                     s_vert[3 * ns[l + 1]], s_vert[3 * ns[l + 1] + 1],
                     s_vert[3 * ns[l + 1] + 2], radius);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", col_edge_p);
                fprintf(drvui->fpoutp," }\n");
                for (m=0; m<3; m++) {
                  at[m] = s_vert[3 * ns[l] + m];
                  df[m] = s_vert[3 * ns[l + 1] + m] - at[m];
                }
                push_cylinder(df, at, radius, col_edge_v);   // do the openGL edge
              }
              fprintf (drvui->fpoutp,
                   " cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                   s_vert[3 * ns[0]], s_vert[3 * ns[0] + 1],
                   s_vert[3 * ns[0] + 2], s_vert[3 * ns[numb_list - 1]],
                   s_vert[3 * ns[numb_list - 1] + 1],
                   s_vert[3 * ns[numb_list - 1] + 2], radius);
              fprintf (drvui->fpoutp," texture{pigment{color %s  }}\n", col_edge_p);
              fprintf(drvui->fpoutp," }\n");
              for (m=0; m<3; m++) {
                at[m] = s_vert[3 * ns[0] + m];
                df[m] = s_vert[3 * ns[numb_list - 1] + m] - at[m];
              }
              push_cylinder(df, at, radius, col_edge_v);   // do the openGL edge
            }
            draw_list = 0;
            free (no);
            free (ns);
          }           /* numb_list > 2 */
        }           /* loop on j through cations */
      }             /* nvert_start < nvert */
    }
  }             /* end of loop through cations */
  fprintf (drvui->fcns,"%4d planes.\n", Plane_Count);
  fprintf (drvui->flout, "Generated %4d planes.\n", Plane_Count);
}

/* ************************************************************** */
/* ************************************************************** */

void generate_gl_texts (void)
/* procedure to render all textual labels in the OpenGL view - this has
   to be done outside the display list so that the labels always face
   toward the viewer during rotations */
{
  int i, n;
  float vert[3];
  char *p;
  float scale;

 // if (nvert == 0) return; // happens when the mapread progress bar forces a 
                          // (re)draw before the structure has been read

  display_cursor_text();
  if (drvui->nlabel == 1)
    return;

  if (Display_axes) {
      nvert = 0;
      generate_triple();
      draw_GL_triple();
  }

  for (n = 1; n < drvui->nlabel; n++) {
    if (drvui->label_fn[n] != drvui->frame_no) continue;
    nvert = 0;           /* initialize vertex list */
    for (i = 0; i < 3; i++) {
        vert[i] = drvui->label_x[n][i];
        if (Display_axes && ((n==drvui->triple[0]) || (n==drvui->triple[1])
                 || (n == drvui->triple[2]) || (n==drvui->triple[3])))
            vert[i] += origin[i];
    }
    add_vert_nc (vert);

    if (Display_axes && ((n==drvui->triple[0]) || (n==drvui->triple[1])
                || (n == drvui->triple[2]) || (n==drvui->triple[3])))
        for (i=0; i<3; i++)
            s_vert[i] += offset[i];

    glPushMatrix();
    glLoadName(n);
    glTranslatef(s_vert[0],s_vert[1],s_vert[2]);
    scale = 0.005f;
    if (!strcmp(drvui->label_label[n],"triple_vect"))
        scale *= 0.00001f;                  // make "triple_vect" really small
    glRotatef (-(float)atof(drvui->X_Rot->value()), 1.0f, 0.0f, 0.0f);
    glRotatef (-(float)atof(drvui->Y_Rot->value()), 0.0f, 1.0f, 0.0f);
    glRotatef (-(float)atof(drvui->Z_Rot->value()), 0.0f, 0.0f, 1.0f);
    glColor3f (0.0f, 0.0f, 0.0f);
    glDisable(GL_LIGHTING);
    glLineWidth(2.0);
    glScalef(drvui->label_scale * scale, drvui->label_scale * scale,
        drvui->label_scale * scale);
    for (p=drvui->label_label[n];*p;p++)
      glutStrokeCharacter(GLUT_STROKE_ROMAN,*p);
    glLineWidth(1.0);
    glEnable(GL_LIGHTING);
    glPopMatrix();
  }             /* end of loop on n   */
}

/* ************************************************************** */
/* ************************************************************** */

void generate_poly (void)

/* procedure to generate vertices of polyhedra */
{
  int i, j, k;             /* loop counters, etc */
  int nvert_start;         /* starting number for cations */
  int polyno;              /* number of current polyhedron */
  int Poly_Count;          /* Number of polyhedra output */
  int Max_Vertices;        /* Maximum number of vertices */
  float radius;
  int ii, outside;
  int type_check;          /* if atom type checking needs to be done */
  int omit,l1;
  static int warn = 0;

  radius = rad_edge * 0.01f *Scale;

  if (drvui->npoly == 1)
    return;                /* exit NOW if no work */
  Poly_Count = 0;
  for (i = 0; i < 3; ++i)
    boxlim[i] += 3.0f;     /* expand box for anion generation */
  nvert = 0;               /* make list empty */
  for (i = 0; i < natom; ++i) {       /* loop through atoms */
      if (drvui->atom_fn[i] != drvui->frame_no) continue;
      find_all_in_box (i); /* generate all potential anions */
  }
  for (i = 0; i < 3; ++i)
    boxlim[i] -= 3.0f;     /*shrink box to original size */
  if (nvert == 0)
    return;                /* if no anions, no polyhedra */
  nvert_start = nvert;     /* save start of cation list */
  for (i = 0; i < natom; ++i) {       /* start loop for cations */
    if (drvui->atom_fn[i] != drvui->frame_no) continue;
    nvert = nvert_start;   /* get rid of previous cations */
    type_check = 0;
    polyno = (drvui->atom_n[i] >> 8) & 255;  // save polyhedron number
    if (polyno > 0 && drvui->poly_fn[polyno] == drvui->frame_no) {  // greater than zero for polyhedral centers
      if (drvui->poly_t[polyno][0] != '\0')
        type_check = 1;
      find_all_in_box (i);       /* add all polyhedral centers in box */
      if (nvert_start < nvert) {     /* check for some found */
        draw_list = 0;         /* initialize the draw list */
        Max_Vertices = 0;
        for (j = nvert_start; j < nvert; ++j) {   /* first loop through cations */
          numb_list = 0;       /* list clear */
          outside = 0;
          if (clipflag != 0) {
            for (ii = 0; ii < 3; ii++) {
              if (o_vert[3 * j + ii] < drvui->clip_lim[drvui->frame_no][ii] - 0.01 || 
                  o_vert[3 * j + ii] > drvui->clip_lim[drvui->frame_no][ii + 3] + 0.01)
                outside = 1;
            }
          }
          if (outside == 1)
            continue;
          for (k = 0; k < nvert_start; ++k) {   /* loop through anions */
            float d;
            int kk;

            kk = drvui->orig_atom_no[k];
            if ((d = dist (j, k)) <= drvui->poly_size[polyno]) {  /* add vertex k to poly list */
              if ((!type_check) || (check_atom_name(drvui->atom_l[kk], drvui->poly_t[polyno]))) {
                if (d >= drvui->poly_min[polyno]) {
                  numb_list++;
                }
              }
            }
          }
          if (numb_list > Max_Vertices)
            Max_Vertices = numb_list;
        }                      /* end of first pass through list */
        draw_list = 0;         /* initialize the draw list */
        for (j = nvert_start; j < nvert; ++j) {   /* second loop through cations */
          numb_list = 0;       /* list clear */
          for (k = 0; k < nvert_start; ++k) {   /* loop through anions */
            float d;
            int kk;

            kk = drvui->orig_atom_no[k];
            outside = 0;
            if (domolcomp && packflag) {
              for (ii = 0; ii < 3; ii++) {
                if (o_vert[3 * k + ii] < drvui->cryst_lim[drvui->frame_no][ii] - 0.01 ||
                    o_vert[3 * k + ii] > drvui->cryst_lim[drvui->frame_no][ii + 3] + 0.01)
                  outside = 1;
              }
            }
            if (clipflag != 0) {
              for (ii = 0; ii < 3; ii++) {
                if (o_vert[3 * k + ii] < drvui->clip_lim[drvui->frame_no][ii] - 0.01 || 
                    o_vert[3 * k + ii] > drvui->clip_lim[drvui->frame_no][ii + 3] + 0.01)
                  outside = 1;
              }
            }
            if ((d = dist (j, k)) <= drvui->poly_size[polyno] + 0.000001 && !outside) {  /* add vertex k to poly list */
              if (d >= drvui->poly_min[polyno]) {
                if ((type_check == 0) || check_atom_name(drvui->atom_l[kk], drvui->poly_t[polyno]))
                  vertex_list[numb_list++] = k;
                if (numb_list > 4 * NvertM) {
                  vertex_list = (int *)malloc ((unsigned) (4 * (nvert+2) * sizeof (int)));
                  if (!warn) {
                  fprintf (drvui->flout,"Overrun of vertex_list.\n");
                  fprintf (drvui->flout, "Overrun of vertex_list.\n");
                  Error_Box("Overrun of vertex_list.\nPlease send 'str' file "
                      "to Larry.Finger@@lwfinger.net.");
                  warn = 1;
                  }
                } 
              }
            }
          }

          omit = 0;
          for (l1=0;l1<Omit->nomits;l1++){
            if (Omit->omit1[l1]==(i+1)*100000+(i+1) && Omit->omit2[l1] == j) omit = 1;
          }  

          if (numb_list > 3 && numb_list == Max_Vertices && omit==0) {  /* only do if polyhedron complete */
            fprintf (drvui->flout,"Polyhedron with %d vertices around %c%c%c%c at %8.5f %8.5f %8.5f \n",
                  numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1],drvui->atom_l[i][2], drvui->atom_l[i][3]
                  , s_vert[3 * j], s_vert[3 * j + 1], s_vert[3 * j + 2]);
            fprintf (drvui->fpoutp," /* Polyhedron (%d) around %c%c%c%c at %8.5f %8.5f %8.5f */ \n",
                  numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1],drvui->atom_l[i][2], drvui->atom_l[i][3]
                  , o_vert[3 * j], o_vert[3 * j + 1], o_vert[3 * j + 2]);
            if (no_comment == 0)
              fprintf (drvui->fpoutv,"# Polyhedron (%d) around %c%c%c%c at %8.5f %8.5f %8.5f \n",
                  numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1],drvui->atom_l[i][2], drvui->atom_l[i][3]
                  , o_vert[3 * j], o_vert[3 * j + 1], o_vert[3 * j + 2]);
            if (Vrml2) {
              fprintf (drvui->fpoutv, " Shape {");
              fprintf (drvui->fpoutv,"appearance Appearance {\n  material Material { diffuseColor  %s} \n \n",
                   drvui->poly_col_v[polyno]);
            } else {
              fprintf (drvui->fpoutv, " Separator {\n    ");
              fprintf (drvui->fpoutv,"  Material {\n   diffuseColor  %s \n  }\n",drvui->poly_col_v[polyno]);
            }
            if (drvui->p_rad_edge[polyno] != 0.)
              radius = drvui->p_rad_edge[polyno] * (float) 0.01 *Scale;

            build_poly_list (numb_list, polyno,i+1,j); /* select polygons */
            poly_list[draw_list++] = 0;     /* terminate polygon list */
            plot_vrml_poly (polyno);     /* pov and openGL part already done in build_poly_list */
            draw_list = 0;
            Poly_Count++;
          }
        }
      }             /* nvert_start < nvert */
    }
  }             /* end of loop through cations */
  fprintf (drvui->fcns,"%4d polyhedra.\n", Poly_Count);
  fprintf (drvui->flout, "Generated %4d polyhedra.\n", Poly_Count);
}

/* ************************************************************** */
/* ************************************************************** */

void generate_texts (void)

/* procedure to create POV and VRML command sequences for rendering of
   user-defined labels (labeltext command) - the GL rendering is done
   in generate_gl_texts */
{
  int i, n;
  int Label_Count = 0;
  float vert[3];

  if (drvui->nlabel == 1)
    return;

  for (n = 1; n < drvui->nlabel; n++) {

    float size;

    if (drvui->label_fn[n] != drvui->frame_no) continue;
    if (drvui->triple[0] == n) continue;              // skip "vector_trip"

    nvert = 0;           /* initialize vertex list */
    for (i = 0; i < 3; i++) {
      vert[i] = drvui->label_x[n][i];
      if (Display_axes && ((n==drvui->triple[0]) || (n==drvui->triple[1])
                 || (n == drvui->triple[2]) || (n==drvui->triple[3])))
          vert[i] += origin[i];
    }
    add_vert_nc (vert);

    if (Display_axes && ((n==drvui->triple[0]) || (n==drvui->triple[1])
                 || (n == drvui->triple[2]) || (n==drvui->triple[3])))
        for (i=0; i<3; i++)
            s_vert[i] += offset[i];
    Label_Count++;
    fprintf (drvui->flout, "Label[%d] %g %g %g\n", n, vert[0], vert[1], vert[2]);
    fprintf (drvui->fpoutp, "/* Labels */\n");
    fprintf (drvui->fpoutv, "# Labels\n");
    size = drvui->label_scale * 8.0f * Text_Size;
    fprintf (drvui->fpoutp, "  text { ttf \"crystal.ttf\",\"%s\" 0.15,0\n",
        drvui->label_label[n]);
    fprintf (drvui->fpoutp, "    scale <%4.2f,%4.2f,%4.2f>\n",
        size, size, size);
    fprintf (drvui->fpoutp, "    rotate <0, 0, -zrot>\n");
    fprintf (drvui->fpoutp, "    rotate <0, -yrot, 0>\n");
    fprintf (drvui->fpoutp, "    rotate <-xrot,0, 0>\n");
    fprintf (drvui->fpoutp, "    translate <%8.5f,%8.5f,%8.5f> pigment{color Black}\n",
        s_vert[0], s_vert[1], s_vert[2]);
    fprintf (drvui->fpoutp, "  }\n");
    if (Vrml2) {
      fprintf (drvui->fpoutv, "   Transform {");
      fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f \n",
          s_vert[0], s_vert[1], s_vert[2]);
      fprintf (drvui->fpoutv, " rotation 1 0 0 %f\n",-xrot/RAD);
      fprintf (drvui->fpoutv, "children Transform { rotation 0 1 0 %f\n",-yrot/RAD);
      fprintf (drvui->fpoutv, "children Transform { rotation 0 0 1 %f\n",-zrot/RAD); 
      fprintf (drvui->fpoutv, "  children [ \n");
      fprintf (drvui->fpoutv, "   Billboard {\n    axisOfRotation 0 1 0\n");
      fprintf (drvui->fpoutv, "    children [ \n");
      fprintf (drvui->fpoutv, "     Shape {\n");
      fprintf (drvui->fpoutv, "      appearance Appearance{\n");
      fprintf (drvui->fpoutv, "       material Material {diffuseColor 0 0 0}\n");
      fprintf (drvui->fpoutv, "      }\n");
      fprintf (drvui->fpoutv, "      geometry Text{string [\"%s\"]}\n", drvui->label_label[n]);
      fprintf (drvui->fpoutv, "     }\n    ]\n");
      fprintf (drvui->fpoutv, "   }\n  ]\n }}}\n");
    } else {
      fprintf (drvui->fpoutv, " Separator {\n  Translation {");
      fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f }\n",
          s_vert[0], s_vert[1], s_vert[2]);
      fprintf (drvui->fpoutv, "  Rotation{ rotation 0 0 1 %8.5f}\n", -zrot / RAD);
      fprintf (drvui->fpoutv, "  Rotation{ rotation 0 1 0 %8.5f}\n", -yrot / RAD);
      fprintf (drvui->fpoutv, "  Rotation{ rotation 1 0 0 %8.5f}\n", -xrot / RAD);
      fprintf (drvui->fpoutv, "  Scale{ scaleFactor %4.2f %4.2f %4.2f }\n",Text_Size, Text_Size, Text_Size);
      fprintf (drvui->fpoutv, "  Material{ diffuseColor 0. 0. 0. }\n");
      fprintf (drvui->fpoutv, "  AsciiText { string  \"%s\" }\n", drvui->label_label[n]);
      fprintf (drvui->fpoutv, " }\n");
    }
  }             /* end of loop on n   */
  fprintf (drvui->fcns,"%4d labels.\n", Label_Count);
  fprintf (drvui->flout, "Generated %4d labels.\n", Label_Count);
}

