// module Generate_Drawing.cxx - part of DRAWxtl V5.1
// Coded using the FLTK 1.1.6 widget set
//
//     Larry W. Finger, Martin Kroeker and Brian Toby
//
// This module contains the following routines:
//
//  do_drawing - routine to put together all the elements of the drawing
//  Generate_Drawing - create the drawing for the GUI version of DRAWxtl
//  generate_slab - routine to generate slab parameters from the input values
//  show_slab - draw outline if selected slab if slabmode == 2
//
#include "drawxtl.h"
#include <stdio.h>
#include "draw_ext.h"
#ifndef Win32
#include <stdlib.h>
#include <string.h>
#endif
#include <FL/glut.h>
#include <FL/gl.h>
#include <FL/fl_ask.H>
#include <math.h>
#include "DRAWxtlViewUI.h"
#include "CrystalView.h"

#include "DRAWxtl_proto.h"

/* ************************************************************** */
/* ************************************************************** */

void do_drawing (void)

// routine to put together the elements of the drawing
{
  int i;

  nvert = 0;                         // empty vertex list
  if (drvui->frame_no == 1)
      draw_cell (docell);            // generate unit-cell corners and draw it
  nvert = 0;                         // empty vertex list
  generate_spheres ();               // generate all spheres to be drawn within box
  if (do_ellipsoids && n_ellips > 1)
      generate_ellipsoids ();        // now the ellipsoids
  if (drvui->nmag >0 )
      generate_arrows();             // magnetic moment arrows, if any
  generate_bonds ();                 // generate all bonds to be drawn within box
  generate_cones ();                 // generate all lonepairs to be drawn within box
  generate_poly ();                  // generate all polyhedra to be drawn
  generate_planes ();                // now the planes
  generate_texts ();                 // any textual elements
  if (slabmode > 1)
      show_slab();                   // show outline of cutout section 
  if (ReadFourMap) {
      for (i = 1; i <= numOfFourierContours; ++i) {
// generate & display a Fourier map
          generate_map (FourierContourLevel[i], FourierContourSolid[i],
            FourierContourColor[i]);
      }
  }
}

/* ************************************************************** */
/* ************************************************************** */

void Generate_Drawing(int keepmatrix)
{
  char filename[256] = "";
  int i, m;
  float temp;
  float rm[16];

  if (strlen(drvui->Cur_File ) == 0){
      return;
  }

  if (drvui->crystalDL) {
      glDeleteLists(drvui->crystalDL, 1);               // flush display list
      drvui->crystalDL = glGenLists (1);                // create a new one
  }
  if (!(drvui->fpin = fopen (drvui->Cur_Temp, "r"))) {
      Error_Box("Cannot open structure file");
      return;
  }
  strcpy(filename,drvui->Cur_Root);
  strcat (filename, ".out");
  if (!(drvui->flout = fopen (filename, "w"))) {
      Error_Box("Cannot open listing file");
      return;
  }
  strcpy(filename,drvui->Cur_Root);
  strcat (filename, ".wrl");
  if (!(drvui->fpoutv = fopen (filename, "w"))) {
      Error_Box ("Cannot open VRML output file\n");
      return;
  }
  strcpy(filename,drvui->Cur_Root);
  strcat (filename, ".pov");
  if (!(drvui->fpoutp = fopen (filename, "w"))) {
      Error_Box ("Cannot open POV output file\n");
      return;
  }
  strcpy(filename,drvui->Cur_Root);
  strcat (filename, ".cns");
  if (!(drvui->fcns = fopen (filename, "w"))) {
      Error_Box ("Cannot open console listing file\n");
      return;
  }

  if (drvui->atom_no) {
      free(drvui->atom_no);
      drvui->atom_no = NULL;
  }
  if (!(drvui->atom_no = (int *) malloc ((unsigned) ((long) (MAX_VERTS * sizeof (int)))))) {
      Error_Box("Unable to get initial atom_no allocation\n");
      return;
  }
  if (xypos) {
      free(xypos);
      xypos = NULL;
  }
  if (!(xypos = (float *) malloc ((unsigned) (3 * (long) (MAX_VERTS * sizeof (float)))))) {
      Error_Box("Unable to get initial xypos allocation\n");
      return;
  }
  fprintf (drvui->flout, "\n\n                DRAWxtl\n");
  fprintf (drvui->flout, "  Crystal Structure Drawing Program %s\n\n", VERSION);
  fprintf (drvui->flout,
         "  Copyright 1996, 1997, 2000-2005 by Larry Finger, Martin Kroeker and Brian Toby\n");
  fprintf (drvui->flout, "  (all rights reserved).\n\n");
  fprintf (drvui->flout, "     Combined POVxtl and VRMLxtl\n");
  fprintf (drvui->flout,
         "POVxtl  - Scenes for Persistence of Vision Ray Tracer (POV-Ray) V2.0\n");
  fprintf (drvui->flout,
         "VRMLxtl - Virtual Reality Modeling Language V1.0/V2.0\n");
  fprintf (drvui->fcns,"\n\n                DRAWxtl\n");
  fprintf (drvui->fcns,"  Crystal Structure Drawing Program %s\n\n", VERSION);
  fprintf (drvui->fcns,"  Copyright 1996, 1997, 2000-2005 by Larry Finger, Martin Kroeker and Brian Toby\n");
  fprintf (drvui->fcns,"  (all rights reserved).\n\n");
  fprintf (drvui->fcns,"     Combined POVxtl and VRMLxtl\n");
  fprintf (drvui->fcns,"POVxtl  - Scenes for Persistence of Vision Ray Tracer (POV-Ray) V2.0\n");
  fprintf (drvui->fcns,"VRMLxtl - Virtual Reality Modeling Language V1.0/V2.0\n");
  drvui->frame_no = 0;
  drvui->nsphere = drvui->npoly = drvui->nbond = drvui->nplane = drvui->ncone = 1;

// Initialize POV output limits

  POV_Max[0] = POV_Max[1] = POV_Max[2] = -99999.0f;
  POV_Min[0] = POV_Min[1] = POV_Min[2] = 99999.0f;
  natom = 0;
  NvertM = 1;                     // make master list empty
  drvui->nmag = 0;
  drvui->nlabel = 1;

// start loop through frames

  while (drvui->frame_no < drvui->max_frame) {
    drvui->frame_no++;
    fprintf (drvui->fcns, "\n doing frame %d of %d\n",drvui->frame_no,drvui->max_frame);
    fprintf (drvui->flout, "\n doing frame %d of %d\n",drvui->frame_no,drvui->max_frame);
    clipflag = 0;
    packflag = 0;
    domolcomp = 0;
    drvui->mol_d = 0.0f;
    drvui->mag_matrix[0][0]=1.0f;
    drvui->mag_matrix[1][0]=0.0f;
    drvui->mag_matrix[2][0]=0.0f;
    drvui->mag_matrix[2][1]=0.0f;
    drvui->mag_matrix[1][2]=0.0f;
    drvui->mag_matrix[0][1]=0.0f;
    drvui->mag_matrix[0][2]=0.0f;
    drvui->mag_matrix[1][1]=1.0f;
    drvui->mag_matrix[2][2]=1.0f;
    origin[0] = drvui->X_Origin;
    origin[1] = drvui->Y_Origin;
    origin[2] = drvui->Z_Origin;
    boxlim[0] = drvui->X_Boxlim;
    boxlim[1] = drvui->Y_Boxlim;
    boxlim[2] = drvui->Z_Boxlim;
    docell = Unit_Cell;
    do_ellipsoids = Color_Warning = 0;
    rad_cell = 0.02f;
    drvui->Sphere_Mult = drvui->SpMult;
    drvui->Bond_Mult = drvui->BndMult;
    slabmode = 0;
    n_ellips = 1;
    if (drvui->frame_no != 1) {
      drvui->fpin = fopen (drvui->Cur_Temp, "r");
    }
    get_input (0);

// Calculate Grand rotation matrix

    if (keepmatrix == 0) {
      Rotq = XYZ_Rot_to_Q(xrot, yrot, zrot);
      crystal->calculate(rm);         // get rotation matrix
      for (m = 0; m < 3; m++) {       //  and copy to G_Rot
        G_Rot[0][m] = rm[m];
        G_Rot[1][m] = rm[m + 4];
        G_Rot[2][m] = rm[m+8];
      }
    }
// generate all possible atoms to get the size of the drawing

    if (!s_vert) {
        if ((s_vert = (float *) malloc ((unsigned) (3 * (long) (MAX_VERTS 
                  * sizeof (float))))) == NULL) {
            Error_Box ("Unable to get initial s_vert allocation");
            return;
        }
    }
    get_atom_id ();                     // rework atom id numbers to pointers
    if (slabmode) generate_slab();	// initialize cutout box if needed
    build_box_contents ();              // complete the master atom list
    fprintf (drvui->fcns, "\nMaster atom list contains %5d atoms.\n", NvertM - 1);
    fprintf (drvui->flout, "\nMaster atom list contains %5d atoms.\n", NvertM - 1);

// allocate space for the various arrays (at least 14 cell vertices plus first atom - thus extra 20)

    if (o_vert)
        free(o_vert);
    if (!(o_vert = (float *)malloc ((unsigned) (6 * (NvertM+20) *    sizeof (float))))) {
      Error_Box ("Unable to get o_vert allocation");
      return;
    }
    if (vert_sym_no)
        free(vert_sym_no);
    if (!(vert_sym_no =(int *) malloc ((unsigned) (2 * NvertM * sizeof (int))))) {
      Error_Box ("Unable to get vert_sym_no allocation");
      return;
    }
    if (drvui->orig_atom_no)
        free(drvui->orig_atom_no);
    if (!(drvui->orig_atom_no =(int *) malloc ((unsigned) (2 * NvertM * sizeof (int))))) {
      Error_Box ("Unable to get orig_atom_no allocation");
      return;
    }
    if (!(poly_list = (int *) malloc ((unsigned) (4 * (NvertM + 20)) * sizeof (int)))) {
      Error_Box ("Unable to get poly_list allocation");
      return;
    }
    if (!(vertex_list = (int *) malloc ((unsigned) (4 * NvertM) * sizeof (int)))) {
      Error_Box ("Unable to get vertex_list allocation");
      return;
    }
    analyze_bonds ();                  // print the bond distance table
    fprintf (drvui->fcns, "\nThe 'pov' and 'wrl' files contain:\n");
    if (docell != 0 || natom == 0) {      // if unit-cell corners drawn, add them for scaling
      float vert[3];
      int saved_Nvert;

      saved_Nvert = NvertM;
      for (i = 0; i <= 2; ++i)
        vert[i] = 0.0f;
      add_vert (vert, 0, 1);             // place 0,0,0 in list
      vert[2] = 1.0f;
      add_vert (vert, 0, 1);             // place 0,0,1 in list
      vert[1] = 1.0f;
      add_vert (vert, 0, 1);             // place 0,1,1 in list
      vert[2] = 0.0f;
      add_vert (vert, 0, 1);             // place 0,1,0 in list
      vert[0] = 1.0f;
      add_vert (vert, 0, 1);             // place 1,1,0 in list
      vert[1] = 0.0f;
      add_vert (vert, 0, 1);             // place 1,0,0 in list
      vert[2] = 1.0f;
      add_vert (vert, 0, 1);             // place 1,0,1 in list
      vert[1] = 1.0f;
      add_vert (vert, 0, 1);             // place 1,1,1 in list
      NvertM = saved_Nvert;              // reset list count 
    }
    temp = POV_Max[0] - POV_Min[0];
    if (POV_Max[1] - POV_Min[1] > temp)
      temp = POV_Max[1] - POV_Min[1];
    if (POV_Max[2] - POV_Min[2] > temp)
      temp = POV_Max[2] - POV_Min[2];
    if (drvui->frame_no == 1) {        // only do for first frame
      Scale = 15.0f *temp / Magnification;

      if (Vrml2)
        fprintf (drvui->fpoutv, "#VRML V2.0 utf8\n\n");
      else
        fprintf (drvui->fpoutv, "#VRML V1.0 ascii\n\n");
      fprintf (drvui->fpoutv, "# Scene file created by DRAWxtl %s from %s\n",
           VERSION, drvui->Cur_File);
      fprintf (drvui->fpoutv,
           "# with command line options: -b %6.2f %6.2f %6.2f -o %5.2f %5.2f %5.2f\n",
           boxlim[0], boxlim[1], boxlim[2], origin[0], origin[1],
           origin[2]);
      fprintf (drvui->fpoutv,
           "#  -p %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f -v %5.2f %5.2f %5.2f\n",
           drvui->cryst_lim[drvui->frame_no][0], drvui->cryst_lim[drvui->frame_no][3], 
           drvui->cryst_lim[drvui->frame_no][1], drvui->cryst_lim[drvui->frame_no][4],
           drvui->cryst_lim[drvui->frame_no][2], drvui->cryst_lim[drvui->frame_no][5], xrot, yrot, zrot);

// Output VRML header

      if (!Vrml2) {
        fprintf (drvui->fpoutv,
             "DEF BackgroundColor Info { string \"%s\" }\n", drvui->col_bg_v);
        fprintf (drvui->fpoutv, "Separator {\n");
        fprintf (drvui->fpoutv, " Material {\n");
        fprintf (drvui->fpoutv, "  ambientColor    0.1 0.1 0.1\n");
        fprintf (drvui->fpoutv, "  diffuseColor    1 1 1\n");
        fprintf (drvui->fpoutv, "  specularColor   0.9 0.9 0.9\n");
        fprintf (drvui->fpoutv, "  shininess       0.8\n");
        fprintf (drvui->fpoutv, " }\n");
      }
      Text_Size = 0.0005f *Scale;

      if (!M_cameras && Vrml2) {         // Generate Orthographic Projection for VRML97
        fprintf (drvui->fpoutv, "NavigationInfo {type \"EXAMINE\"}\n");
        fprintf (drvui->fpoutv, "Background { skyColor %s }\n", drvui->col_bg_v);
        fprintf (drvui->fpoutv,
             "Viewpoint { fieldOfView 0.1 position 0. 0. %8.3f }\n",
             Scale);
        fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 -1}\n");
        fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 1}\n");
      } else {
        if (!Vrml2) {                 // simulate orthographic projection for VRML1.0
          float fscale, fangle;

          if (!M_cameras) {
            fscale = Scale * (10.0f / 7.0f);
            fangle = 0.07853f;
          } else {
            fscale = Scale / 7.0f;
            fangle = 0.7853f;
          }
          fprintf (drvui->fpoutv, "PerspectiveCamera {\n");
          fprintf (drvui->fpoutv, "  position        0 0 %8.3f\n", fscale);
          fprintf (drvui->fpoutv, "  orientation     0 0 1 0\n");
          fprintf (drvui->fpoutv, "  focalDistance %8.3f\n", fscale);
          fprintf (drvui->fpoutv, "  heightAngle   %8.5f\n", fangle);
          fprintf (drvui->fpoutv, "}\n");
          fprintf (drvui->fpoutv,
               "DirectionalLight{ direction 0 0 -1 intensity 1.0}\n");
          fprintf (drvui->fpoutv,
               "DirectionalLight{ direction 0 0 1 intensity 1.0}\n");
        } else {
          fprintf (drvui->fpoutv, "NavigationInfo {type \"EXAMINE\"}\n");
          fprintf (drvui->fpoutv, "Background { skyColor %s }\n", drvui->col_bg_v);
          fprintf (drvui->fpoutv, "Viewpoint { position 0. 0. %8.3f }\n",
               Scale / 7.);
          fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 -1}\n");
          fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 1}\n");
        }
      }

      if (Vrml2) {
        fprintf (drvui->fpoutv, "Transform { rotation 0 0 1 %f\n", zrot / RAD);
      } else {
        fprintf (drvui->fpoutv, "Rotation {");
        fprintf (drvui->fpoutv, " rotation 0 0 1 %f}\n", zrot / RAD);
      }
      if (Vrml2) {
        fprintf (drvui->fpoutv, " children Transform {");
        fprintf (drvui->fpoutv, " rotation 0 1 0 %f\n", yrot / RAD);
      } else {
        fprintf (drvui->fpoutv, " Rotation {");
        fprintf (drvui->fpoutv, " rotation 0 1 0 %f}\n", yrot / RAD);
      }
      if (Vrml2) {
        fprintf (drvui->fpoutv, " children Transform {");
        fprintf (drvui->fpoutv, " rotation 1 0 0 %f\n", xrot / RAD);
      } else {
        fprintf (drvui->fpoutv, " Rotation {");
        fprintf (drvui->fpoutv, " rotation 1 0 0 %f}\n", xrot / RAD);
      }

      if (Vrml2)
        fprintf (drvui->fpoutv, "  children [\n");

// Output POV header

      fprintf (drvui->fpoutp, "/*\n\nPOVxtl-Persistence of Vision Ray Tracer\n");
      fprintf (drvui->fpoutp, "Crystal Structure Drawing Program %s\n\n",
           VERSION);
      fprintf (drvui->fpoutp,
           "Copyright 1996, 1997, 2000-2005 by Larry Finger, Martin Kroeker and Brian Toby\n");
      fprintf (drvui->fpoutp, "   (all rights reserved).\n\n");
      fprintf (drvui->fpoutp, "Scene file created from %s \n", drvui->Cur_File);
      fprintf (drvui->fpoutp,
           "with command line options: -b %6.2f %6.2f %6.2f -o %5.2f %5.2f %5.2f \n",
           boxlim[0], boxlim[1], boxlim[2], origin[0], origin[1],
           origin[2]);
      fprintf (drvui->fpoutp,
           "-p %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f -v %5.2f %5.2f %5.2f -m %5.2f\n*/\n\n",
           drvui->cryst_lim[drvui->frame_no][0], drvui->cryst_lim[drvui->frame_no][3], 
           drvui->cryst_lim[drvui->frame_no][1], drvui->cryst_lim[drvui->frame_no][4],
           drvui->cryst_lim[drvui->frame_no][2], drvui->cryst_lim[drvui->frame_no][5], 
           xrot, yrot, zrot, Magnification);

      fprintf (drvui->fpoutp, "#include \"colors.inc\"\n");
      fprintf (drvui->fpoutp, "#include \"chars.inc\"\n");

      if (drvui->ambient > 0.) {
        fprintf (drvui->fpoutp, "\ndefault {\n");
        fprintf (drvui->fpoutp, "        finish {\n");
        fprintf (drvui->fpoutp, "                ambient %5.2f\n",drvui->ambient);
        fprintf (drvui->fpoutp, "                diffuse %5.2f\n",drvui->diffuse);
        fprintf (drvui->fpoutp, "                specular %5.2f\n",drvui->specular);
        fprintf (drvui->fpoutp, "                roughness %5.2f\n",drvui->roughness);
        fprintf (drvui->fpoutp, "               }\n");
        fprintf (drvui->fpoutp, "        }\n\n");
      }	
      if (!M_cameras) {
        fprintf (drvui->fpoutp, "camera{ orthographic\n");
      } else {
        fprintf (drvui->fpoutp, "camera{\n");
      }
      Scale *= 0.9f;
      fprintf (drvui->fpoutp, " location< 0 ,0, %f>\n", 1.6f * Scale);
      if (!M_cameras) {
        fprintf (drvui->fpoutp, " up <0,%f,0>\n", 0.1 * Scale);
        fprintf (drvui->fpoutp, " right <%f,0,0>\n", -0.1 * Scale);
      } else {
        fprintf (drvui->fpoutp, " up <0, 0.1, 0>\n");
        fprintf (drvui->fpoutp, " right <-0.1, 0, 0>\n");
      }
      fprintf (drvui->fpoutp, " look_at <0,0,0>\n");
      fprintf (drvui->fpoutp, "}\n");
      fprintf (drvui->fpoutp, "background{color %s}\n", drvui->col_bg_p);
      fprintf (drvui->fpoutp, "light_source {< 0, 0, %f>\n", Scale);
      fprintf (drvui->fpoutp, "color red 4.0 green 4.0 blue 4.0\n}\n");
      fprintf (drvui->fpoutp, "object{\n union{\n");
      fprintf (drvui->fpoutp, "\n");
      fprintf (drvui->fpoutp, "#declare xrot = %5.1f;\n", xrot);
      fprintf (drvui->fpoutp, "#declare yrot = %5.1f;\n", yrot);
      fprintf (drvui->fpoutp, "#declare zrot = %5.1f;\n", zrot);
      fprintf (drvui->fpoutp, "\n");
      glNewList (drvui->crystalDL, GL_COMPILE);
    } else {
      fprintf (drvui->fpoutp, "\n/************ Frame %d **********/\n", drvui->frame_no);
      fprintf (drvui->fpoutv, "\n############# Frame %d ##########\n", drvui->frame_no);
    }                     // only do for drvui->frame_no == 1

    do_drawing ();      // generate all elements of the drawing for this frame
    free (vertex_list);
    free (poly_list);
  }                    // while drvui->frame_no < drvui->max_frame
  drvui->frame_no = drvui->max_frame;
  glEndList ();
  if (gl_size == 0.0f) {
    gl_size = max(POV_Max[1]-POV_Min[1],POV_Max[0]-POV_Min[0]);
    gl_size = max(gl_size,POV_Max[2]-POV_Min[2]);
  }

  if (drvui->noshadow == 1 ) 
    fprintf (drvui->fpoutp, " } no_shadow\n");
  else
    fprintf (drvui->fpoutp, " }\n");
  fprintf (drvui->fpoutp, "rotate <xrot, 0, 0>\n");
  fprintf (drvui->fpoutp, "rotate <0, yrot, 0>\n");
  fprintf (drvui->fpoutp, "rotate <0, 0, zrot>\n");
  fprintf (drvui->fpoutp, "}\n");
  if (Vrml2)
    fprintf (drvui->fpoutv, "]\n");
  fprintf (drvui->fpoutv, " }\n");
  if (Vrml2) {
      fprintf (drvui->fpoutv, " }\n");
      fprintf (drvui->fpoutv, " }\n");
  }
  fprintf(drvui->flout,"End of File.\n");
  fclose (drvui->fpoutv);
  fclose (drvui->fpoutp);
  fclose (drvui->flout);
  fclose (drvui->fcns);
}                         // end of Generate_Drawing

/* ************************************************************** */
/* ************************************************************** */

void generate_slab(void)
{
// generate the slab parameters from the input data

  float csal,snal,csbe,snbe,csga,snga;
  float rot_mat[3][3];
  float bmat[3][3];
  float vert[3];
  int i,j;

  for (i = 3; i <= 5; ++i)
    if (drvui->slab_con[i] <= 0)
      drvui->slab_con[i] = 90.0f;
  for (i = 0; i <= 2; ++i)
    for (j = 0; j <= 2; ++j)
      bmat[j][i] = 0.0f;         // initialize matrix
  bmat[0][0] = 1.0f;
  csal = (float) cos (drvui->slab_con[3] / RAD);
  snal = (float) sin (drvui->slab_con[3] / RAD);
  csbe = (float) cos (drvui->slab_con[4] / RAD);
  snbe = (float) sin (drvui->slab_con[4] / RAD);
  csga = (float) cos (drvui->slab_con[5] / RAD);
  snga = (float) sin (drvui->slab_con[5] / RAD);
  bmat[0][1] = csga;
  bmat[1][1] = snga;
  bmat[0][2] = csbe;
  bmat[1][2] = (csal - csbe * csga) / snga;
  bmat[2][2] =    (float) sqrt (1.0 - bmat[0][2] * bmat[0][2] -
              bmat[1][2] * bmat[1][2]);
  for (i = 0; i <= 2; ++i)
    for (j = 0; j <= 2; ++j)
      bmat[j][i] *= drvui->slab_con[i];

  csal = (float) cos (drvui->slab_rot[0] / RAD);
  snal = (float) sin (drvui->slab_rot[0] / RAD);
  csbe = (float) cos (drvui->slab_rot[1] / RAD);
  snbe = (float) sin (drvui->slab_rot[1] / RAD);
  csga = (float) cos (drvui->slab_rot[2] / RAD);
  snga = (float) sin (drvui->slab_rot[2] / RAD);
  rot_mat[0][0] = csbe * csga;
  rot_mat[0][1] = snal * snbe * csga + csal * snga;
  rot_mat[0][2] = -csal * snbe * csga + snal * snga;
  rot_mat[1][0] = -csbe * snga;
  rot_mat[1][1] = -snal * snbe * snga + csal * csga;
  rot_mat[1][2] = csal * snbe * snga + snal * csga;
  rot_mat[2][0] = snbe;
  rot_mat[2][1] = -snal * csbe;
  rot_mat[2][2] = csal * csbe;

/* 0 0 0 */
  for (i = 0; i <= 2; ++i)
    vert[i] = (float) 0.0;
  slabx1=0.;
  slaby1=0.;
  slabz1=0.;

/* 0 0 1 */
  vert[2] = (float) 1.0;
  slabx2=0.;
  for (i = 0; i <= 2; ++i)
    slabx2 += (float) (bmat[0][i] * vert[i]);
  slaby2=0.;
  for (i = 0; i <= 2; ++i)
    slaby2 += (float) (bmat[1][i] * vert[i]);
  slabz2=0.;
  for (i = 0; i <= 2; ++i)
    slabz2 += (float) (bmat[2][i] * vert[i]);
 
/* 0 1 0 */
  vert[1] = (float) 1.0;
  vert[2] = (float) 0.0;
  slabx3=0.;
  for (i = 0; i <= 2; ++i)
    slabx3 += (float) (bmat[0][i] * vert[i]);
  slaby3=0.;
  for (i = 0; i <= 2; ++i)
    slaby3 += (float) (bmat[1][i] * vert[i]);
  slabz3=0.;
  for (i = 0; i <= 2; ++i)
    slabz3 += (float) (bmat[2][i] * vert[i]);

/* 1 0 0 */
  vert[0] = (float) 1.0;
  vert[1] = (float) 0.0;
  slabx4=0.;
  for (i = 0; i <= 2; ++i)
    slabx4 += (float) (bmat[0][i] * vert[i]);
  slaby4=0.;
  for (i = 0; i <= 2; ++i)
    slaby4 += (float) (bmat[1][i] * vert[i]);
  slabz4=0.;
  for (i = 0; i <= 2; ++i)
    slabz4 += (float) (bmat[2][i] * vert[i]);


  slabx1 -= drvui->slab_off[0];
  slabx2 -= drvui->slab_off[0];
  slabx3 -= drvui->slab_off[0];
  slabx4 -= drvui->slab_off[0];
  slaby1 -= drvui->slab_off[1];
  slaby2 -= drvui->slab_off[1];
  slaby3 -= drvui->slab_off[1];
  slaby4 -= drvui->slab_off[1];

  slabz1 -= drvui->slab_off[2];
  slabz2 -= drvui->slab_off[2];
  slabz3 -= drvui->slab_off[2];
  slabz4 -= drvui->slab_off[2];

  vert[0]=slabx1 * rot_mat[0][0]+slaby1 * rot_mat[1][0]+slabz1 * rot_mat[2][0];
  vert[1]=slabx1 * rot_mat[0][1]+slaby1 * rot_mat[1][1]+slabz1 * rot_mat[2][1];
  vert[2]=slabx1 * rot_mat[0][2]+slaby1 * rot_mat[1][2]+slabz1 * rot_mat[2][2];
  slabx1=vert[0];
  slaby1=vert[1];
  slabz1=vert[2];
  vert[0]=slabx2 * rot_mat[0][0]+slaby2 * rot_mat[1][0]+slabz2 * rot_mat[2][0];
  vert[1]=slabx2 * rot_mat[0][1]+slaby2 * rot_mat[1][1]+slabz2 * rot_mat[2][1];
  vert[2]=slabx2 * rot_mat[0][2]+slaby2 * rot_mat[1][2]+slabz2 * rot_mat[2][2];
  slabx2=vert[0];
  slaby2=vert[1];
  slabz2=vert[2];
  vert[0]=slabx3 * rot_mat[0][0]+slaby3 * rot_mat[1][0]+slabz3 * rot_mat[2][0];
  vert[1]=slabx3 * rot_mat[0][1]+slaby3 * rot_mat[1][1]+slabz3 * rot_mat[2][1];
  vert[2]=slabx3 * rot_mat[0][2]+slaby3 * rot_mat[1][2]+slabz3 * rot_mat[2][2];
  slabx3=vert[0];
  slaby3=vert[1];
  slabz3=vert[2];
  vert[0]=slabx4 * rot_mat[0][0]+slaby4 * rot_mat[1][0]+slabz4 * rot_mat[2][0];
  vert[1]=slabx4 * rot_mat[0][1]+slaby4 * rot_mat[1][1]+slabz4 * rot_mat[2][1];
  vert[2]=slabx4 * rot_mat[0][2]+slaby4 * rot_mat[1][2]+slabz4 * rot_mat[2][2];
  slabx4=vert[0];
  slaby4=vert[1];
  slabz4=vert[2];
}

/* ************************************************************** */
/* ************************************************************** */

void show_slab(void)
{
// show the outlines of the slab when slabmode == 2
  glPushMatrix ();
  glLineWidth(3);
  glColor3f (1., 0., 0.);
  glBegin (GL_LINES);

  glVertex3f (slabx1, slaby1, slabz1);
  glVertex3f (slabx2, slaby2, slabz2);

  glVertex3f (slabx3, slaby3, slabz3);
  glVertex3f (slabx1, slaby1, slabz1);

  glVertex3f (slabx3, slaby3, slabz3);
  glVertex3f (slabx2+(slabx3-slabx1), slaby2+(slaby3-slaby1),slabz2+(slabz3-slabz1));

  glVertex3f (slabx2, slaby2, slabz2);
  glVertex3f (slabx2+(slabx3-slabx1), slaby2+(slaby3-slaby1),slabz2+(slabz3-slabz1));

  glVertex3f (slabx1, slaby1, slabz1);
  glVertex3f (slabx4, slaby4, slabz4);


  glVertex3f (slabx3, slaby3, slabz3);
  glVertex3f (slabx3+(slabx4-slabx1), slaby3+(slaby4-slaby1),slabz3+(slabz4-slabz1));

  glVertex3f (slabx4, slaby4, slabz4);
  glVertex3f (slabx3+(slabx4-slabx1), slaby3+(slaby4-slaby1),slabz3+(slabz4-slabz1));

  glVertex3f (slabx2, slaby2, slabz2);
  glVertex3f (slabx2+(slabx4-slabx1),slaby2+(slaby4-slaby1),slabz2+(slabz4-slabz1));


  glVertex3f (slabx2+(slabx4-slabx1),slaby2+(slaby4-slaby1),slabz2+(slabz4-slabz1));
  glVertex3f (slabx4, slaby4, slabz4);

  glVertex3f (slabx2+(slabx4-slabx1),slaby2+(slaby4-slaby1),slabz2+(slabz4-slabz1));
  glVertex3f (slabx2+(slabx4-slabx1)+slabx3-slabx1,
              slaby2+(slaby4-slaby1)+slaby3-slaby1,
              slabz2+(slabz4-slabz1)+slabz3-slabz1);


  glVertex3f (slabx3+(slabx4-slabx1), slaby3+(slaby4-slaby1),slabz3+(slabz4-slabz1));
  glVertex3f (slabx2+(slabx4-slabx1)+slabx3-slabx1,
              slaby2+(slaby4-slaby1)+slaby3-slaby1,
              slabz2+(slabz4-slabz1)+slabz3-slabz1);

  glVertex3f (slabx2+(slabx3-slabx1), slaby2+(slaby3-slaby1),slabz2+(slabz3-slabz1));
  glVertex3f (slabx2+(slabx4-slabx1)+slabx3-slabx1,
              slaby2+(slaby4-slaby1)+slaby3-slaby1,
              slabz2+(slabz4-slabz1)+slabz3-slabz1);
  glEnd ();
  glPopMatrix ();
  glLineWidth(1);
}

/* ************************************************************** */
/* ************************************************************** */

void label_cell(void)
{
// generate the labels corresponding to the unit-cell axes
  int i, j;
  int used[4];
  int old_type = 1;
  if (Labels) {

// initialize the saved_x's  - first pass only
      
      if (!drvui->labels_inited) {
          drvui->saved_x_label[0][0][0] = -0.5f/drvui->lat_con[0];
          drvui->saved_x_label[0][0][1] = -0.5f/drvui->lat_con[1];
          drvui->saved_x_label[0][0][2] = -0.5f/drvui->lat_con[2];
          drvui->saved_x_label[0][3][2] = 1.0f+0.5f/drvui->lat_con[2];
          drvui->saved_x_label[1][3][2] = 1.4f/drvui->lat_con[2];
          drvui->saved_x_label[0][2][1] = 1.0f+0.5f/drvui->lat_con[1];
          drvui->saved_x_label[1][2][1] = 1.4f/drvui->lat_con[1];
          drvui->saved_x_label[0][1][0] = 1.0f+0.5f/drvui->lat_con[0];
          drvui->saved_x_label[1][1][0] = 1.4f/drvui->lat_con[0];
          drvui->labels_inited = 1;
          for (i=0; i<4; i++)
              drvui->triple[i] = 0;
      }
      if (fabs(offset[0]) + fabs(offset[1]) + fabs(offset[2]) < 0.01)
          Locate_Triple();
      for (i=0; i<4; i++)
          used[i] = 0;

/* find any labels that are already in the list and save their current position */

      if (drvui->nlabel > 1) {
          for (i=1; i<drvui->nlabel; i++) {
              if (!strcmp(drvui->label_label[i],"o")) {
                  drvui->triple[0] = 0;
                  for (j=0; j<3; j++)
                      drvui->saved_x_label[0][0][j] = drvui->label_x[i][j];
                  used[0] = 1;
                  old_type = 0;
              }
              if (!strcmp(drvui->label_label[i],"triple_vect")) {
                  drvui->triple[0] = i;
                  for (j=0; j<3; j++)
                      drvui->saved_x_label[1][0][j] = drvui->label_x[i][j];
                  used[0] = 1;
                  old_type = 1;
              }
          }
          for (i=1; i<drvui->nlabel; i++) {
              if (!strcmp(drvui->label_label[i],"a")) {
                  drvui->triple[1] = i;
                  for (j=0; j<3; j++)
                      drvui->saved_x_label[old_type][1][j] = drvui->label_x[i][j];
                  used[1] = 1;
              }
              if (!strcmp(drvui->label_label[i],"b")) {
                  drvui->triple[2] = i;
                  for (j=0; j<3; j++)
                      drvui->saved_x_label[old_type][2][j] = drvui->label_x[i][j];
                  used[2] = 1;
              }
              if (!strcmp(drvui->label_label[i],"c")) {
                  drvui->triple[1] = i;
                  for (j=0; j<3; j++)
                      drvui->saved_x_label[old_type][3][j] = drvui->label_x[i][j];
                  used[3] = 1;
              }
          }
      }

/* following section adds labels to the list for those not already supplied */

      if (!used[0]) {
          drvui->label_fn[drvui->nlabel]=1;
          for (j=0; j<3; j++) {
              drvui->label_x[drvui->nlabel][j] = drvui->saved_x_label[Display_axes][0][j];
          }
          if (!Display_axes) {
              drvui->triple[0] = 0;
              strcpy(drvui->label_label[drvui->nlabel++],"o");
          } else {
              drvui->triple[0] = drvui->nlabel;
              strcpy(drvui->label_label[drvui->nlabel++],"triple_vect");
          }
      }
      if (!used[1]) {
          drvui->label_fn[drvui->nlabel]=1;
          for (j=0; j<3; j++) {
              drvui->label_x[drvui->nlabel][j] = drvui->saved_x_label[Display_axes][1][j];
          }
          drvui->triple[1] = drvui->nlabel;
          strcpy(drvui->label_label[drvui->nlabel++],"a");
      }
      if (!used[2]) {
          drvui->label_fn[drvui->nlabel]=1;
          for (j=0; j<3; j++) {
              drvui->label_x[drvui->nlabel][j] = drvui->saved_x_label[Display_axes][2][j];
          }
          drvui->triple[2] = drvui->nlabel;
          strcpy(drvui->label_label[drvui->nlabel++],"b");
      }
      if (!used[3]) {
          drvui->label_fn[drvui->nlabel]=1;
          for (j=0; j<3; j++) {
              drvui->label_x[drvui->nlabel][j] = drvui->saved_x_label[Display_axes][3][j];
          }
          drvui->triple[3] = drvui->nlabel;
          strcpy(drvui->label_label[drvui->nlabel++],"c");
      }

/* The following section will copy the appropriate saved values into the active label parameters */

      for (i=1; i<drvui->nlabel; i++) {
          if (Display_axes) {
              if (!strcmp(drvui->label_label[i],"o")) {
                  strcpy(drvui->label_label[i],"triple_vect");
                  drvui->triple[0] = i;
                  for (j=0; j<3; j++) {
                      drvui->label_x[i][j] = drvui->saved_x_label[1][0][j];
                  }
              }
          } else {
              if (!strcmp(drvui->label_label[i],"triple_vect")) {
                  strcpy(drvui->label_label[i],"o");
                  drvui->triple[0] = 0;
                  for (j=0; j<3; j++) {
                      drvui->label_x[i][j] = drvui->saved_x_label[0][0][j];
                  }
              }
          }
          if (!strcmp(drvui->label_label[i],"c")) {
              drvui->triple[3] = i;
              for (j=0; j<3; j++) {
                  drvui->label_x[i][j] = drvui->saved_x_label[Display_axes][3][j];
              }
          }
          if (!strcmp(drvui->label_label[i],"b")) {
              drvui->triple[2] = i;
              for (j=0; j<3; j++) {
                  drvui->label_x[i][j] = drvui->saved_x_label[Display_axes][2][j];
              }
          }
          if (!strcmp(drvui->label_label[i],"a")) {
              drvui->triple[1] = i;
              for (j=0; j<3; j++) {
                  drvui->label_x[i][j] = drvui->saved_x_label[Display_axes][1][j];
              }
          }
      }
  }
}
