// Read_Input.cxx - Source module for DRAWxtl V5.1
// Coded using the FLTK 1.1.6 widget set
//
//     Larry W. Finger, Martin Kroeker and Brian Toby
//
//
// This module contains the following routines:
//
//  Blank_Strip - strips leading spaces from a string
//  check_atom_name - return 1 if input names match, 0 otherwise
//  check_color - check that a color name is found in POV color file, and generate RGB values
//  expand_atom - generates all atoms of a given type within the unit cell
//  findsys - determines crystal system from symmetry information
//  get_label - gets 1-4 character atom label from input string
//  getsym - obtain symmetry information from a SHELX-style SYMM card
//  import_cif - get structural information from a CIF (Crystallographic Information File)
//  import_fdat - get structural information from a CSD- or FDAT-format file
//  import_gsas - get structural information from a GSAS-format file
//  import_schakal - get structural information from a SCHAKAL-format file
//  import_shelx - get structural information from a SHELX-format file
//  Init_DRAWxtl - initialize all global variables
//  P_to_C - convert ellipsoid probability to scaling parameter
//  read_fcf - read CIF-standard Fo/Fc file and compute density
//  read_grd - read GSAS (grd) Fourier file
//  read_stf - read JANA (stf) Fourier file
//  read_flp - read FullProf (flp) Fourier file
//  read_w2k - read WIEN2k calculated charge density file
//  read_vasp - read VASP calculated charge density file
//  read_inp - read and process input ('.str') file
//  Token_Strip - strips tokens from the beginning of a string
//  trim_string - trim color strings
//  Unique_Atom - checks that atom position is unique
//
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#ifndef WIN32
  #include <ctype.h>
#else
  #define snprintf _snprintf
#endif
#include "drawxtl.h"
#include "draw_ext.h"
#include "DRAWxtlViewUI.h"

#include "DRAWxtl_proto.h"

/* ************************************************************** */
/* ************************************************************** */

void Blank_Strip (char input[100])

/* routine to suppress any leading spaces in the input string */
{
  int i, j;

  for (i = 0; i < 100; i++)
    if (input[i] != ' ')
      break;
  for (j = 0; j < 100; j++) {
    input[j] = input[i];
    if (input[i++] == '\0')
      break;
  }
}

/* ************************************************************** */
/* ************************************************************** */

int check_atom_name(char * name1, char * name2)
{
// determine if two atom names are the same. Return 1 (TRUE) if they are

    if (name1[0] == name2[0] && name1[1] == name2[1] &&
        name1[2] == name2[2] && name1[3] == name2[3])
        return 1;
    else
        return 0;
}

/* ************************************************************** */
/* ************************************************************** */

void check_color (char *input)
{
  char *transp_start;
  char transp[20];
  float fval;
  int i = strlen(input);

char *a,b[30],line[81];
char thecolor[50];
char keyword[80];
float thered,thegreen,theblue;
char POV_incpath[255]="\0";
FILE *colinc;

/*
   VRML routine to check color names and replace them with RGB triples
   derived from 'colors.inc' of POV
*/

  if (input[i-1] < ' ') input[i-1] = 0;        // get rid of any ^J character
  i = strlen(input);
  if (input[i-1] < ' ') input[i-1] = 0;        // get rid of any ^J character
  transp[0] = '\0';
  transp_start = strstr (input, "filter");
  if (transp_start) {
    fval = (float) atof (transp_start + 6);
    sprintf (transp, " transparency %5.2f", fval / 2.);
  }

  if (!strncmp (input, "Black", 5)) {
    strncpy (input, "0 0 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Brass", 5)) {
    strncpy (input, ".71 .65 .26\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Bronze2", 7)) {
    strncpy (input, ".65 .49 .24\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Bronze", 6)) {
    strncpy (input, ".55 .47 .14\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "BlueViolet", 10)) {
    strncpy (input, ".62352 .372549 .623529\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Blue", 3)) {
    strncpy (input, "0 0 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Brown", 5)) {
    strncpy (input, ".647059 .164706 .164706\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Copper", 6)) {
    strncpy (input, ".72 .45 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Cyan", 3)) {
    strncpy (input, "0 1 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DimGray", 7)) {
    strncpy (input, ".329412 .329412 .329412\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DimGrey", 7)) {
    strncpy (input, ".329412 .329412 .329412\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Aquamarine", 10)) {
    strncpy (input, ".439216 .858824 .576471\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "CadetBlue", 9)) {
    strncpy (input, ".372549 .623529 .623529\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Coral", 5)) {
    strncpy (input, "1.0 .498039 .0\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "CornflowerBlue", 14)) {
    strncpy (input, ".258824 .258824 .435294\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DarkGreen", 9)) {
    strncpy (input, ".184314 .309804 .184314\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DarkOliveGreen", 14)) {
    strncpy (input, ".309804 .309804 .184314\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DarkOrchid", 10)) {
    strncpy (input, ".6 .196078 .8\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DarkSlateBlue", 13)) {
    strncpy (input, ".419608 .137255 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DarkSlateGray", 13)) {
    strncpy (input, ".184314 .309804 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DarkSlateGrey", 13)) {
    strncpy (input, ".184314 .309804 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DarkTurquoise", 13)) {
    strncpy (input, ".439216 .576471 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Firebrick", 9)) {
    strncpy (input, ".556863 .137255 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "ForestGreen", 11)) {
    strncpy (input, ".137255 .556863 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Goldenrod", 9)) {
    strncpy (input, ".858824 .858824 .439216\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gold", 4)) {
    strncpy (input, ".8 .498039 .196078\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray05", 6)) {
    strncpy (input, ".05 .05 .05\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray10", 6)) {
    strncpy (input, ".10 .10 .10\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray15", 6)) {
    strncpy (input, ".15 .15 .15\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray20", 6)) {
    strncpy (input, ".20 .20 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray25", 6)) {
    strncpy (input, ".25 .25 .25\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray30", 6)) {
    strncpy (input, ".30 .30 .30\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray35", 6)) {
    strncpy (input, ".35 .35 .35\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray40", 6)) {
    strncpy (input, ".40 .40 .40\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray45", 6)) {
    strncpy (input, ".45 .45 .45\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray50", 6)) {
    strncpy (input, ".50 .50 .50\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray55", 6)) {
    strncpy (input, ".55 .55 .55\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray60", 6)) {
    strncpy (input, ".60 .60 .60\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray65", 6)) {
    strncpy (input, ".65 .65 .65\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray70", 6)) {
    strncpy (input, ".70 .70 .70\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray75", 6)) {
    strncpy (input, ".75 .75 .75\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray80", 6)) {
    strncpy (input, ".80 .80 .80\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray85", 6)) {
    strncpy (input, ".85 .85 .85\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray90", 6)) {
    strncpy (input, ".90 .90 .90\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray95", 6)) {
    strncpy (input, ".95 .95 .95\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Gray", 4)) {
    strncpy (input, ".752941 .752941 .752941\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Grey", 4)) {
    strncpy (input, ".752941 .752941 .752941\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "GreenYellow", 11)) {
    strncpy (input, ".576471 .858824 .439216\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Green", 5)) {
    strncpy (input, "0 1 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "IndianRed", 9)) {
    strncpy (input, ".309804 .184314 .184314\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Khaki", 5)) {
    strncpy (input, ".623529 .623529 .372549\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "LightBlue", 9)) {
    strncpy (input, ".74902 .847059 .847059\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "LightGray", 9)) {
    strncpy (input, ".658824 .658824 .658824\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "LightGrey", 9)) {
    strncpy (input, ".658824 .658824 .658824\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Light_Purple", 12)) {
    strncpy (input, ".87 .58 .98\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "LightSteelBlue", 14)) {
    strncpy (input, ".560784 .560784 .737255\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "LimeGreen", 9)) {
    strncpy (input, ".196078 .8 .196078\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Magenta", 7)) {
    strncpy (input, "1 0 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Maroon", 6)) {
    strncpy (input, ".556863 .137255 .419608\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MediumAquamarine", 16)) {
    strncpy (input, ".196078 .8 .6\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MediumBlue", 10)) {
    strncpy (input, ".196078 .196078 .8\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MediumForestGreen", 17)) {
    strncpy (input, ".419608 .556863 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MediumGoldenrod", 15)) {
    strncpy (input, ".917647 .917647 .678431\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MediumOrchid", 12)) {
    strncpy (input, ".576471 .439216 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MediumSeaGreen", 14)) {
    strncpy (input, ".258824 .435294 .258824\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MediumSlateBlue", 15)) {
    strncpy (input, ".498039 1.0\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MediumSpringGreen", 17)) {
    strncpy (input, ".498039 1.0 0\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MediumTurquoise", 15)) {
    strncpy (input, ".439216 .858824 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MediumVioletRed", 15)) {
    strncpy (input, ".858824 .439216 .576471\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MidnightBlue", 12)) {
    strncpy (input, ".184314 .184314 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "NavyBlue", 8)) {
    strncpy (input, ".137255 .137255 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Navy", 4)) {
    strncpy (input, ".137255 .137255 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "OrangeRed", 9)) {
    strncpy (input, "1.0 .498039\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Orange", 6)) {
    strncpy (input, "1 .5 .0\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Orchid", 6)) {
    strncpy (input, ".858824 .439216 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "PaleGreen", 9)) {
    strncpy (input, ".560784 .737255 .560784\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Pink", 4)) {
    strncpy (input, ".737255 .560784 .560784\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Plum", 4)) {
    strncpy (input, ".917647 .678431 .917647\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Red", 3)) {
    strncpy (input, "1 0 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "RichBlue", 8)) {
    strncpy (input, ".35 .35 .67\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Salmon", 6)) {
    strncpy (input, ".435294 .258824 .258824\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "SeaGreen", 8)) {
    strncpy (input, ".137255 .556863 .419608\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Sienna", 6)) {
    strncpy (input, ".556863 .419608 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "SkyBlue", 7)) {
    strncpy (input, ".196078 .6 .8\0", 14);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "SlateBlue", 9)) {
    strncpy (input, "0 .498039 1\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "SpringGreen", 11)) {
    strncpy (input, "0 1 .498039\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "SteelBlue", 9)) {
    strncpy (input, ".137255 .419608 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "SummerSky", 9)) {
    strncpy (input, ".22 .69 .87\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Tan", 3)) {
    strncpy (input, ".858824 .576471 .439216\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Thistle", 7)) {
    strncpy (input, ".847059 .74902 .847059\0", 23);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Turquoise", 9)) {
    strncpy (input, ".678431 .917647 .917647\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "VioletRed", 9)) {
    strncpy (input, ".8 .196078 .6\0", 14);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Violet", 6)) {
    strncpy (input, ".309804 .184314 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "VLightGrey", 10)) {
    strncpy (input, ".8 .8 .8\0", 9);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Wheat", 5)) {
    strncpy (input, ".847059 .847059 .74902\0", 23);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Silver", 6)) {
    strncpy (input, ".90 .91 .98\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "BrightGold", 10)) {
    strncpy (input, ".85 .85 .10\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "OldGold", 7)) {
    strncpy (input, ".81 .71 .23\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Feldspar", 8)) {
    strncpy (input, ".82 .57 .46\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Quartz", 6)) {
    strncpy (input, ".85 .85 .95\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Mica", 4)) {
    strncpy (input, "0 0 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "NeonPink", 8)) {
    strncpy (input, "1 .43 .78\0", 10);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DarkPurple", 10)) {
    strncpy (input, ".53 .12 .47\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "NeonBlue", 8)) {
    strncpy (input, ".3 .3 1\0", 8);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "CoolCopper", 10)) {
    strncpy (input, ".85 .53 .10\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MandarinOrange", 14)) {
    strncpy (input, ".89 .47 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "LightWood", 9)) {
    strncpy (input, ".91 .76 .65\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "MediumWood", 10)) {
    strncpy (input, ".65 .50 .39\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DarkWood", 8)) {
    strncpy (input, ".52 .37 .26\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "SpicyPink", 9)) {
    strncpy (input, "1 .11 .68\0", 10);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "SemiSweetChoc", 13)) {
    strncpy (input, ".42 .26 .15\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "BakersChoc", 10)) {
    strncpy (input, ".36 .20 .09\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Flesh", 5)) {
    strncpy (input, ".96 .80 .69\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "NewTan", 6)) {
    strncpy (input, ".92 .78 .62\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "NewMidnightBlue", 15)) {
    strncpy (input, "0 0 .61\0", 8);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "VeryDarkBrown", 13)) {
    strncpy (input, ".35 .16 .14\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DarkBrown", 9)) {
    strncpy (input, ".36 .25 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DarkTan", 7)) {
    strncpy (input, ".59 .41 .31\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "GreenCopper", 11)) {
    strncpy (input, ".32 .49 .46\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DkGreenCopper", 13)) {
    strncpy (input, ".29 .46 .43\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "DustyRose", 9)) {
    strncpy (input, ".52 .39 .39\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "HuntersGreen", 12)) {
    strncpy (input, ".13 .37 .31\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Scarlet", 7)) {
    strncpy (input, ".55 .09 .09\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Med_Purple", 10)) {
    strncpy (input, ".73 .16 .96\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "White", 5)) {
    strncpy (input, "1 1 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Very_Light_Purple", 17)) {
    strncpy (input, ".94 .81 .99\0", 12);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "YellowGreen", 11)) {
    strncpy (input, ".6 .8 .196078\0", 14);
    strcat (input, transp);
    return;
  }
  if (!strncmp (input, "Yellow", 6)) {
    strncpy (input, "1 1 0\0", 6);
    strcat (input, transp);
    return;
  }

/* try to find it in colors.inc, if povray options include +L path or if POV_Include is non-blank */
  colinc = NULL;
  if (strlen(drvui->POV_Include) > 10) {
      colinc = fopen(drvui->POV_Include,"r");
  }
  if (!colinc) {
      strncpy(b,drvui->POV_Options,29);
      b[29]='\0';
      a=strstr(b,"+L");
      if (a!=NULL) {
        sscanf(a+2,"%s",POV_incpath);
        if (POV_incpath!=NULL) {
          strcat(POV_incpath,"/colors.inc");
          colinc=fopen(POV_incpath,"r");
          if (colinc) {
            strcpy(drvui->POV_Include,POV_incpath);
            WriteConfig();
          }
        }
      }
  }
  if (colinc!=NULL){
      while (!feof(colinc)){
        if (fgets(line,80,colinc)==NULL)continue;
        sscanf(line,"%s %s = color red %f green %f blue %f",keyword,thecolor,
              &thered,&thegreen,&theblue); 
        if (strcmp(keyword,"#declare"))continue;
        if (!strcmp(thecolor,input)) {
          snprintf(input,30, "%3.2f %3.2f %3.2f%c",thered,thegreen,theblue,0);
          if (transp_start == NULL) {
            transp[0] = '\0';
            transp_start = strstr (line, "filter");
            if (transp_start != NULL) {
              fval = (float) atof (transp_start + 6);
              sprintf (transp, " transparency %5.2f", fval / 2.);
            }
          }
          strcat (input,transp);
          fclose(colinc);
          return;
        }
      }
      fclose(colinc);
  }
  
  if (Color_Warning <= 5) {
      char string[256];
      sprintf(string," **** Warning - Non-standard color \"%s\" encountered. ****",
         input);
      Error_Box(string);
      Color_Warning++;
  }

  if (sscanf(input," %f %f %f",&thered,&thegreen,&theblue) <3) {
    strcpy(input,"1. 1. 1.");    // substitute White if no color triplet
  } else {
    snprintf(input,30, "%3.2f %3.2f %3.2f%c",thered,thegreen,theblue,0);
    strcat (input, transp);
  }
}

/* ************************************************************** */
/* ************************************************************** */

float convert_pos(char * string)
{
// routine to convert an atom string to a float. If only digits, a simple sscanf
//  is done. If the position is of the form n/m, each part is decoded and the
// results of the division is returned.

    float result;
    int i, n, m;
    char * pdest;
    char sub_string1[10], sub_string2[10];

    if (!(pdest = strstr(string,"/"))) {
        sscanf(string,"%f",&result);             // no / in the input
    } else {
        i = pdest - string +1;                   // number of chars to copy
        strncpy(sub_string1, string, i);
        sub_string1[i - 1] = '\0';
        strcpy(sub_string2, pdest + 1);          // copy part after /
        sscanf(sub_string1,"%d", &n);
        sscanf(sub_string2,"%d", &m);
        result = (float)n/(float)m;
    }
    return result;
}

/* ************************************************************** */
/* ************************************************************** */

int end_flip(int value)
{
// routine to flip the endian nature of the input integer
// This routine was copied from a Web article:
//   http://www.codeproject.com/cpp/endianness.asp by Juan Carlos Cobas

   return (((value&0x000000FF)<<24)+((value&0x0000FF00)<<8)+
   ((value&0x00FF0000)>>8)+((value&0xFF000000)>>24));
}

/* ************************************************************** */
/* ************************************************************** */

float end_flip_real(float value)
{
// routine to flip the endian nature of the input float 
// This routine was copied from a Web article:
//   http://www.codeproject.com/cpp/endianness.asp by Juan Carlos Cobas

   union u {float vi; unsigned char c[4];}; 
   union v {float ni; unsigned char d[4];};
   union u un; 
   union v vn; 
   un.vi = value;
   vn.d[0]=un.c[3]; 
   vn.d[1]=un.c[2]; 
   vn.d[2]=un.c[1]; 
   vn.d[3]=un.c[0]; 
   return (vn.ni); 

}

/* ************************************************************** */
/* ************************************************************** */

void expand_atom (int natom)

/* routine to expand an atom position and obtain all equipoints within
   the unit cell */
/* natom - number of atom in list to expand */
{
  float xp[3], xpp[3];                 /* user in generating position */
  int i, j, k, l;                 /* loop variables */

  ncell = 0;                     /* clear list */
  for (i = 0; i < drvui->ng; ++i) {             /* symmetry operations */
      for (j = 0; j <= 2; ++j) {
          xp[j] = drvui->ts[i][j];
          for (k = 0; k <= 2; ++k)
              xp[j] = xp[j] + drvui->ss[i][j][k] * drvui->atom_xyz[natom][k];
      }
      for (k = 0; k < drvui->nlat; ++k) {         /*lattice points */
          for (l = 0; l <= 2; ++l)
              xpp[l] = xp[l] + drvui->lat_pos[k][l];
          add_to_list (xpp, i);
          if (!drvui->acentric) {
             for (l = 0; l <= 2; ++l)
                 xpp[l] = -xpp[l];
             add_to_list (xpp, i);
          }
      }
  }
}

/* ************************************************************** */
/* ************************************************************** */

void findsys(void) 
/* routine determine crystal system from symmetry information */
{
  int i,j,k;
  int n,kk=0,nn,symlist[23];
  int yzx=0,zxy=0,ymxmz=0,mxmyz=0,myxmz=0;

  if (drvui->ng==1) { /* only x,y,z => triclinic */
    drvui->sys = 1;
    return;
  }

  if (drvui->ng==2) { /* only two symops => monoclinic */
    drvui->sys = 2;
    return;
  }

  drvui->sys = 3; /* orthorhombic unless we find specific operators below */

  for (i = 0; i < drvui->ng; ++i) {
    for (j = 0; j <= 2; ++j) {
      if(fabs(drvui->ts[i][j]-0.33333333) <0.00001) drvui->sys=5;
      if(fabs(drvui->ts[i][j]-0.16666667) <0.00001) drvui->sys=5;
      if(fabs(drvui->ts[i][j]-0.66666667) <0.00001) drvui->sys=5;
      if(fabs(drvui->ts[i][j]-0.83333333) <0.00001) drvui->sys=5;
    }
    if (drvui->sys==5) return;
  }

  if (fabs(drvui->lat_con[5] - 120.0) < 0.00001
      && fabs(drvui->lat_con[4] - 90.0) <0.0001) {
    drvui->sys = 5;
    return;
  }

  n=0;
  for (i = 0; i < drvui->ng; ++i) {
    for (j = 0; j <= 2; ++j) {
       if (drvui->ts[i][j] != 0.0) break;
       kk=0;
       for (k = 0; k <= 2; ++k)
         if (drvui->ss[i][j][k] != 0) kk++;
    }
    if (kk >1) {
      drvui->sys = 5;
      return;
    }
    symlist[n++]=i;
  }

  for (nn=0; nn < n; nn++) {
    i=symlist[nn];
    if (drvui->ss[i][0][1]==1 && 
        drvui->ss[i][1][2]==1 &&
        drvui->ss[i][2][0]==1) yzx=1;
    if (drvui->ss[i][0][2]==1 && 
        drvui->ss[i][1][0]==1 &&
        drvui->ss[i][2][1]==1) zxy=1;
    if (drvui->ss[i][0][1]==1 && 
        drvui->ss[i][1][0]==-1 &&
        drvui->ss[i][2][2]==-1) ymxmz=1;
    if (drvui->ss[i][0][0]==-1 && 
        drvui->ss[i][1][1]==-1 &&
        drvui->ss[i][2][2]==1) mxmyz=1;
    if (drvui->ss[i][0][1]==-1 && 
        drvui->ss[i][1][0]==1 &&
        drvui->ss[i][2][2]==-1) myxmz=1;
  }

  if (ymxmz+mxmyz+myxmz==3) drvui->sys=4; /*tetragonal*/
  if (yzx+zxy==2) drvui->sys=6; /*cubic*/
}

/* ************************************************************** */
/* ************************************************************** */

void get_label (char input[], char *c1, char *c2, char *c3, char *c4,
        int strip)

// Routine to get 1 to 4 character atom ID from input line, and shift input
//   line to left, if strip true, skip over first token
{

  int i, j;

  i = 0;
  if (strip)
    while (input[i] != ' ')
      input[i++] = ' ';            // skip over command
  while (input[i] == ' ')
    i++;                           // skip whitespace after command
  *c1 = input[i];                  // return first atom label character
  input[i++] = ' ';
  *c2 = input[i];                  // return 2nd atom label char
  if (input[i] != ' ')
    input[i++] = ' ';              // if 2nd not blank, rub it out
  *c3 = input[i];                  // return 3rd character
  if (input[i] != ' ')
    input[i++] = ' ';              // if 3rd not blank, rub it out
  *c4 = input[i];                  // return 4th character
  if (input[i] != ' ')
    input[i++] = ' ';              // if 4th not blank, rub it out
  while (input[i] == ' ')
    i++;                          // skip white space after atom label
  for (j = 0; i < (int)strlen(input); i++) {
    input[j++] = input[i];        // copy rest of command
  }
  input[j] = 0;                   // put NULL at end
}                                 // end of get_label

/* ************************************************************** */
/* ************************************************************** */

void getsym(char* text, int num, int kk)

/* helper routine to parse SHELX symmetry operator strings */
{ 
    int i,j;
    char *txt;

    txt = (char*)malloc(24*sizeof(char));
 
    j=0;
    for (i=0;i<23;i++)
      if (text[i] != ' ' && text[i] != '\n')
        txt[j++]=toupper(text[i]);

    txt[j]='\0';

    drvui->ts[num][kk] = 0.;

    char *n = strstr(txt,"/");
    if (n != NULL) {
    n++;
    char *d = n-3;
    if (atof(d)==0.) d=n-2;
        drvui->ts[num][kk] = (float)(atof(d) / atof(n));
    }

    n= strstr(txt,".");
    if (n != NULL) {
        drvui->ts[num][kk] = (float)atof(n);
        if (n!=txt) {
            if (txt[&txt-&n+1] == '-' )
                drvui->ts[num][kk] *= -1;
        }
    }

    drvui->ss[num][kk][0]=0;
    drvui->ss[num][kk][1]=0;
    drvui->ss[num][kk][2]=0;

    for (i=0;i<(int)strlen(txt);i++) {
        if (txt[i]=='X') {
            drvui->ss[num][kk][0]=1;
            if (i > 0) {
                if (txt[i-1]=='-') drvui->ss[num][kk][0]=-1;
            }
        }
        if (txt[i]=='Y') {
             drvui->ss[num][kk][1]=1;
             if (i > 0) {
                 if (txt[i-1]=='-') drvui->ss[num][kk][1]=-1;
             }
        }     
        if (txt[i]=='Z') {
            drvui->ss[num][kk][2]=1;
            if (i >0) {
                if (txt[i-1]=='-') drvui->ss[num][kk][2]=-1;
            }
        }
    }
    free(txt);
}

/* ************************************************************** */
/* ************************************************************** */

void import_cif (char input[100], int in_line, int Quick)

/* routine to extract structural information from a CIF */
{
  FILE *impin;
  char string[100], filename[256], tstring[100];
  int U_pos[6], X_pos[3], Label_pos = 0;
  int i, j, sg;

  if (in_line) {
      Error_Box("Cannot read inline CIF data. Please use the import instruction");
      return;
  }
  (void)sscanf (input, "%s %s", filename, filename);     /* get file name */

  if (!(impin = fopen (filename, "r"))) {
      Error_Box("Cannot Open CIF Import File, Run aborted.");
      return;
  }
  
  memset(string,0,100);
  memset(tstring,0,100);

  while (strncmp (string, "_cell_length_a", 14) != 0) {
    if (!fgets (string, 100, impin)) {         /* search for a-axis length */
        Error_Box("Error reading CIF Import File, Run aborted.");
        return;
    }
    Blank_Strip (string);
  }
  (void)sscanf (string, "%s %f", tstring, &drvui->lat_con[0]);
  rewind (impin);
  while (strncmp (string, "_cell_length_b", 14) != 0) {
    if (!fgets (string, 100, impin)) {         /* search for b-axis */
        Error_Box("Error reading CIF Import File, Run aborted.");
        return;
    }
    Blank_Strip (string);
  }
  (void)sscanf (string, "%s %f", tstring, &drvui->lat_con[1]);
  rewind (impin);
  while (strncmp (string, "_cell_length_c", 14) != 0) {
    if (!fgets (string, 100, impin)) {         /* search for c axis */
        Error_Box("Error reading CIF Import File, Run aborted.");
        return;
    }
    Blank_Strip (string);
  }
  (void)sscanf (string, "%s %f", tstring, &drvui->lat_con[2]);
  rewind (impin);
  while (strncmp (string, "_cell_angle_al", 14) != 0) {
    if (!fgets (string, 100, impin)) {         /* search for alpha */
        Error_Box("Error reading CIF Import File, Run aborted.");
        return;
    }
    Blank_Strip (string);
  }
  rewind (impin);
  (void)sscanf (string, "%s %f", tstring, &drvui->lat_con[3]);
  while (strncmp (string, "_cell_angle_be", 14) != 0) {
    if (!fgets (string, 100, impin)) {         /* search for beta */
        Error_Box("Error reading CIF Import File, Run aborted.");
        return;
    }
    Blank_Strip (string);
  }
  rewind (impin);
  (void)sscanf (string, "%s %f", tstring, &drvui->lat_con[4]);
  while (strncmp (string, "_cell_angle_ga", 14) != 0) {
    if (!fgets (string, 100, impin)) {         /* search for gamma */
        Error_Box("Error reading CIF Import File, Run aborted.");
        return;
    }
    Blank_Strip (string);
  }
  rewind (impin);
  j = 1;
  (void)sscanf (string, "%s %f", tstring, &drvui->lat_con[5]);
  if (!Quick)
    fprintf (drvui->flout,
         "******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n",
         drvui->lat_con[0], drvui->lat_con[1], drvui->lat_con[2], drvui->lat_con[3], drvui->lat_con[4],
         drvui->lat_con[5]);

  sg = 0;

  while (strncmp (string, "_symmetry_space", 15) != 0) {
    if (!fgets (string, 100, impin)) {;         /* search for space group */
      j = 0;                     /* indicate not found */
      break;                     /* no error if not found */
    }
    Blank_Strip (string);
  }
  if (j != 0) {
    tstring[0] = 's';
    tstring[1] = 'p';
    tstring[2] = 'g';
    tstring[3] = 'r';
    tstring[4] = ' ';
    for (i = 0; i < 60; i++) {             /* change the keyword to spaces */
      if (string[i] == ' ') {
    break;
      }
      else {
    string[i] = ' ';
      }
    }
    Blank_Strip (string);             /* strip off the spaces */
    if (string[0] != '?') {
      for (i = 0; i < 60; i++)
    if (string[i] == (char) 39)
      break;
      for (i++, j = 5; i < 60; i++)
    if (string[i] == (char) 39)
      break;
    else
      tstring[j++] = string[i];
      tstring[j] = '\0';
      if (!Quick)
    fprintf (drvui->flout, "******      %s\n", tstring);
      for (i = 0; i < 60; i++) {
    if (tstring[i] == 'H')
      tstring[i] = ' ';             /* remove any H from rhomb. spgrps */
    if (tstring[i] == 'S')
      tstring[i] = ' ';             /* remove any S from monoclinic spgrps */
      }
      symop (tstring);
      sg = 1;
    } else {
      if (!Quick)
        fprintf (drvui->flout, "******      this CIF does not contain a space group name\n");
    }
  } else {
    if (!Quick)
      fprintf (drvui->flout, "******      this CIF does not contain a space group record\n");
  }
    
  if (sg == 0) {
    rewind (impin);
    string[0]='\0';
    j = 1;
    while (strncmp (string, " _symmetry_equiv_pos_as_xyz", 27) != 0) {
      if (!fgets (string, 100, impin)) {;         /* search for space group */
        j = 0;                     /* indicate not found */
        break;                     /* no error if not found */
      }
    }
    if ( j == 0) {
      if (!Quick)
        fprintf (drvui->flout, "******      this CIF contains neither a space group nor a list of symmetry operations\n");
    } else {
        char *p,*pp;
        int k;
        drvui->acentric=1;
    /* initialize symmetry arrays */
    for (i = 0; i <= 23; ++i) {
      for (j = 0; j <= 2; ++j) {
        drvui->ts[i][j] = 0.0;
        for (k = 0; k <= 2; ++k)
          drvui->ss[i][j][k] = 0;
      }
    }
    for (i = 0; i <= 2; ++i) {
      drvui->lat_pos[0][i] = 0.0;
      drvui->ss[0][i][i] = 1;
      for (j = 0; j <= 3; ++j)
        drvui->spg[i][j] = ' ';
    }
    drvui->ng=0;

        while (strlen(string)>2) {
          fgets(string,100,impin);
          for (j=0;j<(int)strlen(string);j++) 
            if (string[j]==(unsigned char)39) string[j]=' ';
          Blank_Strip (string);
          if (strstr(string,"-x, -y, -z")) {
            drvui->acentric=0;
            break;
          }
          drvui->ng++;
          p = strstr(string,",");
          *p='\0';
          getsym(string,drvui->ng-1,0);
          *p=' ';
          p++;
          pp = strstr(p,",");
          *pp='\0';
          getsym(p,drvui->ng-1,1);
          pp++;
          getsym(pp,drvui->ng-1,2);
        }
        findsys();

        drvui->nbr=1;

        for (i = 1; i <= drvui->ng; ++i) {
          if (drvui->ss[i][0][2]==1 && drvui->ss[i][1][0]==1) {
		drvui->nbr=5; // Z,X,Y ==> F
		break;
          }
	          
            if (drvui->nbr<4 && drvui->ts[i-1][0]==0. && 
        	      drvui->ts[i-1][1]==0.5 &&
	      	      drvui->ts[i-1][2]==0.5) 
            		drvui->nbr=4;  // X,0.5+Y,0.5+Z ==> C

            if (drvui->nbr <3 && drvui->ts[i-1][0]==0.5 && 
        	      drvui->ts[i-1][1]==0. &&
	      	      drvui->ts[i-1][2]==0.5) 
            		drvui->nbr=3;  // 0.5+X,+Y,0.5+Z ==> B
			
          if (drvui->nbr==1 && drvui->ts[i-1][0]==0.5 && 
              drvui->ts[i-1][1]==0.5 &&
	      drvui->ts[i-1][2]==0.) 
            drvui->nbr=2;  // 0.5+X,0.5+Y,Z ==> A

          if (drvui->ts[i-1][0]==0.5 && 
              drvui->ts[i-1][1]==0.5 &&
	      drvui->ts[i-1][2]==0.5) 
            drvui->nbr=6;  // 0.5+X,0.5+Y,0.5+Z ==> I
			
          if (drvui->ts[i-1][0]<0.5 && 
              drvui->ts[i-1][0]>0.25 )
            drvui->nbr=7;  // 0.33+X ==> R
        }
        switch (drvui->nbr) {
          case 1:
            drvui->nlat = 1;
            break;
          case 2:
            drvui->nlat = 2;
            drvui->lat_pos[1][0] = 0.0;
            drvui->lat_pos[1][1] = 0.5f;
            drvui->lat_pos[1][2] = 0.5f;
            break;
          case 3:
            drvui->nlat = 2;
            drvui->lat_pos[1][0] = 0.5f;
            drvui->lat_pos[1][1] = 0.0;
            drvui->lat_pos[1][2] = 0.5f;
            break;
          case 4:
            drvui->nlat = 2;
            drvui->lat_pos[1][0] = 0.5f;
            drvui->lat_pos[1][1] = 0.5f;
            drvui->lat_pos[1][2] = 0.0;
            break;
          case 5:
            drvui->nlat = 4;
            drvui->lat_pos[1][0] = 0.0;
            drvui->lat_pos[1][1] = 0.5f;
            drvui->lat_pos[1][2] = 0.5f;
            drvui->lat_pos[2][0] = 0.5f;
            drvui->lat_pos[2][1] = 0.0;
            drvui->lat_pos[2][2] = 0.5f;
            drvui->lat_pos[3][0] = 0.5f;
            drvui->lat_pos[3][1] = 0.5f;
            drvui->lat_pos[3][2] = 0.0;
            break;
          case 6:
            drvui->nlat = 2;
            drvui->lat_pos[1][0] = 0.5f;
            drvui->lat_pos[1][1] = 0.5f;
            drvui->lat_pos[1][2] = 0.5f;
            break;
          case 7:
            drvui->nlat = 3;
            drvui->lat_pos[1][0] = (1.f / 3.f);
            drvui->lat_pos[1][1] = (2.f / 3.f);
            drvui->lat_pos[1][2] = (2.f / 3.f);
            drvui->lat_pos[2][0] = (2.f / 3.f);
            drvui->lat_pos[2][1] = (1.f / 3.f);
            drvui->lat_pos[2][2] = (1.f / 3.f);
          }

      }
  }

    
  rewind (impin);
/* skip through file to atom_site */
  while (strncmp (string, "_atom_site_", 11) != 0) {
    if (!fgets (string, 100, impin)) {         /* search for phase info */
        Error_Box("Error reading CIF Import File, Run aborted.");
        return;
    }
    Blank_Strip (string);
  }
/* first _atom_site_ line in string need to find position of label, x, y, and z */
  j = 0;
  while (!strncmp (string, "_atom_site_", 11)) {
    if (!strncmp (string, "_atom_site_label", 16))
      Label_pos = j;
    if (!strncmp (string, "_atom_site_fract_x", 18))
      X_pos[0] = j;
    if (!strncmp (string, "_atom_site_fract_y", 18))
      X_pos[1] = j;
    if (!strncmp (string, "_atom_site_fract_z", 18))
      X_pos[2] = j;
    j++;
    if (!fgets (string, 100, impin)) {         /* search for phase info */
        Error_Box("Error reading CIF Import File, Run aborted.");
        return;
    }
    Blank_Strip (string);
  }
  for (;;) {
    if (strlen (string) < 5)
      break;
    if (!strncmp (string, "loop_", 5))
      break;
    drvui->atom_n[natom] = 0;
    for (i = 0; i < 4; i++)
      drvui->atom_l[natom][i] = ' ';             /* initialize atom name */
    strcpy (tstring, string);
    if (Label_pos > 0)
      Token_Strip (tstring, Label_pos);
    j = 0;
    for (i = 0; i < 10; i++) {
      if (tstring[i] >= '0' && tstring[i] <= '9') {
        drvui->atom_n[natom] = 10 * drvui->atom_n[natom] + (int) tstring[i] - 48;
        drvui->atom_l[natom][i] = ' ';
      } else {
        drvui->atom_l[natom][j++] = tstring[i];
        if (j > 3 || tstring[i] == ' ')
          break;
      }
    }
    strcpy (tstring, string);
    if (X_pos[0] > 0)
      Token_Strip (tstring, X_pos[0]);
    (void)sscanf (tstring, "%f", &drvui->atom_xyz[natom][0]);
    strcpy (tstring, string);
    if (X_pos[1] > 0)
      Token_Strip (tstring, X_pos[1]);
    (void)sscanf (tstring, "%f", &drvui->atom_xyz[natom][1]);
    strcpy (tstring, string);
    if (X_pos[2] > 0)
      Token_Strip (tstring, X_pos[2]);
    (void)sscanf (tstring, "%f", &drvui->atom_xyz[natom][2]);
    if (!Quick)
      fprintf (drvui->flout,
           "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
           drvui->atom_l[natom][0]
           , drvui->atom_l[natom][1], drvui->atom_l[natom][2], drvui->atom_l[natom][3],
           drvui->atom_n[natom]
           , drvui->atom_xyz[natom][0], drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2]);
    for (i = 0; i < 100; i++)
      string[i] = ' ';
    if (!fgets (string, 100, impin)) {
        Error_Box("Error reading CIF Import File, Run aborted.");
        return;
    }
    Blank_Strip (string);
    drvui->atom_fn[natom] = 1;
    if (++natom >= MAX_ATOM) {
        Error_Box("Too many atoms in asymmetric unit - increase MAX_ATOM.");
        return;
    }
  }
  rewind (impin);
/* skip through file to atom_site_aniso */
  while (strncmp (string, "_atom_site_aniso", 16) != 0) {
    if (!fgets (string, 100, impin)) {         /* search for anisotropic U's */
      (void)fclose (impin);
      return;
    }
    Blank_Strip (string);
  }
/* first _atom_site_aniso line in string
   need to find position of label, U11, etc */
  j = 0;
  while (!strncmp (string, "_atom_site_aniso", 16)) {
    if (!strncmp (string, "_atom_site_aniso_label", 22))
      Label_pos = j;
    if (!strncmp (string, "_atom_site_aniso_U_11", 21))
      U_pos[0] = j;
    if (!strncmp (string, "_atom_site_aniso_U_22", 21))
      U_pos[1] = j;
    if (!strncmp (string, "_atom_site_aniso_U_33", 21))
      U_pos[2] = j;
    if (!strncmp (string, "_atom_site_aniso_U_12", 21))
      U_pos[3] = j;
    if (!strncmp (string, "_atom_site_aniso_U_13", 21))
      U_pos[4] = j;
    if (!strncmp (string, "_atom_site_aniso_U_23", 21))
      U_pos[5] = j;
    j++;
    if (!fgets (string, 100, impin)) {         /* search for phase info */
        Error_Box("Error reading CIF Import File, Run aborted.");
        return;
    }
    Blank_Strip (string);
  }
  for (;;) {
    if (strlen (string) < 5)
      break;
    if (!strncmp (string, "_", 1))
      break;
    drvui->ellips_n[n_ellips] = 0;
    for (i = 0; i < 4; i++)
      drvui->ellips_l[n_ellips][i] = ' ';         /* initialize ellipsoid name */
    strcpy (tstring, string);
    if (Label_pos > 0)
      Token_Strip (tstring, Label_pos);
    j = 0;
    for (i = 0; i < 10; i++) {
      if (tstring[i] >= '0' && tstring[i] <= '9') {
        drvui->ellips_n[n_ellips] = 10 * drvui->ellips_n[n_ellips] + (int) tstring[i] - 48;
        drvui->ellips_l[n_ellips][i] = ' ';
      }
      else {
        drvui->ellips_l[n_ellips][j++] = tstring[i];
        if (j > 3 || tstring[i] == ' ')
          break;
      }
    }
    drvui->save_el_number[n_ellips] = drvui->ellips_n[n_ellips];
    do_ellipsoids = 1;
    strcpy (tstring, string);
    if (U_pos[0] > 0)
      Token_Strip (tstring, U_pos[0]);
    (void)sscanf (tstring, "%f", &drvui->ellips[n_ellips][0]);
    strcpy (tstring, string);
    if (U_pos[1] > 0)
      Token_Strip (tstring, U_pos[1]);
    (void)sscanf (tstring, "%f", &drvui->ellips[n_ellips][1]);
    strcpy (tstring, string);
    if (U_pos[2] > 0)
      Token_Strip (tstring, U_pos[2]);
    (void)sscanf (tstring, "%f", &drvui->ellips[n_ellips][2]);
    strcpy (tstring, string);
    if (U_pos[3] > 0)
      Token_Strip (tstring, U_pos[3]);
    (void)sscanf (tstring, "%f", &drvui->ellips[n_ellips][3]);
    strcpy (tstring, string);
    if (U_pos[4] > 0)
      Token_Strip (tstring, U_pos[4]);
    (void)sscanf (tstring, "%f", &drvui->ellips[n_ellips][4]);
    strcpy (tstring, string);
    if (U_pos[5] > 0)
      Token_Strip (tstring, U_pos[5]);
    (void)sscanf (tstring, "%f", &drvui->ellips[n_ellips][5]);
    if (drvui->auto_ellipse) {
        drvui->ell_type[n_ellips] = 1001;         // Uij with ellipcolor
        memset(drvui->ellips_col_p[n_ellips],0,40);
        memset(drvui->ellips_col_v[n_ellips],0,40);
        strcpy(drvui->ellips_col_p[n_ellips],"Gray30");
        strcpy(drvui->ellips_col_v[n_ellips],"0.3 0.3 0.3");
    } else {
        drvui->ell_type[n_ellips] = 1;            // Uij
    }
    if(!Quick)fprintf (drvui->flout,
         "******      Uij  %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %8.5f %8.5f\n",
         drvui->ellips_l[n_ellips][0], drvui->ellips_l[n_ellips][1],
         drvui->ellips_l[n_ellips][2], drvui->ellips_l[n_ellips][3],
         drvui->ellips_n[n_ellips], drvui->ellips[n_ellips][0], drvui->ellips[n_ellips][1], 
         drvui->ellips[n_ellips][2], drvui->ellips[n_ellips][3], drvui->ellips[n_ellips][4],
         drvui->ellips[n_ellips][5]);
    if (++n_ellips >= MAX_EL) {
        Error_Box("Too many ellipsoid types specified - increase MAX_EL.");
        return;
    }
    if (!fgets (string, 100, impin)) {
        Error_Box("Error reading CIF Import File, Run aborted.");
        return;
    }
    Blank_Strip (string);
  }
  drvui->auto_ellipse = 0;
  (void)fclose (impin);
}

/* ************************************************************** */
/* ************************************************************** */

void import_fdat (char input[100], int in_line, int Quick)

/* routine to extract structural information from a CSD 'FDAT' data file */
{
  FILE *impin;
  char string[200];                 /* temporary string */
  char filename[256];                 /* string for filename */
  int i, j;                     /* temporary */
  int k, l, k3, k4;                 /*flags for space group symbol parsing */
  char spstring[100];                 /* space group symbol */
  char lat_s[6][7];                 /*temporary storage for lattice constants */
  int ldec[6];                     /*temporary storage for decimal places of lattice constants */
  char number[4];
  char ncons[4], nrfacs[4], nrems[4], ndiss[4], nerrs[4], noprs[4],
    nrads[4], nats[4];
  char sats[4], nbnds[4];
  int ncards, nrfac = 0, nrem = 0, ndis = 0, nerr = 0, nopr = 0, nrad =
    0, nat = 0, sat = 0, ncon = 0, cellflag;
  int intflag, coordflag, centflag, errflag = 0;
  int nskip;                     /* lines to skip between lattice constants and atoms */
  int natomin;

  memset (spstring, 0, 100);
  memset (string, 0, 100);
  memset (nrfacs, 0, 4);
  memset (nrems, 0, 4);
  memset (ndiss, 0, 4);
  memset (ncons, 0, 4);
  memset (number, 0, 4);
  memset (nerrs, 0, 4);
  memset (noprs, 0, 4);
  memset (nrads, 0, 4);
  memset (nats, 0, 4);
  memset (sats, 0, 4);
  memset (nbnds, 0, 4);
  for (i = 0; i < 6; i++)
    memset (lat_s[i], 0, 7);
  if (in_line) {
    impin = drvui->fpin;
  } else {
    (void)sscanf (input, "%s %s", filename, filename); /* get file name */
    if (!(impin = fopen (filename, "r"))) {
        Error_Box("Cannot Open CSD/FDAT Import File, Run aborted.");
        return;
    }
  }

  if(!Quick)fprintf (drvui->flout, "****** Now interpreting CSD-style data\n");

  if (!fgets (input, 100, impin)) {         /* read a line */
      Error_Box("Error reading CSD/FDAT file.");
      return;
  }

/* the first line should contain the REFCODE of the database entry */
/* and information about the quality of the data and the length of */
/* the following sections                                          */

  if (strncmp (input, "#", 1) == 0) {

    (void)sscanf (input, "%23c%3d%3c%3c%3c%3c%3c%3c%3c%3c%3c%3c%1d%1d%1d%1d%1d",
        string, &ncards, nrfacs, nrems, ndiss, nerrs, noprs, nrads,
        nats, sats, nbnds, ncons, &cellflag, &intflag, &coordflag,
        &centflag, &errflag);

    nrfac = atoi (nrfacs);
    ncon = atoi (ncons);
    nrem = atoi (nrems);
    ndis = atoi (ndiss);
    nerr = atoi (nerrs);
    nopr = atoi (noprs);
    nrad = atoi (nrads);
    nat = atoi (nats);
    sat = atoi (sats);
    input[0] = ' ';
    input[9] = '\0';                 /* truncate title */
    if (!Quick) fprintf (drvui->flout, "******      %s\n", input);
    if (!cellflag) {
        Error_Box("Error - this CSD entry has neither CELL nor COORDINATES");
        return;
    }

    if (!coordflag) {
        Error_Box("Error - this CSD entry has no COORDINATES");
        return;
    }

    if (errflag) {
        Error_Box("Warning - this CSD entry has the ERROR flag set -\n at least some coordinates may be wrong");
    }

    if (!fgets (input, 100, impin)) {         /* read a line */
        Error_Box("Error reading CSD file.");
        return;
    }
  }

/* lattice constants come in a fixed FORTRAN format followed by integers */
/* encoding precision, esd's, etc and finally by the spacegroup symbol */

  for (i = 0; i <= 5; ++i)
    drvui->lat_con[i] = 0.0f;

  (void)sscanf (input, "%6c%6c%6c%6c%6c%6c%1d%1d%1d%1d%1d%1d%21c%s", lat_s[0],
      lat_s[1], lat_s[2], lat_s[3], lat_s[4], lat_s[5], &ldec[0],
      &ldec[1], &ldec[2], &ldec[3], &ldec[4], &ldec[5], string, string);

  for (i = 0; i < 6; ++i) {
    drvui->lat_con[i] = (float) (atof (lat_s[i]) / pow (10, (double)ldec[i]));
  }

/* separate class symbol from other symmetry elements */
  spstring[0] = 's';
  spstring[1] = 'p';
  spstring[2] = 'g';
  spstring[3] = 'r';
  spstring[4] = ' ';
  spstring[5] = string[0];
  spstring[6] = ' ';
  j = 6;
  k = 0;
  k3 = 0;
  k4 = 0;
  l = (int) strlen (string);
  for (i = 1; i <= l; i++) {
    switch (string[i]) {
    case ' ':
      break;
    case '/':
      spstring[++j] = string[i];
      break;
    default:
      j++;
      /* do we have to collate this symbol (for screw axes or mirror planes) ? */
      if (string[i] > '0' && string[i] <= '9') { /* it is a number */
    if (string[0] == 'R' ||             /* no screw axes in R */
        l == 4 ||                 /* no screw axes in four letter names */
        k != 0 ||                 /* have just appended to a number */
        k3 != 0 ||                 /* cannot have further screw axes after 3_x */
        (k4 != 0 && spstring[j - 1] == '3')) {    /* cannot have 3_x after a 4_x */
      spstring[j] = ' ';             /* so separate it */
      k = 0;                 /* and reset the collation flag */
      j++;
    }
    else if (spstring[j - 1] != '-' && (spstring[j - 1] <= string[i]
                        || spstring[j - 1] >= 'a')) {
      /* neither a minus sign nor a higher number precedes it */
      spstring[j] = ' ';             /* so separate it */
      k = 0;                 /* and reset the collation flag */
      j++;
    }
    else {
      k++;                     /* set flag for collation */
      if (spstring[j - 1] == '3')
        k3 = 1;                 /* if 3_x screw, prohibit further screw axes */
      if (spstring[j - 1] == '4')
        k4 = 1;                 /* if 4_x screw, prohibit 3_x screw axes */
    }
      }
      else {                     /* not a number */
    if (spstring[j - 1] != '/') {         /*but not a mirror plane either */
      spstring[j] = ' ';             /* separate it */
      j++;
    }
      }
      spstring[j] = string[i];             /* now add the symbol */
      break;
    }                         /*switch */
  }                         /*for i */
  spstring[j + 1] = '\0';

/* Space group 90, P 4 21 2, is not correctly identified by this code. 
   It accounts for only 14 of the 175000 entries in the CSD (and not all 
   of them have coordinates), so this quick and dirty fix is all it can
   expect from us. Does this spacegroup really exist ? :-) */
  if (!strcmp (string, "P4212"))
    strcpy (spstring, "P 4 21 2");

  if(!Quick)fprintf (drvui->flout, "******      Space group *%s*\n", spstring);
  symop (spstring);

  if(!Quick)fprintf (drvui->flout,
       "******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n",
       drvui->lat_con[0], drvui->lat_con[1], drvui->lat_con[2], drvui->lat_con[3], drvui->lat_con[4],
       drvui->lat_con[5]);

/*skip R value, any textual information, symmetry codes and radii data */

  nskip = (nrfac + nrem + ndis + nerr) / 80 + 1; /* skip comments (80chars per line) */
  nskip += nopr / 5 + 1;             /* skip symmetry operators (five per line) */
  nskip += nrad / 16 + 1;             /* skip atomic radii (sixteen per line) */


  for (i = 0; i < nskip; i++) {
    if (!fgets (input, 100, impin)) {         /* read a line */
        Error_Box("Error reading file while skipping records.");
        if (!Quick) {
          fprintf (drvui->flout,
           "nrfac:%d, nrem:%d, ndis:%d,nerr:%d, nopr:%d, nrad:%d, nskip:%d\n",
           nrfac, nrem, ndis, nerr, nopr, nrad, nskip);
          return;
        }
    }
  }

  natomin = natom;                 /* number of atoms already in list */

/* read coordinates until empty atom symbol encountered */

  for (i = 0; i < (nat + sat) / 3 + 1; i++) {     /*for all expected atom records */

    if (!fgets (input, 100, impin)) {         /* read a line */
        Error_Box("Error reading CSD file while reading atom coordinates.");
        return;
    }

/* atoms come in groups of three */
    (void)sscanf (input,
        "%c%c%c%c %7f %7f %7f %c%c%c%c %7f %7f %7f %c%c%c%c %7f %7f %7f",
        &drvui->atom_l[natom][0], &drvui->atom_l[natom][1], &drvui->atom_l[natom][2],
        &drvui->atom_l[natom][3], &drvui->atom_xyz[natom][0], &drvui->atom_xyz[natom][1],
        &drvui->atom_xyz[natom][2], &drvui->atom_l[natom + 1][0],
        &drvui->atom_l[natom + 1][1], &drvui->atom_l[natom + 1][2],
        &drvui->atom_l[natom + 1][3], &drvui->atom_xyz[natom + 1][0],
        &drvui->atom_xyz[natom + 1][1], &drvui->atom_xyz[natom + 1][2],
        &drvui->atom_l[natom + 2][0], &drvui->atom_l[natom + 2][1],
        &drvui->atom_l[natom + 2][2], &drvui->atom_l[natom + 2][3],
        &drvui->atom_xyz[natom + 2][0], &drvui->atom_xyz[natom + 2][1],
        &drvui->atom_xyz[natom + 2][2]);

/* and the atom list is terminated by numerical data */
    if (drvui->atom_l[natom][0] < 'A') {
      break;
    }

/* scale coordinates */
    for (j = 0; j < 3; j++) {
      drvui->atom_xyz[natom][j] = drvui->atom_xyz[natom][j] / 100000.0f;
      drvui->atom_xyz[natom + 1][j] = drvui->atom_xyz[natom + 1][j] / 100000.0f;
      drvui->atom_xyz[natom + 2][j] = drvui->atom_xyz[natom + 2][j] / 100000.0f;
    }

/* separate numbers from labels and remove trailing identifiers */

    for (l = 0; l < 3; l++) {
      k = 0;
      for (j = 1; j < 4; j++) {
    number[k] = ' ';
    if (drvui->atom_l[natom + l][j] >= '0' && drvui->atom_l[natom + l][j] <= '9') {
      number[k] = drvui->atom_l[natom + l][j];
      k++;
    }
    if (k)
      drvui->atom_l[natom + l][j] = ' ';
      }
      number[k] = '\0';
      drvui->atom_n[natom + l] = atoi (number);
    }

    if (Unique_Atom () == 1) {
/* report converted data of first atom */
      if (!Quick)fprintf (drvui->flout,
           "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
           drvui->atom_l[natom][0], drvui->atom_l[natom][1], drvui->atom_l[natom][2],
           drvui->atom_l[natom][3]
           , drvui->atom_n[natom], drvui->atom_xyz[natom][0], drvui->atom_xyz[natom][1],
           drvui->atom_xyz[natom][2]);
      drvui->atom_fn[natom] = 1;
      natom++;
      if (natom >= MAX_ATOM) {             /* first unique */
          Error_Box("Too many atoms in asymmetric unit - increase MAX_ATOM.");
          return;
      }
    }                         /* first unique */
/* return now if no second atom on the line */
    if (natom == nat + natomin) {
      break;                     /* got all expected atoms */
    }
    if (drvui->atom_l[natom][0] == '\0')
      break;
    if (Unique_Atom () == 1) {
/* report converted data of second atom */
      if (!Quick)fprintf (drvui->flout,
           "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
           drvui->atom_l[natom][0], drvui->atom_l[natom][1], drvui->atom_l[natom][2],
           drvui->atom_l[natom][3]
           , drvui->atom_n[natom], drvui->atom_xyz[natom][0], drvui->atom_xyz[natom][1],
           drvui->atom_xyz[natom][2]);
      drvui->atom_fn[natom] = 1;
      natom++;
      if (natom >= MAX_ATOM) {             /* second unique */
          Error_Box("Too many atoms in asymmetric unit - increase MAX_ATOM.");
          return;
      }
    }                         /* second unique */
/* return now if no third atom on the line */
    if (natom == nat + natomin) {
      break;                     /* got all expected atoms */
    }
    if (drvui->atom_l[natom][0] == '\0')
      break;
    if (Unique_Atom () == 1) {             /* third unique */
/* report converted data of third atom */
      if (!Quick)fprintf (drvui->flout,
           "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
           drvui->atom_l[natom][0], drvui->atom_l[natom][1], drvui->atom_l[natom][2],
           drvui->atom_l[natom][3]
           , drvui->atom_n[natom], drvui->atom_xyz[natom][0], drvui->atom_xyz[natom][1],
           drvui->atom_xyz[natom][2]);
      drvui->atom_fn[natom] = 1;
      natom++;
      if (natom >= MAX_ATOM) {
          Error_Box("Too many atoms in asymmetric unit - increase MAX_ATOM.");
          return;
      }
      if (natom == nat + natomin) {
    break;                     /* got all expected atoms */
      }
    }                         /* third unique */
  }                         /*for all expected atom records */
  for (i++; i < (nat + sat) / 3; i++) {         /* skip remainder of atom records */

    if (!fgets (input, 100, impin)) {         /* read a line */
        Error_Box("Error reading file while skipping extra atom records.");
        return;
    }
  }
  if (!errflag) {                 /* only error-free entries have connectivity records */
    nskip = ncon / 40 + 1;
    if (nat + sat > 100)
      nskip = ncon / 26 + 1;
    for (i = 0; i < nskip; i++) {
      if (!fgets (input, 100, impin)) {         /* read a line */
    if (!in_line)
      (void)fclose (impin);
    return;
      }
    }
  }
  if (!in_line)
    (void)fclose (impin);

}                         /* end of import_fdat */

/* ************************************************************** */
/* ************************************************************** */

void import_gsas (char input[100], int in_line, int Quick)

/* routine to extract structural information from a GSAS exp file */
{
  char string[256], filename[256], temp[5], tstring[20];
  int phase_no, i, j;
  FILE *impin;

  if (in_line) {
      Error_Box("Cannot read inline GSAS data. Please use the import instruction");
      return;
  }
  (void)sscanf (input, "%s %s %d", filename, filename, &phase_no);    /* get file name and phase number */
  if ((phase_no < 1) || (phase_no > 9)) {
      Error_Box("Illegal phase number in GSAS import instruction, Run aborted.");
      return;
  }
  if (!(impin = fopen (filename, "r"))) {
      Error_Box("Cannot Open GSAS Import File, Run aborted.");
      return;
  }
  sprintf (temp, "CRS%d", phase_no);         /* make search object */
  for (;;) {
    if (!fgets (string, 100, impin)) {         /* search for phase info */
        Error_Box("Error reading phase info from GSAS Import File, Run aborted.");
        return;
    }
    if (strncmp (string, temp, 4) == 0)
      break;
  }
  for (;;) {                     /* search for lattice constants */
    if (!fgets (string, 100, impin)) {
        Error_Box("Error reading lattice constants GSAS Import File, Run aborted.");
        return;
    }
    (void)sscanf (string, "%s %s", tstring, tstring);
    if (strncmp (tstring, "ABC", 3) == 0) {
      (void)sscanf (string, "%s %s %f %f %f", tstring, tstring, &drvui->lat_con[0],
          &drvui->lat_con[1]
          , &drvui->lat_con[2]);             /* read lattice constants */
      break;
    }
  }                         /* search for lattice constants */
  for (;;) {                     /* search for angles */
    if (!fgets (string, 100, impin)) {
        Error_Box("Error reading lattice angles GSAS Import File, Run aborted.");
        return;
    }
    (void)sscanf (string, "%s %s", tstring, tstring);
    if (strncmp (tstring, "ANGLES", 6) == 0) {
      (void)sscanf (string, "%s %s %f %f %f", tstring, tstring, &drvui->lat_con[3],
          &drvui->lat_con[4]
          , &drvui->lat_con[5]);             /* read lattice constants */
      break;
    }
  }                         /* search for angles */
  if (!Quick)
    fprintf (drvui->flout,
         "******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n",
         drvui->lat_con[0], drvui->lat_con[1], drvui->lat_con[2], drvui->lat_con[3], drvui->lat_con[4],
         drvui->lat_con[5]);
  for (;;) {                     /* position to atomic coordinates */
    if (!fgets (string, 100, impin)) {
        Error_Box("Error reading atomic coordinates from GSAS Import File, Run aborted.");
        return;
    }
    (void)sscanf (string, "%s %s", tstring, tstring);
    if (strncmp (tstring, "AT", 2) == 0)
      break;
  }                         /* position to atomic coordinates */

/* at this point, positioned at first atom line */
  for (;;) {                     /* process atom parameters */
    (void)sscanf (string, "%s %s", tstring, tstring);
    if (strncmp (tstring, "AT", 2) != 0)
      break;                     /* out when atom stuff done */
    (void)sscanf (string, "%s %s %s %s %f %f %f", tstring, tstring, tstring,
        tstring, &drvui->atom_xyz[natom][0], &drvui->atom_xyz[natom][1],
        &drvui->atom_xyz[natom][2]);
    drvui->atom_n[natom] = 0;
    for (i = 0; i < 4; i++)
      drvui->atom_l[natom][i] = ' ';             /* initialize atom name */
    j = 0;
    if (string[62] >= '0' && string[62] <= '9')
      drvui->atom_n[natom] = (int) string[62] - 48;
    else
      drvui->atom_l[natom][j++] = string[62];
    if (string[63] >= '0' && string[63] <= '9')
      drvui->atom_n[natom] = 10 * drvui->atom_n[natom] + (int) string[63] - 48;
    else
      drvui->atom_l[natom][j++] = string[63];
    if (string[64] >= '0' && string[64] <= '9')
      drvui->atom_n[natom] = 10 * drvui->atom_n[natom] + (int) string[64] - 48;
    else
      drvui->atom_l[natom][j++] = string[64];
    if (string[65] >= '0' && string[65] <= '9')
      drvui->atom_n[natom] = 10 * drvui->atom_n[natom] + (int) string[65] - 48;
    else
      drvui->atom_l[natom][j++] = string[65];
    if (!Quick)
      fprintf (drvui->flout,
           "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
           drvui->atom_l[natom][0]
           , drvui->atom_l[natom][1], drvui->atom_l[natom][2], drvui->atom_l[natom][3],
           drvui->atom_n[natom]
           , drvui->atom_xyz[natom][0], drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2]);
    (void)fgets (string, 100, impin);
    if (string[74] == 'A') {
      drvui->ell_type[n_ellips] = 1;         // Uij
      (void)sscanf (string, "%s %s %s %f %f %f %f %f %f", tstring, tstring, 
            tstring, &drvui->ellips[n_ellips][0], &drvui->ellips[n_ellips][1], &drvui->ellips[n_ellips][2], 
            &drvui->ellips[n_ellips][3], &drvui->ellips[n_ellips][4], &drvui->ellips[n_ellips][5]);    /* read unique coefficients */
      for (i = 0; i < 4; i++)
        drvui->ellips_l[n_ellips][i] = drvui->atom_l[natom][i];
      drvui->ellips_n[n_ellips] = drvui->atom_n[natom];
      drvui->save_el_number[n_ellips] = drvui->ellips_n[n_ellips];
      if (!Quick)
        fprintf (drvui->flout,
             "******      Uij  %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %8.5f %8.5f\n",
             drvui->atom_l[natom][0], drvui->atom_l[natom][1], drvui->atom_l[natom][2],
             drvui->atom_l[natom][3]
             , drvui->atom_n[natom], drvui->ellips[n_ellips][0],
             drvui->ellips[n_ellips][1], drvui->ellips[n_ellips][2]
             , drvui->ellips[n_ellips][3], drvui->ellips[n_ellips][4],
             drvui->ellips[n_ellips][5]);
      if (++n_ellips >= MAX_EL) {
          Error_Box("Too many ellipsoid types specified - increase MAX_EL.");
          return;
      }
    }                         /* string[74] == 'A' */
    drvui->atom_fn[natom] = 1;
    if (++natom >= MAX_ATOM) {
        Error_Box("Too many atoms in asymmetric unit - increase MAX_ATOM.");
        return;
    }
    if (!fgets (string, 100, impin)) {
        Error_Box("Error reading GSAS Import File, Run aborted.");
        return;
    }
  }                         /* process atom parameters */
  for (;;) {                     /* start loop for Space Group information */
    if (!fgets (string, 100, impin)) {
        Error_Box("Error reading GSAS Import File, Run aborted.");
        return;
    }
    (void)sscanf (string, "%s %s", tstring, tstring);
    if (strncmp (tstring, "SG", 2) == 0)
      break;                     /* out when space group found */
  }                         /* start loop for Space Group information */
  string[0] = 's';
  string[1] = 'p';
  string[2] = 'g';
  string[3] = 'r';
  for (i = 3; i < 60; i++)
    string[i + 1] = string[i + 10];
  string[60] = '\0';
  if (!Quick)fprintf (drvui->flout, "******      %s\n", string);
  symop (string);
  (void)fclose (impin);
}

/* ************************************************************** */
/* ************************************************************** */

void import_schakal (char input[100], int in_line, int Quick)

/* routine to extract structural information from a SCHAKAL data file */
{
  FILE *impin;
  char string[256];               /* temporary string */
  char filename[256];             /* string for filename */
  int intype;                     /* type of datum read */
  int i, j, k;                     /* temporary */

  if (in_line) {
    impin = drvui->fpin;
  }
  else {
    (void)sscanf (input, "%s %s", filename, filename); /* get file name */
    if (!(impin = fopen (filename, "r"))) {
        Error_Box("Cannot Open SCHAKAL Import File, Run aborted.");
        return;
    }
  }
  if (!Quick)
    fprintf (drvui->flout, "****** Now interpreting SCHAKAL-style data\n");
  while (!feof (impin)) {
    if (!fgets (input, 100, impin)) {         /* read a line */
        Error_Box("Error reading SCHAKAL file.");
        return;
    }
    intype = 0;
    for (i = 0; i < 4; i++) {
      if (input[i] == '\0')
    intype = 3;
    }
    if (strncmp (input, "TITL", 4) == 0)
      intype = 1;                 /* echo and ignore */
    if (strncmp (input, "titl ", 4) == 0)
      intype = 1;                 /* echo and ignore */
    if (strncmp (input, "CELL", 4) == 0)
      intype = 2;
    if (strncmp (input, "cell", 4) == 0)
      intype = 2;
    if (strncmp (input, "ATOM", 4) == 0)
      intype = 3;                 /* atom coordinate card */
    if (strncmp (input, "atom", 4) == 0)
      intype = 3;
    if (strncmp (input, "END", 3) == 0)
      intype = 4;                 /* trigger exit */
    if (strncmp (input, "end", 3) == 0)
      intype = 4;                 /* trigger exit */


    switch (intype) {
    case 1:
      for (i = 0; i < 51; i++)
    if (input[i] == '\n')
      input[i] = ' ';
      input[50] = '\0';                 /* truncate title */
      if (!Quick)
    fprintf (drvui->flout, "******      %s\n", input);
      break;                     /* ignore title except for listing */
    case 2:
      for (i = 3; i <= 5; ++i)
          drvui->lat_con[i] = 0.0f;
      (void)sscanf (input, "%s %f %f %f %f %f %f", string, &drvui->lat_con[0], &drvui->lat_con[1]
          , &drvui->lat_con[2], &drvui->lat_con[3], &drvui->lat_con[4], &drvui->lat_con[5]);
      if (!Quick)
          fprintf (drvui->flout,
             "******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n",
             drvui->lat_con[0], drvui->lat_con[1], drvui->lat_con[2], drvui->lat_con[3],
             drvui->lat_con[4], drvui->lat_con[5]);
      break;
    case 3:                     /* read an atom line */
      (void)sscanf (input, "%s %c%c%c%c %f %f %f", string,
          &drvui->atom_l[natom][0], &drvui->atom_l[natom][1]
          , &drvui->atom_l[natom][2], &drvui->atom_l[natom][3],
          &drvui->atom_xyz[natom][0], &drvui->atom_xyz[natom][1], &drvui->atom_xyz[natom][2]);
      drvui->atom_n[natom] = 0;
      k = 0;
      for (j = 0; j < 4; j++) {            /* extract atom number from SCHAKAL name */
          if (drvui->atom_l[natom][j] >= '0' && drvui->atom_l[natom][j] <= '9') {
              drvui->atom_n[natom] = drvui->atom_n[natom] * 10 + (int) drvui->atom_l[natom][j] - 48;
          } else {
              drvui->atom_l[natom][k++] = drvui->atom_l[natom][j];
          }
      }
      for (j = k; j < 4; j++)
          drvui->atom_l[natom][j] = ' ';
      if (drvui->atom_n[natom] == 0)
          drvui->atom_n[natom] = 1;
      if (!Quick)
          fprintf (drvui->flout,
              "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
              drvui->atom_l[natom][0], drvui->atom_l[natom][1], drvui->atom_l[natom][2],
              drvui->atom_l[natom][3], drvui->atom_n[natom], drvui->atom_xyz[natom][0],
              drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2]);
      drvui->atom_fn[natom] = 1;
      natom++;
      if (natom >= MAX_ATOM) {
          Error_Box("Too many atoms in asymmetric unit - increase MAX_ATOM.");
          return;
      }
      break;
    case 4:
      if (!Quick)
    fprintf (drvui->flout, "****** END - end of SCHAKAL dataset\n");
      if (!in_line)
    (void)fclose (impin);
      return;
    default:
      break;                     /* end of default */
    }                         /* end of switch */
  }
}                         /* end of import_schakal */

/* ************************************************************** */
/* ************************************************************** */

void import_shelx (char string[100], int in_line, int Quick)

/* routine to extract structural information from a SHELX data file */
{
  FILE *impin;
  char input[200];                /* temporary for continuation cards */
  char filename[100];             /* string for filename */
  int intype;                     /* type of datum read */
  int i, j, k;                    /* temporary */
  float xray;                     /* xray wavelength (read, but not used) */
  int read_start;                 /* flag for start of atom list */
  int read_uij;                   /* flag for uij card */
  float u11, u22;                 /* temporary storage for uij */
  float site_occup;               /* occupancy (read, but not used) */
  static char shcmds[][68] =
    { "ZERR", "LATT", "SYMM", "SFAC", "DISP", "UNIT", "LAUE",
    "REM ", "MORE", "TIME", "OMIT", "SHEL", "BASF", "TWIN",
    "EXTI", "SWAT", "HOPE", "MERG", "SPEC", "RESI", "MOVE",
    "ANIS", "AFIX", "HFIX", "FRAG", "FEND", "EXYZ", "EADP",
    "EQIV", "CONN", "PART", "BIND", "FREE", "DFIX", "DANG",
    "BUMP", "SAME", "SADI", "CHIV", "FLAT", "DELU", "SIMU",
    "DEFS", "ISOR", "NCSY", "SUMP", "L.S.", "CGLS", "BLOC",
    "DAMP", "STIR", "WGHT", "FVAR", "BOND", "CONF", "MPLA",
    "RTAB", "HTAB", "LIST", "ACTA", "SIZE", "TEMP", "WPDB",
    "FMAP", "GRID", "PLAN", "MOLE", "REM\0"
  };
  static char cmd[] = "    ";             /*temporary storage for start of line */
  float fvarpar[30];
  int nfvar=0;
  int fv;
  int latt;
  char *p,*pp;
    
  memset (input, ' ', (size_t) 199);
  latt = 0;

  if (in_line) {
    impin = drvui->fpin;
  } else {
    (void) sscanf (string, "%*s %s", filename);     /* get file name */
    if (!(impin = fopen (filename, "r"))) {
      Error_Box("Cannot Open SHELX Import File, Run aborted.");
      return;
    }
  }
  read_start = 0;
  read_uij = 0;

  if (drvui->sys == 0) { /* only do this if no spgr command given */
    /* initialize symmetry arrays */
    for (i = 0; i <= 23; ++i) {
      for (j = 0; j <= 2; ++j) {
        drvui->ts[i][j] = 0.0;
        for (k = 0; k <= 2; ++k)
          drvui->ss[i][j][k] = 0;
      }
    }
    for (i = 0; i <= 2; ++i) {
      drvui->lat_pos[0][i] = 0.0;
      drvui->ss[0][i][i] = 1;
      for (j = 0; j <= 3; ++j)
        drvui->spg[i][j] = ' ';
    }

  /* add trivial x,y,z symmetry */
    drvui->ts[0][0]=0;
    drvui->ts[0][1]=0;
    drvui->ts[0][2]=0;
    drvui->ss[0][0][0]=1;
    drvui->ss[0][1][1]=1;
    drvui->ss[0][2][2]=1;
    drvui->ng=1;
  }

  if (!Quick)
    fprintf (drvui->flout, "****** Now interpreting SHELX-style data\n");
  while (!feof (impin)) {
    if (!fgets (string, 100, impin)) {         /* read a line */
      if (!Quick)
         fprintf (drvui->flout, "Error reading input file.");
      return;
    }
    strcpy (input, string);
    if (strchr (input, '=') && (strncmp(input,"TITL",4) && strncmp(input,"REM",3))) {
      (*strchr (input, '=')) = '\0';
      if (!fgets (string, 100, impin)) {
          Error_Box("Error reading str file.");
          return;
      }
      strcat (input, string);
    }
    intype = 0;
    for (i = 0; i < 4; i++) {
      cmd[i] = (char) toupper (input[i]);
      if (input[i] == '\0')
        intype = 3;
    }

    if (strcmp (cmd, "TITL") == 0)
      intype = 1;                 /* echo and ignore */
    if (strcmp (cmd, "CELL") == 0)
      intype = 2;

    if (strcmp (cmd, "HKLF") == 0) 
    intype = 4;                 /* trigger exit */
    if (strncmp (cmd, "END", 3) == 0)
      intype = 4;                 /* trigger exit */

    if (strncmp (cmd, "FVAR", 4) == 0)
      intype = 5;                 /* store fixed parameters */

    if (drvui->sys == 0) { /* ignore LATT/SYMM if spgr already seen */
      if (strncmp (cmd, "LATT", 4) == 0)
        intype = 6;

      if (strncmp (cmd, "SYMM", 4) == 0)
        intype = 7;
    }
  
    if (intype == 0) {                 /* check against list of 'other' commands */
      for (i = 0; i < 68; i++) {
        if (strlen (cmd) < 4 || strcmp (cmd, shcmds[i]) == 0) {
          intype = 3;
          break;
        }
      }
    }
    if ((strncmp (cmd, "REM", 3) == 0) || (strncmp(cmd,"    ",4) == 0))
      intype = 3;
    switch (intype) {
    case 1:
      trim_string(input,51);
      if (!Quick)
        fprintf (drvui->flout, "******      %s\n", input);
      break;                     /* ignore title except for listing */
    case 2:
      for (i = 3; i <= 5; ++i)
        drvui->lat_con[i] = 0.0f;
      (void) sscanf (input, "%*s %f %f %f %f %f %f %f", &xray, &drvui->lat_con[0]
             , &drvui->lat_con[1], &drvui->lat_con[2], &drvui->lat_con[3], &drvui->lat_con[4]
             , &drvui->lat_con[5]);
      if (!Quick)
        fprintf (drvui->flout,
             "******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n",
             drvui->lat_con[0], drvui->lat_con[1], drvui->lat_con[2], drvui->lat_con[3],
             drvui->lat_con[4], drvui->lat_con[5]);
      break;
    case 3:                     /* just ignore most SHELX commands */
      break;
    case 4:
      read_start = 0;
      if (!Quick)
        fprintf (drvui->flout, "****** HKLF or END - end of SHELX dataset\n");
      if (!in_line)
        (void) fclose (impin);
      if (latt == 0 && drvui->sys == 0) {
          drvui->acentric = 0;    /* No LATT line given, spgp = P -1 */
          latt = 1;
          drvui->nlat = 1;
          drvui->nbr = 1;
      }
      if (latt == 3) drvui->sys = 5;
      if (latt == 4) drvui->sys = 6;
      if (drvui->sys == 0) findsys();
      for (i=1; i<n_ellips; i++) {
        if (drvui->ellips[i][0] <= 0.0f)
          drvui->ell_type[i] = -100;
      }
      if (drvui->autolabel==2) {
        for (i=0;i<natom;i++) {
          drvui->label_fn[drvui->nlabel]=drvui->frame_no;
          drvui->label_x[drvui->nlabel][0]=drvui->atom_xyz[i][0];
          drvui->label_x[drvui->nlabel][1]=drvui->atom_xyz[i][1];
          drvui->label_x[drvui->nlabel][2]=drvui->atom_xyz[i][2];
          int nn=0;
          for (j=0;j<4;j++) {
            if (drvui->atom_l[i][j] != ' ')
              drvui->label_label[drvui->nlabel][nn++]=drvui->atom_l[i][j];
          }
          drvui->label_label[drvui->nlabel][nn]='\0';
          sprintf(cmd,"%d",drvui->atom_n[i]);
          strcat(drvui->label_label[drvui->nlabel],cmd);

          drvui->nlabel++;
        }
        drvui->autolabel=1;
      }

      drvui->auto_ellipse = 0;
      return;
      break;
    case 5: 
      i = sscanf(input,"%*s %f %f %f %f %f %f %f",
            &fvarpar[nfvar],&fvarpar[nfvar+1],&fvarpar[nfvar+2],
            &fvarpar[nfvar+3],&fvarpar[nfvar+4],&fvarpar[nfvar+5],
            &fvarpar[nfvar+6]);
            nfvar +=7;
      if (i<7) nfvar -= (7-i);
      break;
    case 6:
      (void) sscanf(input,"%*s %d",&latt);
      if (latt <0) {
        drvui->acentric = 1;
        latt *= -1;
      } else
        drvui->acentric = 0;
      switch (latt) {
      case 1:
        drvui->nlat = 1;
        drvui->nbr = 1;
        break;
      case 2:
        drvui->nlat = 2;
        drvui->nbr = 6;
        drvui->lat_pos[1][0] = 0.5f;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.5f;
        break;
      case 3:
        drvui->nlat = 3;
        drvui->nbr = 7;
        drvui->lat_pos[1][0] = (1.f / 3.f);
        drvui->lat_pos[1][1] = (2.f / 3.f);
        drvui->lat_pos[1][2] = (2.f / 3.f);
        drvui->lat_pos[2][0] = (2.f / 3.f);
        drvui->lat_pos[2][1] = (1.f / 3.f);
        drvui->lat_pos[2][2] = (1.f / 3.f);
        break;
      case 4:
        drvui->nlat = 4;
        drvui->nbr = 5;
        drvui->lat_pos[1][0] = 0.0;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.5f;
        drvui->lat_pos[2][0] = 0.5f;
        drvui->lat_pos[2][1] = 0.0;
        drvui->lat_pos[2][2] = 0.5f;
        drvui->lat_pos[3][0] = 0.5f;
        drvui->lat_pos[3][1] = 0.5f;
        drvui->lat_pos[3][2] = 0.0;
        break;
      case 5:
        drvui->nlat = 2;
        drvui->nbr = 2;
        drvui->lat_pos[1][0] = 0.0;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.5f;
        break;
      case 6:
        drvui->nlat = 2;
        drvui->nbr = 3;
        drvui->lat_pos[1][0] = 0.5f;
        drvui->lat_pos[1][1] = 0.0;
        drvui->lat_pos[1][2] = 0.5f;
        break;
      case 7:
        drvui->nlat = 2;
        drvui->nbr = 4;
        drvui->lat_pos[1][0] = 0.5f;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.0;
      }
      break;

    case 7: /* SYMM */
      drvui->ng++;
      input[0]=' ';
      input[1]=' ';
      input[2]=' ';
      input[3]=' ';
      p = strstr(input,",");
      *p='\0';
      getsym(input,drvui->ng-1,0);
      p++;
      pp = strstr(p,",");
      *pp='\0';
      getsym(p,drvui->ng-1,1);
      pp++;
      getsym(pp,drvui->ng-1,2);
      break;

    default:
      site_occup=1.;
      u11=u22=0.05f;
      i = sscanf (input, "%c%c%c%c %d %f %f %f %f %f %f %f %f %f %f",
          &drvui->atom_l[natom][0], &drvui->atom_l[natom][1],
          &drvui->atom_l[natom][2], &drvui->atom_l[natom][3],
          &drvui->atom_n[natom], &drvui->atom_xyz[natom][0],
          &drvui->atom_xyz[natom][1], &drvui->atom_xyz[natom][2],
          &site_occup, &u11, &u22, &drvui->ellips[n_ellips][2],
          &drvui->ellips[n_ellips][5], &drvui->ellips[n_ellips][4],
          &drvui->ellips[n_ellips][3]);
      for (j = 0; j < 3; j++)
        if (fabs(drvui->atom_xyz[natom][j]) >15.) {
          fv=((int)fabs(drvui->atom_xyz[natom][j]))/10;
      if (drvui->atom_xyz[natom][j] <0.)
        drvui->atom_xyz[natom][j] = (float)(1.- fvarpar[fv -1]* (drvui->atom_xyz[natom][j]-fv*10));
      else
        drvui->atom_xyz[natom][j] = (float)fvarpar[fv -1]* (drvui->atom_xyz[natom][j]-fv*10);
    } else  
      if (drvui->atom_xyz[natom][j] > 9.)
        drvui->atom_xyz[natom][j] -= 10.0f;
      
    if (fabs(site_occup) >11.) { 
      fv= ((int)fabs(site_occup))/10;
      if (site_occup <0)
        site_occup = (float)(1.0-fvarpar[fv-1]);
      else
        site_occup = fvarpar[fv-1];
    }
    if (fabs(u11) >10.) { 
      fv = ((int)fabs(u11))/10;
      if (u11 <0.)
        u11 = (float)(1.0- fvarpar[fv-1] *(fabs(u11)-fv));
      else
        u11 = fvarpar[fv-1] * (u11-fv);
    }    

      drvui->atom_n[natom] = 0;
      int k = 0;
      for (j = 0; j < 4; j++)             /* extract atom number if part of SHELX name */
        if (drvui->atom_l[natom][j] >= '0' && drvui->atom_l[natom][j] <= '9') {
          drvui->atom_n[natom] = 10 * drvui->atom_n[natom] + (int) drvui->atom_l[natom][j] - 48;
        } else {
            drvui->atom_l[natom][k++] = drvui->atom_l[natom][j];
        }
        for (j=k;j<4;j++)
            drvui->atom_l[natom][j] = ' ';
        drvui->atom_l[natom][4] = '\0';
        if (drvui->atom_n[natom] == 0)
          drvui->atom_n[natom] = 1;

        if (i > 11) {
          strcpy(drvui->ellips_l[n_ellips],drvui->atom_l[natom]);
          drvui->ellips_n[n_ellips] = drvui->atom_n[natom];
          drvui->save_el_number[n_ellips] = drvui->ellips_n[n_ellips];
          if (fabs(u22) >10.) { 
            fv = ((int)fabs(u22))/10;
            if (u22 <0.)
              u22 = (float)(1. - fvarpar[fv-1] * (fabs(u22)-fv*10));
            else
              u22 = fvarpar[fv-1] * (u22-fv*10);
          }
          if (fabs(drvui->ellips[n_ellips][2]) >10.) { 
            fv= ((int)fabs(drvui->ellips[n_ellips][2]))/10;
            if (drvui->ellips[n_ellips][2] <0. )
              drvui->ellips[n_ellips][2] = (float)(1.0 -fvarpar[fv-1] * (fabs(drvui->ellips[n_ellips][2])-fv*10));
            else
              drvui->ellips[n_ellips][2] =  fvarpar[fv-1] * (drvui->ellips[n_ellips][2]-fv*10);
          }
          if (fabs(drvui->ellips[n_ellips][5]) >10.) { 
            fv= ((int)fabs(drvui->ellips[n_ellips][5]))/10;
            if (drvui->ellips[n_ellips][5] <0. )
              drvui->ellips[n_ellips][5] = (float)(1. -fvarpar[fv-1] * (fabs(drvui->ellips[n_ellips][5])-fv*10));
            else
              drvui->ellips[n_ellips][5] =  fvarpar[fv-1] * (drvui->ellips[n_ellips][5]-fv*10);
          }
          if (drvui->ellips[n_ellips][4] >10.) { 
            fv= ((int)fabs(drvui->ellips[n_ellips][4]))/10;
            if (drvui->ellips[n_ellips][4] <0. )
              drvui->ellips[n_ellips][4] = (float)(1.0 -fvarpar[fv-1] * (fabs(drvui->ellips[n_ellips][4])-fv*10));
            else
              drvui->ellips[n_ellips][4] =  fvarpar[fv-1] * (drvui->ellips[n_ellips][4]-fv*10);
          }
          if (drvui->ellips[n_ellips][3] >10.) { 
            fv= ((int)fabs(drvui->ellips[n_ellips][3]))/10;
            if (drvui->ellips[n_ellips][3] <0. )
              drvui->ellips[n_ellips][3] = (float)(1.0 -fvarpar[fv-1] * (fabs(drvui->ellips[n_ellips][3])-fv*10));
            else
              drvui->ellips[n_ellips][3] =  fvarpar[fv-1] * (drvui->ellips[n_ellips][3]-fv*10);
          }

          drvui->ellips[n_ellips][0] = u11;
          drvui->ellips[n_ellips][1] = u22;
          if (!Quick)
            fprintf (drvui->flout,
                 "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f \n",
                 drvui->atom_l[natom][0], drvui->atom_l[natom][1],
                 drvui->atom_l[natom][2], drvui->atom_l[natom][3]
               , drvui->atom_n[natom], drvui->atom_xyz[natom][0],
                 drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2]);
          if (!Quick)
            fprintf (drvui->flout,
               "******      Uij  %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %8.5f %8.5f\n",
               drvui->atom_l[natom][0], drvui->atom_l[natom][1],
               drvui->atom_l[natom][2], drvui->atom_l[natom][3]
             , drvui->atom_n[natom], drvui->ellips[n_ellips][0],
               drvui->ellips[n_ellips][1], drvui->ellips[n_ellips][2]
             , drvui->ellips[n_ellips][3], drvui->ellips[n_ellips][4],
               drvui->ellips[n_ellips][5]);
          drvui->save_el_number[n_ellips] = drvui->ellips_n[n_ellips];
          do_ellipsoids = 1;
          if (drvui->auto_ellipse) {
              drvui->ell_type[n_ellips] = 1001;       // Uij with ellipcolor
              memset(drvui->ellips_col_p[n_ellips],0,40);
              memset(drvui->ellips_col_v[n_ellips],0,40);
              strcpy(drvui->ellips_col_p[n_ellips],"Gray30");
              strcpy(drvui->ellips_col_v[n_ellips],"0.3 0.3 0.3");
          } else {
              drvui->ell_type[n_ellips] = 1;          // Uij
          }
          n_ellips++;
          if (n_ellips >= MAX_EL) {
            Error_Box("Too many ellipsoid types specified - increase MAX_EL.");
            return;
          }
        } else {
          if (i < 10) {
            if (!Quick)
              fprintf (drvui->flout,
                     "******WARNING - too few parameters given for following atom:\n");
            if (site_occup < 0)
              site_occup = 0.;
            if (u11 < 0)
              u11 = 0.;
          }
          if (!Quick)
            fprintf (drvui->flout,
                "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %8.5f\n",
                drvui->atom_l[natom][0], drvui->atom_l[natom][1],
                drvui->atom_l[natom][2], drvui->atom_l[natom][3]
              , drvui->atom_n[natom], drvui->atom_xyz[natom][0],
                drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2], site_occup, u11);
        }
        drvui->atom_fn[natom] = 1;
        natom++;
        if (natom >= MAX_ATOM) {
           Error_Box("Too many atoms in asymmetric unit - increase MAX_ATOM.");
           return;
        }
        break;                     /* end of default */
      }                         /* end of switch */
  }
  if (!in_line)
    (void) fclose (impin);
  
  if (latt == 3) drvui->sys = 5;
  if (latt == 4) drvui->sys = 6;
  
  if (drvui->sys == 0) findsys();
  for (i=1; i<n_ellips; i++) {
      if (drvui->ellips[i][0] <= 0.0f)
          drvui->ell_type[i] = -100;
  }
  if (drvui->autolabel==2) {
    for (i=0;i<natom;i++) {
      if(drvui->atom_fn[i] != drvui->frame_no) continue;
      drvui->label_fn[drvui->nlabel]=drvui->frame_no;
      drvui->label_x[drvui->nlabel][0]=drvui->atom_xyz[i][0];
      drvui->label_x[drvui->nlabel][1]=drvui->atom_xyz[i][1];
      drvui->label_x[drvui->nlabel][2]=drvui->atom_xyz[i][2];
      int nn=0;
      for (j=0;j<4;j++) {
        if (drvui->atom_l[i][j] != ' ')
          drvui->label_label[drvui->nlabel][nn++]=drvui->atom_l[i][j];
      }
      drvui->label_label[drvui->nlabel][nn]='\0';
      sprintf(cmd,"%d",drvui->atom_n[i]);
      strcat(drvui->label_label[drvui->nlabel],cmd);

      drvui->nlabel++;
    }
    drvui->autolabel=1;
    drvui->auto_ellipse = 0;
  }
}                         /* end of import_shelx */

/* ************************************************************** */
/* ************************************************************** */

void import_wien (char string[100], int in_line, int Quick)

/* routine to extract structural information from a WIEN2k struct file */
{
  FILE *impin;
  char filename[100];             /* string for filename */
  char line[85];
  int i, j, k, l, k3, k4;                    /* temporary */
  double bohrfactor=1.0;
  char spgp[50], spstring[40],*spgr;
  char xs[15],ys[15],zs[15];
  int nat=0;
  char lattice[4];

  if (in_line) {
    impin = drvui->fpin;
  } else {
    (void) sscanf (string, "%*s %s", filename);     /* get file name */
    if (!(impin = fopen (filename, "r"))) {
      Error_Box("Cannot Open WIEN2k Import File, Run aborted.");
      return;
    }
  }

  if (!Quick)
    fprintf (drvui->flout, "****** Now interpreting WIEN2k-style data\n");

   memset(string,0,85);


    if (!fgets (string, 85, impin)) {         /* skip title */
      if (!Quick)
         fprintf (drvui->flout, "Error reading title.");
      return;
    }
    trim_string(string,85);          // kill any ^M or ^J at end of line
    string[85]='\0';

    if (!Quick) fprintf(drvui->flout,"%s\n",string);

//fprintf(stderr,"title ##%s##(%d)\n",string,strlen(string));

    if (!fgets (line, 85, impin)) {         
      if (!Quick)
         fprintf (drvui->flout, "Error reading header.");
      return;
    }


    memset(spgp,0,40);
    sscanf(line,"%s %*s %1d%s",lattice,&natom,spgp);
    spgr=strchr(spgp,'_')+1;

    strcpy(line,spgr);

/* separate class symbol from other symmetry elements */
   spstring[0] = 's';
   spstring[1] = 'p';
   spstring[2] = 'g';
   spstring[3] = 'r';
   spstring[4] = ' ';
   spstring[5] = line[0];
   spstring[6] = ' ';
   j = 6;
   k = 0;
   k3 = 0;
   k4 = 0;
   l = (int) strlen (line);
   for (i = 1; i <= l; i++) {
     switch (line[i]) {
     case ' ':
       break;
     case '/':
       spstring[++j] = line[i];
       break;
     default:
       j++;
       /* do we have to collate this symbol (for screw axes or mirror planes) ? */
       if (line[i] > '0' && line[i] <= '9') { /* it is a number */
         if (line[0] == 'R' ||             /* no screw axes in R */
           l == 4 ||                 /* no screw axes in four letter names */
           k != 0 ||                 /* have just appended to a number */
           k3 != 0 ||                 /* cannot have further screw axes after 3_x */
             (k4 != 0 && spstring[j - 1] == '3')) {    /* cannot have 3_x after a 4_x */
             spstring[j] = ' ';             /* so separate it */
             k = 0;                 /* and reset the collation flag */
             j++;
         } else if (spstring[j - 1] != '-' && (spstring[j - 1] <= string[i]
                        || spstring[j - 1] >= 'a')) {
       /* neither a minus sign nor a higher number precedes it */
           spstring[j] = ' ';             /* so separate it */
           k = 0;                 /* and reset the collation flag */
           j++;
         } else {
           k++;                     /* set flag for collation */
           if (spstring[j - 1] == '3')
             k3 = 1;                 /* if 3_x screw, prohibit further screw axes */
           if (spstring[j - 1] == '4')
             k4 = 1;                 /* if 4_x screw, prohibit 3_x screw axes */
         }
       } else {                     /* not a number */
         if (spstring[j - 1] != '/') {         /*but not a mirror plane either */
           spstring[j] = ' ';             /* separate it */
           j++;
         }
       }
       spstring[j] = line[i];             /* now add the symbol */
       break;
     }                         /*switch */
   }                         /*for i */
   spstring[j + 1] = '\0';
 
/* Space group 90, P 4 21 2, is not correctly identified by this code. */
   if (!strcmp (line, "P4212"))
     strcpy (spstring, "spgp P 4 21 2");

   if (!Quick) fprintf(drvui->flout,"%s Lattice, %d atoms, space group %s\n",lattice,natom,spstring);

   symop(spstring);

   if (!fgets (line, 85, impin)) {         /* read a line */
     if (!Quick)
       fprintf (drvui->flout, "Error reading units flag.");
     return;
  }

  if ( strstr(line,"bohr") )  {
	 bohrfactor=0.592;
    if (!Quick) fprintf(drvui->flout,"Converting units from Bohr to Angstrom\n");
  }

  if (!fgets (line, 85, impin)) {         /* read a line */
    if (!Quick)
         fprintf (drvui->flout, "Error reading lattice constants.");
    return;
  }

  sscanf(line,"%f %f %f %f %f %f",
		&drvui->lat_con[0],&drvui->lat_con[1],&drvui->lat_con[2],
		&drvui->lat_con[3],&drvui->lat_con[4],&drvui->lat_con[5]); 
  drvui->lat_con[0] *= (float) bohrfactor;
  drvui->lat_con[1] *= (float) bohrfactor;
  drvui->lat_con[2] *= (float) bohrfactor;

  if (!Quick) fprintf(drvui->flout,"Lattice constants %f %f %f %5.2f %5.2f %5.2f\n",
		drvui->lat_con[0],drvui->lat_con[1],drvui->lat_con[2],
		drvui->lat_con[3],drvui->lat_con[4],drvui->lat_con[5]); 

  nat=0;

  while (!feof (impin) &&nat <natom) {
    if (!fgets (line, 85, impin)) {         /* read a line */
      if (!Quick)
        fprintf (drvui->flout, "Error reading atom record.");
      return;
    }
    if (!strncmp(line,"ATOM",4)) {
      sscanf(line,"%*s %*s %s %s %s",xs,ys,zs);
      drvui->atom_xyz[nat][0]=(float) atof(xs+2); 
      drvui->atom_xyz[nat][1]=(float) atof(ys+2); 
      drvui->atom_xyz[nat][2]=(float) atof(zs+2); 
    } else if (strstr(line,"NPT=")) {
      sscanf (line,"%c%c%c%c",
         &drvui->atom_l[nat][0], &drvui->atom_l[nat][1],
         &drvui->atom_l[nat][2], &drvui->atom_l[nat][3]);
      drvui->atom_l[nat][4]='\0';

      drvui->atom_n[nat] = 0;

      nat++;
    }
  }

  if (!Quick) {
    for (i=0;i<natom;i++) fprintf(drvui->flout,"Atom %c%c%c%c %d  %f %f %f\n",
          drvui->atom_l[i][0], drvui->atom_l[i][1],
          drvui->atom_l[i][2], drvui->atom_l[i][3],
          drvui->atom_n[i],
          drvui->atom_xyz[i][0], drvui->atom_xyz[i][1], 
          drvui->atom_xyz[i][2]); 
  }

  if (!in_line)
    (void) fclose (impin);
}                         /* end of import_wien */

/* ************************************************************** */
/* ************************************************************** */

void Init_DRAWxtl(void)
{
/* Initialize */

  drvui->X_Origin = drvui->Y_Origin = drvui->Z_Origin = 0.5f;
  drvui->X_Boxlim = drvui->Y_Boxlim = drvui->Z_Boxlim = 20.0f;
  drvui->Trans[0] = drvui->Trans[1] = drvui->Trans[2] = 0.0f;
  xrot = yrot = zrot = 0.0f;
  boxflag = packflag = clipflag = 0;
  do_ellipsoids = edges = 0;
  Ellipsoid_Prob = 0.5f;
  cur_cen[0] = 0.25f;
  cur_cen[1] = 0.25f;
  cur_cen[2] = 0.25f;                // location of cursor (fractional coordinates)
  drvui->cur_step = 0.5f;            // amount to jump cursor in A
  Unit_Cell = M_cameras = no_comment = Labels = Vrml2 = 1;
  Display_axes = El_Cutout = 0;
  strcpy (drvui->Cutout_color, "Gray20");
  strcpy (Ellipaxis_color, "Gray20");
  drvui->SpMult = drvui->BndMult = Magnification = 1.0f;
  Ellipsoid_Scale = P_to_C (0.5f);
  Options = 0;
  FourierMapType = 0;
  Map_Info.info_valid = 0;
  Map_Info.map_type = -1;
  drvui->sys = 0;
  drvui->polylimit = 0.1f;
  drvui->Phong_Value = 0.2f;
  drvui->Phong_Size = 1.0f;
  DepthCue = 0.0f;
  drvui->noshadow = 0;
/*
  drvui->ambient = 0.0f;
  drvui->diffuse = 0.0f;
  drvui->specular = 0.0f;
  drvui->roughness = 0.0f;
*/
  printdist = 3.5f;
  origin[0] = drvui->X_Origin;
  origin[1] = drvui->Y_Origin;
  origin[2] = drvui->Z_Origin;
  boxlim[0] = drvui->X_Boxlim;
  boxlim[1] = drvui->Y_Boxlim;
  boxlim[2] = drvui->Z_Boxlim;
  memset(drvui->p_rad_edge,0,sizeof(drvui->p_rad_edge));
  docell = Unit_Cell;
  Color_Warning = 0;
  El_Cutout = 0;
  strcpy(drvui->Cutout_color,"");
  rad_cell = 0.02f;
  drvui->Sphere_Mult = drvui->SpMult;
  drvui->Bond_Mult = drvui->BndMult;
  natom = 0;
  drvui->nmag = 0;
  drvui->nsphere = drvui->npoly = drvui->nbond = drvui->nplane = 1;
  n_ellips = drvui->ncone = drvui->nlabel = 1;
  drvui->nlabel = 1;
  drvui->poly_size[0] = 2.5f;
  drvui->poly_min[0] = 0.005f;
  memset(drvui->poly_l[0],' ',4);
  drvui->plane_size[0] = 2.5f;
  memset(drvui->plane_l[0],' ',4);
  drvui->sphere_size[0] = 0.0f;
  memset(drvui->sphere_l[0],' ',4);
//  POV_Max[0] = POV_Max[1] = POV_Max[2] = -99999.0f;
//  POV_Min[0] = POV_Min[1] = POV_Min[2] = 99999.0f;
  drvui->glback[0] = drvui->glback[1] = drvui->glback[2] = 1.0f;
  slabmode = 0;
  domolcomp = 0;
  drvui->mol_d = 0.0f;
  Omit->nomits=0;
  memset(Omit->omit1,0,1000);
  memset(Omit->omit2,0,1000);
  memset(drvui->slab_con,0,sizeof(drvui->slab_con));
  memset(drvui->slab_rot,0,sizeof(drvui->slab_rot));
  memset(drvui->slab_off,0,sizeof(drvui->slab_off));
// clear all the color variables
  memset(Ellipaxis_color,0,sizeof(Ellipaxis_color));
  memset(col_cell_p,0,sizeof(col_cell_p));
  memset(col_edge_p,0,sizeof(col_edge_p));
  memset(drvui->col_cone_p,0,sizeof(drvui->col_cone_p));
  memset(drvui->col_arrow_p,0,sizeof(drvui->col_arrow_p));
  memset(drvui->col_bg_p,0,sizeof(drvui->col_bg_p));
  memset(drvui->col_bond_p,0,sizeof(drvui->col_bond_p));
  memset(drvui->Cutout_color,0,sizeof(drvui->Cutout_color));
  memset(drvui->ellips_col_p,0,sizeof(drvui->ellips_col_p));
  memset(drvui->p_col_edge_p,0,sizeof(drvui->p_col_edge_p));
  memset(drvui->poly_col_p,0,sizeof(drvui->poly_col_p));
  memset(drvui->plane_col_p,0,sizeof(drvui->plane_col_p));
  memset(drvui->sphere_col_p,0,sizeof(drvui->sphere_col_p));
// clear all label variables
  memset(drvui->atom_l,0,sizeof(drvui->atom_l));
  memset(drvui->bond_l1,0,sizeof(drvui->bond_l1));
  memset(drvui->bond_l2,0,sizeof(drvui->bond_l2));
  memset(drvui->cone_l1,0,sizeof(drvui->cone_l1));
  memset(drvui->ellips_l,0,sizeof(drvui->ellips_l));
  memset(drvui->plane_l,0,sizeof(drvui->plane_l));
  memset(drvui->poly_l,0,sizeof(drvui->poly_l));
  memset(drvui->sphere_l,0,sizeof(drvui->sphere_l));

  strcpy (col_cell_v, "0 0 0");             /* preset unit-cell colors to Black */
  strcpy (col_cell_p, "Black");
  strcpy (drvui->col_bg_p, "White");
  strcpy(Map_Info.title," Title not given in map file");
  strcpy (drvui->col_bg_v, "1. 1. 1.");

}

/* ************************************************************** */
/* ************************************************************** */

float P_to_C (float prob)

/* Routine to convert probability to ellipsoid scaling, C.  Given probability
   'prob', calculate value of C required for ellipsoid,

   C^2 = (delx/sigx)^2 + (dely/sigy)^2 + (delz/sigz)^2,

   where 'prob' is the probability that a random point in the distribution
   will fall inside the ellipsoid.  Table values taken from D.B. Owen,
   "Handbook of Statistical Tables", Addison-Wesley, Reading, Mass., 1962.

   N.B. The present version of the tables only handles probabilities from
   0.01 to 0.99.
*/
#define N_C 99
{
  static double C_table[N_C + 1] =
    { 0.3389, 0.4299, 0.4951, 0.5479, 0.5932, 0.6334, 0.6699, 0.7035,
    0.7349, 0.7644,
    0.7924, 0.8192, 0.8447, 0.8694, 0.8932, 0.9162, 0.9386, 0.9605,
    0.9818, 1.0026,
    1.0230, 1.0430, 1.0627, 1.0821, 1.1012, 1.1200, 1.1386, 1.1570,
    1.1751, 1.1932,
    1.2110, 1.2288, 1.2464, 1.2638, 1.2812, 1.2985, 1.3158, 1.3330,
    1.3501, 1.3672,
    1.3842, 1.4013, 1.4183, 1.4354, 1.4524, 1.4695, 1.4866, 1.5037,
    1.5209, 1.5382,
    1.5555, 1.5729, 1.5904, 1.6080, 1.6257, 1.6436, 1.6616, 1.6797,
    1.6980, 1.7164,
    1.7351, 1.7540, 1.7730, 1.7924, 1.8119, 1.8318, 1.8519, 1.8724,
    1.8932, 1.9144,
    1.9360, 1.9580, 1.9804, 2.0034, 2.0269, 2.0510, 2.0757, 2.1012,
    2.1274, 2.1544,
    2.1824, 2.2114, 2.2416, 2.2730, 2.3059, 2.3404, 2.3767, 2.4153,
    2.4563, 2.5003,
    2.5478, 2.5997, 2.6571, 2.7216, 2.7955, 2.8829, 2.9912, 3.1365,
    3.3682, 5.9503
  };

  int i;
  double C;

  i = (int) (100.0 * prob);             /* find correct table entry */
  if (i > N_C)
    i = N_C;
/* interpolate (linear) */
  C = C_table[i - 1] + (prob - 0.01 * i) * (C_table[i] - C_table[i - 1]);
  return ((float) C);
}


/* ************************************************************** */
/* ************************************************************** */

void read_fcf(char * infile, int Quick)
{
// read CIF-standard Fo/Fc file and compute density

    FILE * mapin;
    float *fo,*fc,*f2,*ftmp1, *ftmp2, *ftmp3;
    short *ih,*ik,*il,*itmp1, *itmp2, *itmp3;
    float * Cklx;
    float * Dklx;
    float * Elxy;
    float * Flxy;
    double x,y,z;
    char string[200];
    char maptitle[61];
    int squared,phases,ab;
    int nval=200;
    int nr;
    int i,j,k,l;
    int ijk = 0;
    float rhomin=1.0e15f, rhomax=-1.0e15f;
    float factor;
    short ih0,ik0,il0;
    short ih1,ik1,il1;
    float fo0,fc0,fc1,f20,f21;
    int kmin, kmax, nk;
    int lmin, lmax, nl;
    float phase;
    int skip;
    int progress;

    if (FourierPt != NULL) {
        return;
    }

    if ((mapin = fopen (infile, "r")) == NULL) {
        sprintf(string,"Cannot open Fo/Fc (CIF fcf) file %s", infile);
        Error_Box(string);
        return;
    }

    fo=(float*)malloc(200*sizeof(float));
    if (fo==NULL) {
        sprintf(string,"Unable to obtain storage for Fo data");
        Error_Box(string);
        return;
    }
    fc=(float*)malloc(200*sizeof(float));
    if (fc==NULL) {
        sprintf(string,"Unable to obtain storage for Fc data");
        Error_Box(string);
        return;
    }
    f2=(float*)malloc(200*sizeof(float));
    if (f2==NULL) {
        sprintf(string,"Unable to obtain storage for B or phase data");
        Error_Box(string);
        return;
    }
    ih=(short*)malloc(200*sizeof(short));
    if (ih==NULL) {
        sprintf(string,"Unable to obtain storage for h data");
        Error_Box(string);
        return;
    }
    ik=(short*)malloc(200*sizeof(short));
    if (ik==NULL) {
        sprintf(string,"Unable to obtain storage for k data");
        Error_Box(string);
        return;
    }
    il=(short*)malloc(200*sizeof(short));
    if (il==NULL) {
        sprintf(string,"Unable to obtain storage for l data");
        Error_Box(string);
        return;
    }
 
    do {
      fgets(string,80,mapin);
      if (strstr(string,"title")) {
        memset(maptitle,0,61); 
        sscanf(string,"%*s %60c",maptitle); 
        trim_string(maptitle,60);
        strcpy(Map_Info.title,maptitle);
        Map_Info.info_valid = 1;
      }
    } while (strncmp(string," _refln",6) ); // search for start of _refln group

    if (strncmp(string," _refln_index_h",14)) {
      sprintf(string,"Unsupported field sequence in fcf file - need h k l first");
      Error_Box(string);
      return;
    }

    squared = 0; 
    phases = 0;
    ab = 0;
    while (!strncmp(string," _refln",6) ) {
      fgets(string,80,mapin);
      if (strstr(string,"squared")) squared=1;
      if (strstr(string,"_phase_")) phases=1;
      if (strstr(string,"A_calc")) ab = 1;
    }

    if (phases == 0 && ab == 0) {
      sprintf(string,"fcf file does not contain phase angle or A,B data");
      Error_Box(string);
      return;
    }

    nr = 0;
    do {
      i=sscanf(string,"%hd %hd %hd %f %*f %f %f",&ih0,&ik0,&il0,&fo0,&fc0,&f20);
      if (i<6) break;
      if (squared==1) fo0=(float)sqrt(fo0);
      if (ab) {       // change all data to 'phase' type
          fc1 = (float) sqrt(fc0 * fc0 + f20 * f20);
          f20 = (float)atan2(f20/fc1, fc0/fc1);
          fc0 = fc1;
      } else {
          f20 *= (float)(PI/180.0);     // phase in radians
      }

// apply spacegroup symmetry (including xyz - catch duplicates in input list)
      for (k = 0; k < drvui->ng; ++k) {
          skip=0;

          ih1=drvui->ss[k][0][0]*ih0 + drvui->ss[k][1][0]*ik0 + drvui->ss[k][2][0]*il0;
          ik1=drvui->ss[k][0][1]*ih0 + drvui->ss[k][1][1]*ik0 + drvui->ss[k][2][1]*il0;
          il1=drvui->ss[k][0][2]*ih0 + drvui->ss[k][1][2]*ik0 + drvui->ss[k][2][2]*il0;
          phase = 2.0f * (float)PI * (drvui->ts[k][0]*ih0 + drvui->ts[k][1]*ik0 
              + drvui->ts[k][2]*il0);
          f21 = f20 + phase;
          f21 = (float)atan2(sin(f21),cos(f21));    // get f21 in range -PI to PI
       
          for (j=0;j<nr;j++) {
              if (ih[j] == ih1 && ik[j] == ik1 && il[j] == il1) { 
                skip=1;
                break;
              }
          }
          if (skip==0) {
            ih[nr] = ih1;
            ik[nr] = ik1;
            il[nr] = il1;
            fo[nr] = fo0;
            fc[nr] = fc0;
            f2[nr] = f21;
            nr++;

          }
          ih1 *= -1;               // add Friedel opposite
          ik1 *= -1;
          il1 *= -1;
          f21 *= -1.0;
          skip = 0;

          for (j=0;j<nr;j++) {
              if (ih[j] == ih1 && ik[j] == ik1 && il[j] == il1) {
                skip=1;
                break;
              }
          }
          if (skip==0) {
            ih[nr]=ih1;
            ik[nr]=ik1;
            il[nr]=il1;
            fo[nr]=fo0;
            fc[nr] = fc0;
            f2[nr] = f21;
            nr++;
          }

          if (nr > nval-10) {
            nval +=100;
            ftmp1=(float*)realloc(fo,nval*sizeof(float));
            ftmp2=(float*)realloc(fc,nval*sizeof(float));
            ftmp3=(float*)realloc(f2,nval*sizeof(float));
            itmp1=(short*)realloc(ih,nval*sizeof(short));
            itmp2=(short*)realloc(ik,nval*sizeof(short));
            itmp3=(short*)realloc(il,nval*sizeof(short));
            if (!ftmp1 || !ftmp2 || !ftmp3 || !itmp1 || !itmp2 || !itmp3) {
              sprintf(string,"Unable to expand storage for h, k, l, fo, fc or phase data");
              free(fo);
              free(fc);
              free(f2);
              free(ih);
              free(ik);
              free(il);
              Error_Box(string);
              return;
            } else {
              fo = ftmp1;
              fc = ftmp2;
              f2 = ftmp3;
              ih = itmp1;
              ik = itmp2;
              il = itmp3;
            }
          } // if we need to realloc
      } // loop over all symmetry operators
      fgets(string,80,mapin);
    } while (i>0 && !feof(mapin));

    fclose(mapin);


    factor = 1.0f/(drvui->lat_con[0]*drvui->lat_con[1]*drvui->lat_con[2]
        * (float)sqrt(1.0 - cos(drvui->lat_con[3] * PI / 180.0) *
                     cos(drvui->lat_con[3] * PI / 180.0) *
                     cos(drvui->lat_con[4] * PI / 180.0) *
                     cos(drvui->lat_con[4] * PI / 180.0) *
                     cos(drvui->lat_con[5] * PI / 180.0) *
                     cos(drvui->lat_con[5] * PI / 180.0)));
    kmin = kmax = ik[0];
    lmin = lmax = il[0];

    for (l=0; l < nr; l++) {
        if (ik[l] < kmin) kmin = ik[l];     // find range of k and l
        if (ik[l] > kmax) kmax = ik[l];
        if (il[l] < lmin) lmin = il[l];
        if (il[l] > lmax) lmax = il[l];
        switch (Map_Info.map_type) {        // get map coefficients/V
        default:
        case 0:                             // Fo map
            break;
        case 1:                             // Fc map
            fo[l] = fc[l];
            break;
        case 2:                             // Fo - Fc map
            fo[l] = fo[l] - fc[l];
            break;
        case 3:                             // 2Fo - Fc map
            fo[l] = 2.0f * fo[l] - fc[l];
            break;
        case 4:                             // Fo2 (Patterson)
            fo[l] = fo[l] * fo[l];
            f2[l] = 0.0f;
            break;
        }
        fo[l] *= factor;                          // update coefficients times 1/V
    }

    nk = kmax - kmin +1;
    nl = lmax - lmin +1;
    Cklx = (float *)malloc(nk * nl * sizeof(float));
    Dklx = (float *)malloc(nk * nl * sizeof(float));
    Elxy = (float *)malloc(nl * sizeof(float));
    Flxy = (float *)malloc(nl * sizeof(float));

    if (!Cklx || !Dklx || !Elxy || !Flxy) {
        Error_Box("Unable to allocate Beevers-Lipson arrays.");
        return;
    }
    mapstep_a = (int) (4.0f * drvui->lat_con[0] + 0.5f);
    mapstep_b = (int) (4.0f * drvui->lat_con[1] + 0.5f);
    mapstep_c = (int) (4.0f * drvui->lat_con[2] + 0.5f);

    FourierPt = (float *) malloc( mapstep_a*mapstep_b*mapstep_c * sizeof(float));
    if (FourierPt == NULL) {
      Error_Box("ERROR -- Unable to allocate space for map.");
      return;
    }
    memset(FourierPt, 0, mapstep_a*mapstep_b*mapstep_c * sizeof(float));

// start the Beevers-Lipson expansion

    Progress_Window(-1,"Computing Density",(float)mapstep_a*mapstep_b);
    
    progress=0;
    for(i=0; i < mapstep_a; i++) {
        x = (float)i/(float)mapstep_a;
        memset(Cklx, 0, nk * nl * sizeof(float));
        memset(Dklx, 0, nk * nl * sizeof(float));
        for (l=0; l<nr; l++) {
            ijk = (il[l] - lmin) * nk + ik[l] - kmin;
            Cklx[ijk] += fo[l] * (float)cos(2.0 * PI * ih[l] * x - f2[l]);
            Dklx[ijk] -= fo[l] * (float)sin(2.0 * PI * ih[l] * x - f2[l]);
        }
        for(j=0; j < mapstep_b; j++) {
            Progress_Window(0,NULL,(float)++progress);
            y = (float)j/(float)mapstep_b;
            memset(Elxy, 0, nl * sizeof(float));
            memset(Flxy, 0, nl * sizeof(float));
            for (l=0; l<nl; l++) {
                for (k=0; k<nk; k++) {
                    ijk = l * nk + k;
                    Elxy[l] += Cklx[ijk] * (float)cos(2.0 * PI * (k+kmin) * y)
                             + Dklx[ijk] * (float)sin(2.0 * PI * (k+kmin) * y);
                    Flxy[l] += Dklx[ijk] * (float)cos(2.0 * PI * (k+kmin) * y)
                             - Cklx[ijk] * (float)sin(2.0 * PI * (k+kmin) * y);
                }
            }
            for(k=0; k < mapstep_c; k++) {
              z = (float)k/(float)mapstep_c;
              ijk =  i*mapstep_b*mapstep_c + j*mapstep_c + k;

              for (l=0; l<nl; l++) {
                FourierPt[ijk] += Elxy[l] * (float)cos(2.0 * PI * (l + lmin) * z)
                                + Flxy[l] * (float)sin(2.0 * PI * (l + lmin) * z);
              }
              if (FourierPt[ijk] < rhomin) 
                  rhomin = FourierPt[ijk];
              if (FourierPt[ijk] > rhomax) 
                  rhomax = FourierPt[ijk];
            }
        }
    }
    if(!Quick)
        fprintf(drvui->fcns,"Reading Fourier map file %s\n  "
            "map steps: a=%d b=%d c=%d\n  "
            "Range of rho values: %f to %f, gridsize %d\n", infile, 
 	        mapstep_a, mapstep_b, mapstep_c, rhomin, rhomax, ijk);


    Progress_Window(-2, NULL, 0.0f);
    ReadFourMap = 1;
    Map_Info.rhomn = rhomin;
    Map_Info.rhomx = rhomax;
    Map_Info.map_int[0] = mapstep_a;
    Map_Info.map_int[1] = mapstep_b;
    Map_Info.map_int[2] = mapstep_c;
    Map_Info.lat_con[0] = drvui->lat_con[0];
    Map_Info.lat_con[1] = drvui->lat_con[1];
    Map_Info.lat_con[2] = drvui->lat_con[2];
    Map_Info.lat_con[3] = drvui->lat_con[3];
    Map_Info.lat_con[4] = drvui->lat_con[4];
    Map_Info.lat_con[5] = drvui->lat_con[5];
    Map_Info.xlim[0] = 0.0f;
    Map_Info.xlim[1] = 1.0f;
    Map_Info.ylim[0] = 0.0f;
    Map_Info.ylim[1] = 1.0f;
    Map_Info.zlim[0] = 0.0f;
    Map_Info.zlim[1] = 1.0f;
	
    free(Cklx);
    free(Dklx);
    free(Elxy);
    free(Flxy);
    free(fo);
    free(fc);
    free(f2);
    free(ih);
    free(ik);
    free(il);
}

/* ************************************************************** */
/* ************************************************************** */

void read_flp(char * infile, int Quick)
{
// read FullProf binary map file

#include "read_flp.h"                // define all the struct's

    char titulo[81];
    char version[6];
    struct crystal_cell_type celda;
    FILE * mapin;
    float rhomin=1.0e+15f, rhomax=-1.0e+15f;
    int ireslt = 0;
    char string[200];
    int  icount;
    struct space_group spg;
    char temp[16384];
    int i, j, k, ijk=0;
    int natoms;
    struct FFT_param map_param;
    struct Atom_Type atm[200];
    float dens[200];
    int new_mapstep_a, astart, aend;
    int new_mapstep_b, bstart, bend;
    int new_mapstep_c, cstart, cend;
    int wrong_end = 0;

    memset(version, 0, 6);
    memset(titulo, 0, 81);
    if ((mapin = fopen (infile, "r+b")) == NULL) {
        sprintf(string,"Cannot open Fourier map (FulProf) file %s", infile);
        Error_Box(string);
        return;
    }
    ireslt = fread(&icount, sizeof(int), 1, mapin);         // read record length
	if (!ireslt) {
          char string[200];
          sprintf(string, "Error reading Fourier map file %s",infile);
          Error_Box(string);
          fclose(mapin);
    }
    if (icount != 80) {
        icount = end_flip(icount);
        wrong_end = 1;
    }
    if (icount != 80) {
        Error_Box("Map does not have the correct format\nJob aborted.");
        return;
    }
    fread(titulo, sizeof(char), icount, mapin);    // read title
    fread(&icount, sizeof(int), 1, mapin);         // skip trailer
    trim_string(titulo,80);
    strcpy(Map_Info.title,titulo);

    fread(&icount, sizeof(int), 1, mapin);         // read header for next record
    if (wrong_end) icount = end_flip(icount);
    fread(version, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         // skip trailer

    fread(&icount, sizeof(int), 1, mapin);         // read header for next record
    if (wrong_end) icount = end_flip(icount);
    fread(&celda, sizeof(char), icount, mapin);
    if (wrong_end) {
        for (i=0; i<3;i++) {
            celda.ang[i] = end_flip_real(celda.ang[i]);
            celda.cell[i] = end_flip_real(celda.cell[i]);
        }
    }
    fread(&icount, sizeof(int), 1, mapin);         // skip trailer
    for (i=0;i<3;i++) {
        Map_Info.lat_con[i] = celda.cell[i];
        Map_Info.lat_con[i+3] = celda.ang[3];
    }

    fread(&icount, sizeof(int), 1, mapin);         // read header for next record
    if (wrong_end) icount = end_flip(icount);
    fread(&spg, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         // skip trailer

    fread(&icount, sizeof(int), 1, mapin);         // read header for next record
    if (wrong_end) icount = end_flip(icount);
    fread(&natoms, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         // skip trailer

    fread(&icount, sizeof(int), 1, mapin);         // read header for next record
    if (wrong_end) icount = end_flip(icount);
    fread(&atm, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         // skip trailer

	for (i=0; i<2; i++) {                          // skip next 2 records
        fread(&icount, sizeof(int), 1, mapin);     // read header for next record
        if (wrong_end) icount = end_flip(icount);
        fread(&temp, sizeof(char), icount, mapin);
        fread(&icount, sizeof(int), 1, mapin);     // skip trailer
	}

    fread(&icount, sizeof(int), 1, mapin);         // read header for next record
    if (wrong_end) icount = end_flip(icount);
    fread(&map_param, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         // skip trailer
    if (wrong_end) {
        for (i=0; i<3; i++) {
            map_param.ngrid[i] = end_flip(map_param.ngrid[i]);
        }
    }
    Map_Info.map_int[0] = mapstep_a = map_param.ngrid[0];
    Map_Info.map_int[1] = mapstep_b = map_param.ngrid[1];
    Map_Info.map_int[2] = mapstep_c = map_param.ngrid[2];
    for (i=0; i<2; i++) {
        if (wrong_end) {
            map_param.xlim[i] = end_flip_real(map_param.xlim[i]);
            map_param.ylim[i] = end_flip_real(map_param.ylim[i]);
            map_param.zlim[i] = end_flip_real(map_param.zlim[i]);
        }
        Map_Info.xlim[i] = map_param.xlim[i];
        Map_Info.ylim[i] = map_param.ylim[i];
        Map_Info.zlim[i] = map_param.zlim[i];
    }
    Map_Info.info_valid = 1;

    fread(&icount, sizeof(int), 1, mapin);         // read header for next record
    if (wrong_end) icount = end_flip(icount);
    fread(&temp, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         // skip trailer

	if ((fabs(Map_Info.xlim[0]) > 0.001) || (fabs(Map_Info.xlim[1] - 1.0f) > 0.001)) {
// not full cell along x
		new_mapstep_a = (int)(1.0f/(Map_Info.xlim[1] - Map_Info.xlim[0])+0.5f) * mapstep_a;
		astart = (int)(Map_Info.xlim[0] * new_mapstep_a);
		aend = (int)(Map_Info.xlim[1] * new_mapstep_a);
    } else {
		new_mapstep_a = mapstep_a;
		astart = 0;
		aend = mapstep_a;
	}
	if ((fabs(Map_Info.ylim[0]) > 0.001) || (fabs(Map_Info.ylim[1] - 1.0f) > 0.001)) {
// not full cell along y
		new_mapstep_b = (int)(1.0f/(Map_Info.ylim[1] - Map_Info.ylim[0])+0.5f) * mapstep_b;
		bstart = (int)(Map_Info.ylim[0] * new_mapstep_b);
		bend = (int)(Map_Info.ylim[1] * new_mapstep_b);
    } else {
		new_mapstep_b = mapstep_b;
		bstart = 0;
		bend = mapstep_b;
	}
	if ((fabs(Map_Info.zlim[0]) > 0.001) || (fabs(Map_Info.zlim[1] - 1.0f) > 0.001)) {
// not full cell along z
		new_mapstep_c = (int)(1.0f/(Map_Info.zlim[1] - Map_Info.zlim[0])+0.5f) * mapstep_c;
		cstart = (int)(Map_Info.zlim[0] * new_mapstep_c);
		cend = (int)(Map_Info.zlim[1] * new_mapstep_c);
    } else {
		new_mapstep_c = mapstep_c;
		cstart = 0;
		cend = mapstep_c;
	}
    if (FourierPt) free(FourierPt);
    FourierPt = (float *) malloc( new_mapstep_a*new_mapstep_b*new_mapstep_c * sizeof(float));
    if (FourierPt == NULL) {
        ireslt = 0;
        Error_Box("ERROR -- Unable to allocate space for map.");
        return;
    }
    memset(FourierPt,0, new_mapstep_a*new_mapstep_b*new_mapstep_c * sizeof(float));

// read the Rho values

     for (i=astart; i< aend; i++) {
        for (j=bstart; j<bend; j++) {
            fread(&icount, sizeof(int), 1, mapin); // read header for next record
            if (wrong_end) icount = end_flip(icount);
            fread(&dens, sizeof(char), icount, mapin);
            fread(&icount, sizeof(int), 1, mapin); // skip trailer
            for (k=cstart; k<cend; k++) {
				ijk = i * new_mapstep_c * new_mapstep_b + j * new_mapstep_c + k;
                if (wrong_end) dens[k] = end_flip_real(dens[k]);
                FourierPt[ijk] = dens[k];
                if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
                if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
            }
        }
    }
    fclose(mapin);
    if(!Quick)
        fprintf(drvui->fcns,"Reading Fourier map file %s\n  "
            "cell = %f %f %f %f %f %f\n  map steps: a=%d b=%d c=%d\n  "
            "Range of rho values: %f to %f, gridsize %d\n", infile, 
 	        celda.cell[0], celda.cell[1], celda.cell[2], 
            celda.ang[0], celda.ang[1], celda.ang[2], 
 	        mapstep_a, mapstep_b, mapstep_c, rhomin, rhomax, ijk);
    ReadFourMap = 1;
	Map_Info.rhomn = rhomin;
	Map_Info.rhomx = rhomax;
    mapstep_a = new_mapstep_a;
    mapstep_b = new_mapstep_b;
    mapstep_c = new_mapstep_c;
	Map_Info.map_int[0] = mapstep_a;
	Map_Info.map_int[1] = mapstep_b;
	Map_Info.map_int[2] = mapstep_c;
}

/* ************************************************************** */
/* ************************************************************** */

void read_grd(char * infile, int Quick)
{
      FILE * mapin;
      char line[81];
      char * reslt;
      float rhomin=1.0e15f, rhomax=-1.0e15f;
      int ireslt = 0;
      char string[200];
      int new_mapstep_a = 0, astart, aend;
	  int new_mapstep_b = 0, bstart, bend;
	  int new_mapstep_c = 0, cstart, cend;

      if ((mapin = fopen (infile, "r")) == NULL) {
          sprintf(string,"Cannot open Fourier map (.grd) file %s", infile);
        Error_Box(string);
        return;
      }
      reslt = fgets(line, sizeof(line), mapin);
      trim_string(line,81);
      strcpy(Map_Info.title,line);
      Map_Info.info_valid = 1;
      if (reslt != NULL) {
        ireslt = fscanf(mapin, "%f %f %f %f %f %f", 
 			&map_a, &map_b, &map_c, &map_alpha, &map_beta, &map_gamma);
        Map_Info.lat_con[0] = map_a;
        Map_Info.lat_con[1] = map_b;
        Map_Info.lat_con[2] = map_c;
        Map_Info.lat_con[3] = map_alpha;
        Map_Info.lat_con[4] = map_beta;
        Map_Info.lat_con[5] = map_gamma;
      }
      if (ireslt > 0)
        ireslt = fscanf(mapin, "%d %d %d", 
 			&mapstep_a, &mapstep_b, &mapstep_c);
      if (ireslt > 0) {
        int i, j , k, ijk;
        if ((fabs(Map_Info.xlim[0]) > 0.001) || (fabs(Map_Info.xlim[1] - 1.0f) > 0.001)) {
// not full cell along x
            new_mapstep_a = (int)(1.0f/(Map_Info.xlim[1] - Map_Info.xlim[0])+0.5f) * mapstep_a;
            astart = (int)(Map_Info.xlim[0] * new_mapstep_a);
            aend = (int)(Map_Info.xlim[1] * new_mapstep_a);
        } else {
            new_mapstep_a = mapstep_a;
            astart = 0;
            aend = mapstep_a;
        }
        if ((fabs(Map_Info.ylim[0]) > 0.001) || (fabs(Map_Info.ylim[1] - 1.0f) > 0.001)) {
// not full cell along y
            new_mapstep_b = (int)(1.0f/(Map_Info.ylim[1] - Map_Info.ylim[0])+0.5f) * mapstep_b;
            bstart = (int)(Map_Info.ylim[0] * new_mapstep_b);
            bend = (int)(Map_Info.ylim[1] * new_mapstep_b);
        } else {
            new_mapstep_b = mapstep_b;
            bstart = 0;
            bend = mapstep_b;
        }
        if ((fabs(Map_Info.zlim[0]) > 0.001) || (fabs(Map_Info.zlim[1] - 1.0f) > 0.001)) {
// not full cell along z
            new_mapstep_c = (int)(1.0f/(Map_Info.zlim[1] - Map_Info.zlim[0])+0.5f) * mapstep_c;
            cstart = (int)(Map_Info.zlim[0] * new_mapstep_c);
            cend = (int)(Map_Info.zlim[1] * new_mapstep_c);
        } else {
            new_mapstep_c = mapstep_c;
            cstart = 0;
            cend = mapstep_c;
        }

        if (FourierPt) free(FourierPt);

        FourierPt = (float*)malloc( new_mapstep_a*new_mapstep_b*new_mapstep_c * sizeof(float));
        if (FourierPt == NULL) {
          ireslt = 0;
          Error_Box("ERROR -- Unable to allocate space for map.\n");
          return;
        }
        memset(FourierPt,0,new_mapstep_a*new_mapstep_b*new_mapstep_c*sizeof(float));

/* read the Rho values */
        for(i=astart; i < aend; i++) {
          for(j=bstart; j < bend; j++) {
            for(k=cstart; k < cend; k++) {
              ijk =  i*new_mapstep_b*new_mapstep_c + j*new_mapstep_c + k;
              ireslt = fscanf(mapin,"%f", &FourierPt[ijk]);
              if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
              if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
            }
          }
        }
      }
      Map_Info.rhomn = rhomin;
      Map_Info.rhomx = rhomax;
      Map_Info.map_int[0] = mapstep_a = new_mapstep_a;
      Map_Info.map_int[1] = mapstep_b = new_mapstep_b;
      Map_Info.map_int[2] = mapstep_c = new_mapstep_c;
       
      if (ireslt > 0) {
          if (!Quick) {
              fprintf(drvui->flout,"Reading Fourier map file %s\n\tTitle= %s\n  "
                    "cell = %f %f %f %f %f %f\n  map steps: a=%d b=%d c=%d\n  "
                    "Range of rho values: %f to %f\n", infile, line, 
 	                 map_a, map_b, map_c, map_alpha, map_beta, map_gamma, 
 	                 mapstep_a, mapstep_b, mapstep_c, rhomin, rhomax);
          }
          ReadFourMap = 1;
      } else {
          char string[200];
          sprintf(string, "Error reading Fourier map file %s",infile);
          Error_Box(string);
      }
      fclose(mapin);
}

/* ************************************************************** */
/* ************************************************************** */

void read_stf(char * infile, int Quick)
{
      FILE * mapin;
      char line[81];
      char * reslt;
      float rhomin=1.0e15f, rhomax=-1.0e15f;
      int ireslt = 0;
      int ijk = 0;
      int i,j,k;
      float bx1,bx2,by1,by2,bz1,bz2;
      char string[200];
	  int new_mapstep_a, astart, aend;
	  int new_mapstep_b, bstart, bend;
	  int new_mapstep_c, cstart, cend;
            
      if ((mapin = fopen (infile, "r")) == NULL) {
        sprintf (string,
           "Cannot open Fourier map (.stf) file %s\n", infile);
        Error_Box(string);
        return;
      }
      reslt = fgets(line, sizeof(line), mapin);
      reslt = fgets(line, sizeof(line), mapin);
      reslt = fgets(line, sizeof(line), mapin);
      if (reslt){
        ireslt = sscanf(line, "%*s %d %d %d", 
 			&mapstep_a, &mapstep_b, &mapstep_c);
      }		
      if (ireslt > 0) {
        reslt = fgets(line, sizeof(line), mapin);
        sscanf (line,"%*s %f %f %f %f %f %f",&bx1,&bx2,&by1,&by2,&bz1,&bz2);
        Map_Info.xlim[0] = bx1;
		Map_Info.xlim[1] = bx2;
        Map_Info.ylim[0] = by1;
		Map_Info.ylim[1] = by2;
        Map_Info.zlim[0] = bz1;
		Map_Info.zlim[1] = bz2;
        Map_Info.lat_con[0] = map_a = drvui->lat_con[0];
        Map_Info.lat_con[1] = map_b = drvui->lat_con[1];
        Map_Info.lat_con[2] = map_c = drvui->lat_con[2];
        Map_Info.lat_con[3] = map_alpha = drvui->lat_con[3];
        Map_Info.lat_con[4] = map_beta = drvui->lat_con[4];
        Map_Info.lat_con[5] = map_gamma = drvui->lat_con[5];

        if ((fabs(Map_Info.xlim[0]) > 0.001) || (fabs(Map_Info.xlim[1] - 1.0f) > 0.001)) {
// not full cell along x
            new_mapstep_a = (int)(1.0f/(Map_Info.xlim[1] - Map_Info.xlim[0])+0.5f) * mapstep_a;
            astart = (int)(Map_Info.xlim[0] * new_mapstep_a);
            aend = (int)(Map_Info.xlim[1] * new_mapstep_a);
        } else {
            new_mapstep_a = mapstep_a;
            astart = 0;
            aend = mapstep_a;
        }
        if ((fabs(Map_Info.ylim[0]) > 0.001) || (fabs(Map_Info.ylim[1] - 1.0f) > 0.001)) {
// not full cell along y
            new_mapstep_b = (int)(1.0f/(Map_Info.ylim[1] - Map_Info.ylim[0])+0.5f) * mapstep_b;
            bstart = (int)(Map_Info.ylim[0] * new_mapstep_b);
            bend = (int)(Map_Info.ylim[1] * new_mapstep_b);
        } else {
            new_mapstep_b = mapstep_b;
            bstart = 0;
            bend = mapstep_b;
        }
        if ((fabs(Map_Info.zlim[0]) > 0.001) || (fabs(Map_Info.zlim[1] - 1.0f) > 0.001)) {
// not full cell along z
            new_mapstep_c = (int)(1.0f/(Map_Info.zlim[1] - Map_Info.zlim[0])+0.5f) * mapstep_c;
            cstart = (int)(Map_Info.zlim[0] * new_mapstep_c);
            cend = (int)(Map_Info.zlim[1] * new_mapstep_c);
        } else {
            new_mapstep_c = mapstep_c;
            cstart = 0;
            cend = mapstep_c;
        }

        if (FourierPt) free(FourierPt);

        FourierPt = (float*)malloc( new_mapstep_a*new_mapstep_b*new_mapstep_c * sizeof(float));
        if (FourierPt == NULL) {
          ireslt = 0;
          Error_Box("ERROR -- Unable to allocate space for map.\n");
          return;
        }
        memset(FourierPt,0,new_mapstep_a*new_mapstep_b*new_mapstep_c*sizeof(float));
        reslt = fgets(line, sizeof(line), mapin);
        reslt = fgets(line, sizeof(line), mapin);
        reslt = fgets(line, sizeof(line), mapin);

        if (map_alpha == 90. && map_beta == 90. && map_gamma == 90.) {
        /* read the Rho values */ /* C blocks of A*B values */
          for(k=cstart; k < cend; k++) {
            for(j=bstart; j < bend-1; j++) {
              for(i=astart; i < aend-1; i++) {
                ijk =  k*(new_mapstep_a-1)*(new_mapstep_b-1) + j*(new_mapstep_a-1) + i;
                ireslt = fscanf(mapin,"%f", &FourierPt[ijk]);
                if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
                if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
              }
              fscanf(mapin,"%*f");
            }
            for (i=0;i<mapstep_a;i++) fscanf(mapin,"%*f");
          }

        } else {
          for(i=astart; i < aend; i++) {
            for(j=bstart; j < bend-1; j++) {
              for(k=cstart; k < cend-1; k++) {
                ijk =  k*(new_mapstep_a-1)*(new_mapstep_b-1) + j*(new_mapstep_a-1) + i;
                ireslt = fscanf(mapin,"%f", &FourierPt[ijk]);
                if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
                if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
              }
              fscanf(mapin,"%*f");
            }
            for (k=0;k<mapstep_c;k++)fscanf(mapin,"%*f");
          }
        } /* orthorhombic or oblique */ 
        mapstep_a = --new_mapstep_a;
        mapstep_b = --new_mapstep_b;
        mapstep_c = --new_mapstep_c;
		Map_Info.map_int[0] = mapstep_a;
		Map_Info.map_int[1] = mapstep_b;
		Map_Info.map_int[2] = mapstep_c;
        Map_Info.rhomn = rhomin;
        Map_Info.rhomx = rhomax;

      } /* if mapstep line present */  
       
      if (ijk > 0) {
        if (!Quick)
          fprintf(drvui->flout,"Reading Fourier map file %s\n  "
            "cell = %f %f %f %f %f %f\n  map steps: a=%d b=%d c=%d\n  "
            "Range of rho values: %f to %f, gridsize %d\n", infile, 
 	        map_a, map_b, map_c, map_alpha, map_beta, map_gamma, 
 	        mapstep_a, mapstep_b, mapstep_c, rhomin, rhomax,ijk);
        ReadFourMap = 1;
      } else {
        sprintf(string, "Error reading Fourier map file %s", 
 	       infile);
        Error_Box(string);
      }
//for (i=0;i<=ijk;i++) fprintf(stderr,"pt %d = %f\n",i,FourierPt[i]);
      fclose(mapin);
}

/* ************************************************************** */
/* ************************************************************** */

void read_vasp(char * infile, int Quick)
{
      FILE * mapin;
      char line[200];
      char * reslt;
      float rhomin=1.0e15f, rhomax=-1.0e15f;
      int ireslt = 0;
      int ijk = 0;
      int i,j,k,l,m;
	  float tmp_rho[10];
      char string[200];
      float scale;
      float cmat[3][3];
      int new_mapstep_a = 0, astart, aend;
      int new_mapstep_b = 0, bstart, bend;
      int new_mapstep_c = 0, cstart, cend;
            
      map_a = drvui->lat_con[0];
      map_b = drvui->lat_con[1];
      map_c = drvui->lat_con[2];
      map_alpha = drvui->lat_con[3];
      map_beta = drvui->lat_con[4];
      map_gamma = drvui->lat_con[5];
	  if ((mapin = fopen (infile, "r")) == NULL) {
        sprintf (string,
           "Cannot open VASP file %s\n", infile);
        Error_Box(string);
        return;
      }
	  fgets(line, sizeof(line), mapin);
      trim_string(line,80);
      strcpy(Map_Info.title,line);
	  fgets(line, sizeof(line), mapin);
      sscanf(line, "%f", &scale);
	  for(i=0; i<3; i++) {   // read the cell matrix
		  if (!(reslt = fgets(line, sizeof(line), mapin))) {
			  sprintf(string,"Error reading VASP file.");
			  Error_Box(string);
			  return;
		  }
          sscanf(line,"%f %f %f",&cmat[i][0],&cmat[i][1],&cmat[i][2]);
	  }
      for (i=0; i<3; i++)
          for (j=0; j<3; j++)
              cmat[i][j] *= scale;
	  fgets(line, sizeof(line), mapin);
      Map_Info.lat_con[0] = (float) sqrt(cmat[0][0]);   // this section needs fixing
      Map_Info.lat_con[1] = (float) sqrt(cmat[1][1]);
      Map_Info.lat_con[2] = (float) sqrt(cmat[2][2]);
      Map_Info.lat_con[3] = 90.0f;            
      Map_Info.lat_con[4] = 90.0f;
      Map_Info.lat_con[5] = 90.0f;
	  j = 0;
	  i = strlen(line)/4;
	  for (k=0; k<i; k++) {
		  sscanf(line + 4 * k,"%4d",&l);
		  j += l;
	  }
	  for(i=0; i<j+2; i++) {   // set to skip k-point stuff
		  if (!(reslt = fgets(line, sizeof(line), mapin))) {
			  sprintf(string,"Error reading VASP file.");
			  Error_Box(string);
			  return;
		  }
	  }
	  fgets(line, sizeof(line), mapin);
      ireslt = sscanf(line, "%d %d %d", 
 			&mapstep_a, &mapstep_b, &mapstep_c);
      if (ireslt > 0) {
        if ((fabs(Map_Info.xlim[0]) > 0.001) || (fabs(Map_Info.xlim[1] - 1.0f) > 0.001)) {
// not full cell along x
            new_mapstep_a = (int)(1.0f/(Map_Info.xlim[1] - Map_Info.xlim[0])+0.5f) * mapstep_a;
            astart = (int)(Map_Info.xlim[0] * new_mapstep_a);
            aend = (int)(Map_Info.xlim[1] * new_mapstep_a);
        } else {
            new_mapstep_a = mapstep_a;
            astart = 0;
            aend = mapstep_a;
        }
        if ((fabs(Map_Info.ylim[0]) > 0.001) || (fabs(Map_Info.ylim[1] - 1.0f) > 0.001)) {
// not full cell along y
            new_mapstep_b = (int)(1.0f/(Map_Info.ylim[1] - Map_Info.ylim[0])+0.5f) * mapstep_b;
            bstart = (int)(Map_Info.ylim[0] * new_mapstep_b);
            bend = (int)(Map_Info.ylim[1] * new_mapstep_b);
        } else {
            new_mapstep_b = mapstep_b;
            bstart = 0;
            bend = mapstep_b;
        }
        if ((fabs(Map_Info.zlim[0]) > 0.001) || (fabs(Map_Info.zlim[1] - 1.0f) > 0.001)) {
// not full cell along z
            new_mapstep_c = (int)(1.0f/(Map_Info.zlim[1] - Map_Info.zlim[0])+0.5f) * mapstep_c;
            cstart = (int)(Map_Info.zlim[0] * new_mapstep_c);
            cend = (int)(Map_Info.zlim[1] * new_mapstep_c);
        } else {
            new_mapstep_c = mapstep_c;
            cstart = 0;
            cend = mapstep_c;
        }

        if (FourierPt) free(FourierPt);

        FourierPt = (float*)malloc( new_mapstep_a*new_mapstep_b*new_mapstep_c * sizeof(float));
        if (FourierPt == NULL) {
          ireslt = 0;
          Error_Box("ERROR -- Unable to allocate space for map.\n");
          return;
        }
        memset(FourierPt,0,new_mapstep_a*new_mapstep_b*new_mapstep_c*sizeof(float));

// read the Rho values

		m = 11;
        for(i=cstart; i < cend; i++) {
          for(j=bstart; j < bend; j++) {
            for(k=astart; k < aend; k++) {
			    if (m > 9) {
					reslt = fgets(line, sizeof(line), mapin);
					sscanf(line,"%f %f %f %f %f %f %f %f %f %f",&tmp_rho[0],
						&tmp_rho[1],&tmp_rho[2],&tmp_rho[3],&tmp_rho[4],
						&tmp_rho[5],&tmp_rho[6],&tmp_rho[7],&tmp_rho[8],
						&tmp_rho[9]);
					m = 0;
				}
                ijk =  k * new_mapstep_b * new_mapstep_c + j * new_mapstep_c + i;
				FourierPt[ijk] = tmp_rho[m++];
                if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
                if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
            }
          }
        }
      }
       
      if (reslt) {
          ReadFourMap = 1;
      } else {
          char string[200];
          sprintf(string, "Error reading Fourier map file %s",infile);
          Error_Box(string);
      }
      Map_Info.rhomn = rhomin;
      Map_Info.rhomx = rhomax;
      Map_Info.map_int[0] = mapstep_a = new_mapstep_a;
      Map_Info.map_int[1] = mapstep_b = new_mapstep_b;
      Map_Info.map_int[2] = mapstep_c = new_mapstep_c;
	  Map_Info.info_valid = 1;
      fclose(mapin);

}

/* ************************************************************** */
/* ************************************************************** */

void read_w2k(char * infile, int Quick)
{
      FILE * mapin;
      char line[81];
      char * reslt;
      float rhomin=1.0e15f, rhomax=-1.0e15f;
      int ireslt = 0;
      int dimen=0;
      int ijk = 0;
      int i,j,k;
      char string[200];
      int new_mapstep_a = 0, astart, aend;
      int new_mapstep_b = 0, bstart, bend;
      int new_mapstep_c = 0, cstart, cend;

      Map_Info.lat_con[0] = map_a = drvui->lat_con[0];
      Map_Info.lat_con[1] = map_b=drvui->lat_con[1];
      Map_Info.lat_con[2] = map_c=drvui->lat_con[2];
      Map_Info.lat_con[3] = map_alpha=drvui->lat_con[3];
      Map_Info.lat_con[4] = map_beta=drvui->lat_con[4];
      Map_Info.lat_con[5] = map_gamma=drvui->lat_con[5];
            
      if ((mapin = fopen (infile, "r")) == NULL) {
        sprintf (string,
           "Cannot open Fourier map (.w2k) file %s\n", infile);
        Error_Box(string);
        return;
      }

      do {
        reslt = fgets(line, sizeof(line), mapin);
        if (!strncmp(line,"(@3 [3]",7)) dimen=1;
      } while (!feof(mapin) && dimen==0);
 
      if (dimen==0) {
        sprintf (string,
           "Cannot read Fourier map (.w2k) file %s\n", infile);
        Error_Box(string);
        return;
      }

      ireslt=sscanf(line, "%*s %*s %d %d %d",&mapstep_a,&mapstep_b,&mapstep_c);

      if (ireslt > 0) {
        if ((fabs(Map_Info.xlim[0]) > 0.001) || (fabs(Map_Info.xlim[1] - 1.0f) > 0.001)) {
// not full cell along x
            new_mapstep_a = (int)(1.0f/(Map_Info.xlim[1] - Map_Info.xlim[0])+0.5f) * mapstep_a;
            astart = (int)(Map_Info.xlim[0] * new_mapstep_a);
            aend = (int)(Map_Info.xlim[1] * new_mapstep_a);
        } else {
            new_mapstep_a = mapstep_a;
            astart = 0;
            aend = mapstep_a;
        }
        if ((fabs(Map_Info.ylim[0]) > 0.001) || (fabs(Map_Info.ylim[1] - 1.0f) > 0.001)) {
// not full cell along y
            new_mapstep_b = (int)(1.0f/(Map_Info.ylim[1] - Map_Info.ylim[0])+0.5f) * mapstep_b;
            bstart = (int)(Map_Info.ylim[0] * new_mapstep_b);
            bend = (int)(Map_Info.ylim[1] * new_mapstep_b);
        } else {
            new_mapstep_b = mapstep_b;
            bstart = 0;
            bend = mapstep_b;
        }
        if ((fabs(Map_Info.zlim[0]) > 0.001) || (fabs(Map_Info.zlim[1] - 1.0f) > 0.001)) {
// not full cell along z
            new_mapstep_c = (int)(1.0f/(Map_Info.zlim[1] - Map_Info.zlim[0])+0.5f) * mapstep_c;
            cstart = (int)(Map_Info.zlim[0] * new_mapstep_c);
            cend = (int)(Map_Info.zlim[1] * new_mapstep_c);
        } else {
            new_mapstep_c = mapstep_c;
            cstart = 0;
            cend = mapstep_c;
        }

        if (FourierPt) free(FourierPt);

        FourierPt = (float*)malloc( new_mapstep_a*new_mapstep_b*new_mapstep_c * sizeof(float));
        if (FourierPt == NULL) {
          ireslt = 0;
          Error_Box("ERROR -- Unable to allocate space for map.\n");
          return;
        }
        memset(FourierPt,0,new_mapstep_a*new_mapstep_b*new_mapstep_c*sizeof(float));

        reslt = fgets(line, sizeof(line), mapin);

        for(k=cstart; k < cend; k++) {
          for(j=bstart; j < bend-1; j++) {
            for(i=astart; i < aend-1; i++) {
              ijk =  j*(new_mapstep_a-1)*(new_mapstep_b-1) + k*(new_mapstep_a-1) + i;
              ireslt = fscanf(mapin,"%f", &FourierPt[ijk]);
              if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
              if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
            }
            fscanf(mapin,"%*f");
          }
          for (i=0;i<mapstep_a;i++) fscanf(mapin,"%*f");
        }

     }

     if (ijk > 0) {
       if (!Quick)
         fprintf(drvui->flout,"Reading Fourier map file %s\n  "
            "cell = %f %f %f %f %f %f\n  map steps: a=%d b=%d c=%d\n  "
            "Range of rho values: %f to %f, gridsize %d\n", infile, 
            map_a, map_b, map_c, map_alpha, map_beta, map_gamma, 
            mapstep_a, mapstep_b, mapstep_c, rhomin, rhomax,ijk);
        ReadFourMap = 1;
      } else {
        sprintf(string, "Error reading Fourier map file %s", 
 	       infile);
        Error_Box(string);
      }
      fclose(mapin);
      Map_Info.rhomn = rhomin;
      Map_Info.rhomx = rhomax;
	  Map_Info.map_int[0] = mapstep_a = --new_mapstep_a;
	  Map_Info.map_int[1] = mapstep_b = --new_mapstep_b;
	  Map_Info.map_int[2] = mapstep_c = --new_mapstep_c;
	  Map_Info.info_valid = 1;
}

/* ************************************************************** */
/* ************************************************************** */

void read_inp (int Quick)

/* routine to read and process input file */
{
  char input[100];                 /* input buffer for data read from file */
  char input2[100];                /* input buffer for data read from file */
  char string[100];                /* temporary string */
  char filename[100];              /* used for map file name */
  int intype;                      /* type of datum read */
  int i, j, k;                     /* temporary */
  char t_name[100];                /* temporary atom name storage */
  char t_color[40];                /* temporary ellipsoid color storage */
  int in_line = 0;                 /* flag for inline import of foreign files */
  float t_rad;
  char el_color_tmp[MAX_EL][100];
  int n_el_tmp = 0;
  int done;
  int tmp_frame_no = 1;            // number of frame being processed
  FILE * tout = 0;
  int curframe;drvui->lat_con[0] = 0.0f;
  int MapRegionRead = 0;
  char atom_pos[3][30];
  

/* ************************************************************** */
/* ************************************************************** */
/*
    Commands Implemented

    arrow
    atom
    axislines
    background
    betaij
    bij,Bij
    bond
    bounds
    box
    cell
    clip
    cutout
    dash
    depthcue
    edges
    ellipcolor
    ellipsoids
    finish
    frame
    import
    inline
    labelscale
    labeltext
    list
    lonepair
    lookat
    magtrans
    magnification
    mapcalclimits
    mapcontour
    mapread
    mapregion
    molcomp
    nolabels
    noshadow
    origin
    orthographic
    pack
    phong
    plane
    polyedge
    polysz
    polytolerance
    polylimit
    polyfudge
    polyvert
    rem, REM
    shell
    slab
    special
    spgp, sgrp, spgr
    sphere
    title, titl
    uij,Uij
    vectors    
    view
    vrml1
    vrml2
    vrml97
    xyzoff
    end, END
                                                                  */
/* ************************************************************** */
/* ************************************************************** */

  drvui->xyzoff_read = 0;
  Map_Info.xlim[0] = 0.0f;
  Map_Info.xlim[1] = 1.0f;
  Map_Info.ylim[0] = 0.0f;
  Map_Info.ylim[1] = 1.0f;
  Map_Info.zlim[0] = 0.0f;
  Map_Info.zlim[1] = 1.0f;
  numOfFourierContours = 0;
  curframe = drvui->frame_no;
  if (curframe > 1 && packflag) {
      for (i=0; i<6; i++) {
          drvui->cryst_lim[curframe][i] = drvui->cryst_lim[curframe-1][i];
      }
  } 
  for (i = 0; i < 3; i++)
    drvui->xyzoff[i] = 0.0f;
  memset(el_color_tmp, 0, sizeof(el_color_tmp));
  while (!feof (drvui->fpin)) {
    memset(t_color,0,40);
    memset(string,0,100);
    memset(input,0,sizeof(input));
    if (fgets (input, 100, drvui->fpin)) {         /* read a line */
      trim_string(input,100);          // kill any ^M or ^J at end of line
      Blank_Strip (input);             // suppress any leading blanks
      strcpy(input2,input);            // make a copy for listing file
      for (i=strlen(input); i< 99; i++) { //pad line with blanks
          input[i] = ' ';
      }
      input[99] = 0;                   // force a terminator
    } else {
      Error_Box("Error reading file.");
      return;
    }
    intype = 1000;
    if (!strncmp (input, "title", 5))
        if (drvui->frame_no == 1)
            intype = 1;                // only echo title in first frame
        else
            intype = 0;
    if (strncmp (input, "titl", 4) == 0)
        if (drvui->frame_no == 1)
            intype = 1;                // only echo title in first frame
        else
            intype = 0;
    if (strncmp (input, "TITL", 4) == 0)
        if (drvui->frame_no == 1)
            intype = 1;                // only echo title in first frame
        else
            intype = 0;
    if (strncmp (input, "rem", 3) == 0)
        if (tmp_frame_no == drvui->frame_no)
            intype = 1;                // only echo rem in its frame
        else
            intype = 0;
    if (strncmp (input, "REM", 3) == 0)
        if (tmp_frame_no == drvui->frame_no)
            intype = 1;                // only echo rem in its frame
        else
            intype = 0;
    if (strncmp (input, "cell", 4) == 0)
      intype = 2;
    if (strncmp (input, "sgrp", 4) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 3;
        else
          intype = 0;                       // ignore sgrp command if not current frame
    if (strncmp (input, "spgp", 4) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 3;
      else
          intype = 0;                       // ignore spgp command if not current frame
    if (strncmp (input, "spgr", 4) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 3;
      else
          intype = 0;                       // ignore spgr command if not current frame
    if (strncmp (input, "ellipsoids", 10) == 0)
      intype = 4;
    if (strncmp (input, "lookat", 6) == 0)
      intype = 0; //5;
    if (strncmp (input, "atom", 4) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 6;
        else
          intype = 0;                     // ignore atom command if not current frame
    if (strncmp (input, "polysz", 6) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 7;
        else
          intype = 0;                     // ignore polysz command if not current frame
    if (strncmp (input, "sphere", 6) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 8;
        else
          intype = 0;                     // ignore sphere command if not current frame
    if (strncmp (input, "end", 3) == 0)
      intype = 9;
    if (strncmp (input, "END", 3) == 0)
      intype = 9;
    if (strncmp (input, "xyzoff", 6) == 0)
      intype = 10;
    if (strncmp (input, "bond", 4) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 11;
        else
          intype = 0;                     // ignore bond command if not current frame
    if (strncmp (input, "box", 3) == 0)
      intype = 12;
    if (strncmp (input, "edges", 5) == 0)
      intype = 13;
    if (strncmp (input, "view", 4) == 0)
      intype = 14;
    if (strncmp (input, "plane", 5) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 15;
        else
          intype = 0;                     // ignore plane command if not current frame
    if (strncmp (input, "bij", 3) == 0)
      intype = 16;
    if (strncmp (input, "Bij", 3) == 0)
      intype = 16;
    if (strncmp (input, "betaij", 6) == 0)
      intype = 17;
    if (strncmp (input, "uij", 3) == 0)
      intype = 18;
    if (strncmp (input, "Uij", 3) == 0)
      intype = 18;
    if (strncmp (input, "bounds", 6) == 0)
      intype = 1; //19;
    if (strncmp (input, "pack", 4) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 20;
        else
          intype = 0;                     // ignore pack command if not current frame
    if (strncmp (input, "magnification", 13) == 0)
      intype = 21;
    if (strncmp (input, "origin", 6) == 0)
      intype = 22;
    if (strncmp (input, "ellipcolor", 10) == 0)
      intype = 23;
    if (strncmp (input, "import", 6) == 0)
      intype = 24;
    if (strncmp (input, "inline", 6) == 0)
      intype = 25;
    if (strncmp (input, "orthographic", 12) == 0)
      intype = 26;
    if (strncmp (input, "vectors", 7) == 0) {
        if (tmp_frame_no == 1) {
            intype = 27;
        } else {
            intype = 0;
        }
    }
    if (strncmp (input, "phong", 5) == 0)
      intype = 28;
    if (strncmp (input, "molcomp", 7) == 0)
      intype = 29;
    if (strncmp (input, "cutout", 6) == 0)
      intype = 30;
    if (strncmp (input, "depthcue", 8) == 0)
      intype = 31;
    if (strncmp (input, "vrml1", 5) == 0)
      intype = 32;
    if (strncmp (input, "vrml2", 5) == 0)
      intype = 1;
    if (strncmp (input, "vrml97", 6) == 0)
      intype = 1;
    if (strncmp (input, "polytol", 7) == 0)
      intype = 33;
    if (strncmp (input, "polylimit", 9) == 0)
      intype = 33;
    if (strncmp (input, "polyfudge", 9) == 0)
      intype = 33;
    if (strncmp (input, "clip", 4) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 34;
        else
          intype = 0;                     // ignore clip command if not current frame
    if (strncmp (input, "list", 4) == 0)
      intype = 35;
    if (strncmp (input, "polyedge", 8) == 0)
      intype = 36;
    if (strncmp (input, "frame", 5) == 0)
      intype = 37;
    if (strncmp (input, "axislines", 9) == 0)
      intype = 38;
    if (strncmp (input, "shell", 5) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 39;
        else
          intype = 0;                     // ignore shell command if not current frame
    if (strncmp (input, "nolabels", 8) == 0)
      intype = 40;
    if (strncmp (input, "lonepair", 8) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 41;
        else
          intype = 0;                     // ignore lonepair command if not current frame
    if (strncmp (input, "labeltext", 9) == 0)
      intype = 42;
    if (strncmp (input, "polyvert", 8) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 43;
        else
          intype = 0;                     // ignore polyvert command if not current frame
    if (strncmp (input, "background", 10) == 0)
      intype = 44;
    if (strncmp (input, "slab", 4) == 0)
      intype = 45;
    if (strncmp (input, "mag_trans", 9) == 0)
      intype = 46;
    if (strncmp (input, "special", 7) == 0)
      intype = 47;
    if (strncmp (input, "noshadow", 8) == 0)
      intype = 48;
    if (strncmp (input, "finish", 6) == 0)
      intype = 49;
    if (strncmp (input, "arrow", 5) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 50;
        else
          intype = 0;                     // ignore arrow command if not current frame
    if (strncmp (input, "dash", 4) == 0)
      if(tmp_frame_no == drvui->frame_no)
          intype = 51;
        else
          intype = 0;                     // ignore bond command if not current frame
    if (strncmp (input, "mapcontour", 10) == 0)
      intype = 52;
    if (strncmp (input, "mapread", 7) == 0)
      intype = 53;
    if (strncmp (input, "mapregion", 9) == 0)
      intype = 54;
    if (strncmp (input, "mapcalclimits", 13) == 0)
      intype = 55;
    if (strncmp (input, "labelscale", 10) == 0)
      intype = 56;

    switch (intype) {

    case 0:
      break;
    case 1:
      break;                    // ignore title except for listing
    case 2:                     // cell command
      for (i = 3; i <= 5; ++i)
         drvui->lat_con[i] = 0.0f;
         sscanf (input, "%s %f %f %f %f %f %f", string, &drvui->lat_con[0]
          , &drvui->lat_con[1], &drvui->lat_con[2], &drvui->lat_con[3], &drvui->lat_con[4]
          , &drvui->lat_con[5]);
      break;
    case 3:
      symop (input);                 /* process space group information */
      break;
    case 4:                     /* ellipsoids command */
      do_ellipsoids = 1;
      (void)sscanf (input, "%s %f", string, &Ellipsoid_Prob);
      if (Ellipsoid_Prob > 1)
        Ellipsoid_Prob *= 0.01f;         /* Probability to fraction */
      break;
    case 5:                     // read vectors for lookat command
//      sscanf (input, "%s %f %f %f %f %f %f", string, &drvui->lookat_v1[0],
//          &drvui->lookat_v1[1], &drvui->lookat_v1[2], &drvui->lookat_v2[0],
//          &drvui->lookat_v2[1], &drvui->lookat_v2[2]);
//      Options = Options | L_OPT;
//      break;
    case 6:                     /* atom command */
      char t_atom[5];
      get_label (input, &t_atom[0], &t_atom[1], &t_atom[2], &t_atom[3], 1);
      t_atom[4] = 0;
      strcpy(drvui->atom_l[natom],t_atom);
      (void)sscanf (input, "%d %s %s %s",&drvui->atom_n[natom],atom_pos[0],
          atom_pos[1],atom_pos[2]);
      drvui->sv_atom_n[natom] = drvui->atom_n[natom];
      for (i=0; i<3; i++)
          drvui->atom_xyz[natom][i] = convert_pos(atom_pos[i]);
      drvui->atom_fn[natom] = tmp_frame_no;
      natom++;
      if (natom >= MAX_ATOM) {
        Error_Box("Too many atoms in asymmetric unit - increase MAX_ATOM.");
        return;
      }
      break;
    case 7:                      /* polysz command */
    case 39:                     /* shell command */
    case 43:                     /* polyvert command */
      get_label (input, &drvui->poly_l[drvui->npoly][0], &drvui->poly_l[drvui->npoly][1]
           , &drvui->poly_l[drvui->npoly][2], &drvui->poly_l[drvui->npoly][3], 1);
      if (intype == 7) {
          (void)sscanf (input, "%f %39c", &drvui->poly_size[drvui->npoly], string);
      } else if (intype == 39) { 
          (void)sscanf (input, "%f %f %39c",  &drvui->poly_min[drvui->npoly],
              &drvui->poly_size[drvui->npoly], string);
      } else if (intype == 43) {
          get_label (input, &drvui->poly_t[drvui->npoly][0], &drvui->poly_t[drvui->npoly][1]
             , &drvui->poly_t[drvui->npoly][2], &drvui->poly_t[drvui->npoly][3], 0);
         (void)sscanf (input, "%f %39c", &drvui->poly_size[drvui->npoly], string);
      }
      trim_string(string,40);
      strcpy(drvui->poly_col_p[drvui->npoly],string);
      strcpy (drvui->poly_col_v[drvui->npoly], drvui->poly_col_p[drvui->npoly]);
      check_color (drvui->poly_col_v[drvui->npoly]);
      if (intype == 7 || intype == 43)
          drvui->poly_min[drvui->npoly] = 0.005f;
      drvui->poly_fn[drvui->npoly] = tmp_frame_no;
      if (intype == 7 || intype == 39)
          strcpy(drvui->poly_t[drvui->npoly],"");
      drvui->npoly++;
      if (drvui->npoly >= MAX_POLY) {
          Error_Box("Too many polyhedra specified - increase MAX_POLY");
          return;
      }
      break;
    case 8:                     /* sphere command */
      get_label (input, &drvui->sphere_l[drvui->nsphere][0], &drvui->sphere_l[drvui->nsphere][1]
         , &drvui->sphere_l[drvui->nsphere][2], &drvui->sphere_l[drvui->nsphere][3], 1);
      char str[5][40];
      i = sscanf(input,"%s %s %s %s %s", str[0], str[1], str[2], str[3], str[4]);
      if (i==2 || (i == 4 && strstr(str[2],"filter"))) {       // old-style input
          (void)sscanf(str[0],"%f",&drvui->sphere_size[drvui->nsphere]);
          drvui->sphere_n[drvui->nsphere] = -1;
          strcpy(string,str[1]);
          if (i >= 3) {
              strcat(string," ");
              strcat(string,str[2]);
          }
      } else if (i>=3) {                   // new-style input with atom number
          (void)sscanf(str[0],"%d",&drvui->sphere_n[drvui->nsphere]);
          (void)sscanf(str[1],"%f",&drvui->sphere_size[drvui->nsphere]);
          strcpy(string,str[2]);
      } else {
          Error_Box("Sphere input line malformed.");
      }
      if (i>=4) {
          strcat(string," ");
          strcat(string,str[3]);
      }
      if (i==5) {
          strcat(string," ");
          strcat(string,str[4]);
      }
      trim_string(string,40);
      strcpy(drvui->sphere_col_p[drvui->nsphere],string);
      strcpy (drvui->sphere_col_v[drvui->nsphere], drvui->sphere_col_p[drvui->nsphere]);
      check_color (drvui->sphere_col_v[drvui->nsphere]);
      drvui->sphere_fn[drvui->nsphere] = tmp_frame_no;
      drvui->p_rad_edge[drvui->npoly]=0.0f;
      drvui->nsphere++;
      if (drvui->nsphere >= MAX_SPHERE) {
          Error_Box("Too many sphere types specified - increase MAX_SPHERE.");
          return;
      }
      break;
    case 9:                     /* end or END */
      if (!drvui->lat_con[0]) {
          Error_Box("There is no 'cell' line in your input file.");
      }
      label_cell();
      if (Quick == 2) {
          drvui->max_frame = tmp_frame_no;
          drvui->Old_Xrot = xrot;
          drvui->Old_Yrot = yrot;
          drvui->Old_Zrot = zrot;
      }
      int frame;
      for (frame = 1; frame<=drvui->max_frame; frame++) {
          if (!Quick) {
              strcpy(input,drvui->Cur_Root);
              sprintf(input2,".frm%d",frame);
              strcat(input,input2);
              tout = fopen(input,"w");
          }
          for (i = 0; i < natom; i++)  {       /* shift atom coordinates for origin offset */
              if (drvui->atom_fn[i] != frame) continue;
              drvui->TF_status[i] = -1;                 // initialize TF type/status
              for (j = 0; j < 3; j++)
                  drvui->atom_xyz[i][j] -= drvui->xyzoff[j];
              strncpy(t_atom,drvui->atom_l[i],4);
              t_atom[4] = 0;
              if (!Quick)
                  fprintf(tout,"%s %d\n",t_atom,drvui->sv_atom_n[i]);
          }
          if (!Quick) {
              fclose(tout);
          }
      }
      if (n_el_tmp > 0) {
          strcpy(input,"");
          for (i = 0; i < n_el_tmp; i++) {
            get_label (el_color_tmp[i], &t_name[0], &t_name[1],
                 &t_name[2], &t_name[3], 0);
            t_name[4] = 0;
            if (el_color_tmp[i][0] == '*') {
              (void)sscanf (el_color_tmp[i], "%s %39c", string, t_color);
              j = -1;
            } else {
              (void)sscanf (el_color_tmp[i], "%d %39c", &j, t_color);
            }
            trim_string(t_color,40);
            for (k = 1; k < n_ellips; k++) {     /* find this atom in ellipsoid list */
                if (check_atom_name(t_name,drvui->ellips_l[k])) {
                  if ((j == -1) || (j == drvui->ellips_n[k])) {
                    drvui->ell_type[k] += 1000;                  // >= 1000 means ellipcolor read
                    strcpy (drvui->ellips_col_p[k], t_color);    /* copy color */
                    strcpy (drvui->ellips_col_v[k], drvui->ellips_col_p[k]);
                    check_color (drvui->ellips_col_v[k]);
                    drvui->save_el_number[k]=j;
                  }
                }
            } 
          }
      }
      for (i=1; i<drvui->nsphere; i++) {        // remove any auto-generated ellipsoids that
          for (j=1; j<n_ellips; j++) {          //     interfere with spheres
              if (drvui->ell_type[j] == 1001) {
                  if (check_atom_name(drvui->sphere_l[i],drvui->ellips_l[j])) {
                      if (drvui->sphere_n[i] == -1
                          || drvui->sphere_n[i] == drvui->save_el_number[j]) {
                          drvui->ell_type[j] = 1;
                      }
                  }
              }
          }
      }
      switch (FourierMapType) {
      case 1:
          read_grd( FourierFileName, Quick);
          break;
      case 2:
          read_stf( FourierFileName, Quick);
          break;
      case 3:
          read_w2k( FourierFileName, Quick);
          break;
      case 4:
          read_vasp( FourierFileName, Quick);
          break;
      case 5:
          read_flp( FourierFileName, Quick);
          break;
      case 6:
          read_fcf( FourierFileName, Quick);
          break;
      default:
          break;
      }
      if (!MapRegionRead) {             // mapregion line NOT read
          xMin = Map_Info.xlim[0];
          xMax = Map_Info.xlim[1];
          yMin = Map_Info.ylim[0];
          yMax = Map_Info.ylim[1];
          zMin = Map_Info.zlim[0];
          zMax = Map_Info.zlim[1];
      }
      return;                     /* thru with end */
    case 10:                     /* xyzoff command */
      (void)sscanf (input, "%s %s %s %s",string,atom_pos[0],atom_pos[1],atom_pos[2]);
      for (i=0; i<3; i++)
          drvui->xyzoff[i] = convert_pos(atom_pos[i]);
      drvui->xyzoff_read = 1;
      break;
    case 51:                     /* dash command */
    case 11:                     /* bond command */
      get_label (input, &drvui->bond_l1[drvui->nbond][0], &drvui->bond_l1[drvui->nbond][1]
         , &drvui->bond_l1[drvui->nbond][2], &drvui->bond_l1[drvui->nbond][3], 1);
      get_label (input, &drvui->bond_l2[drvui->nbond][0], &drvui->bond_l2[drvui->nbond][1]
         , &drvui->bond_l2[drvui->nbond][2], &drvui->bond_l2[drvui->nbond][3], 0);
      (void)sscanf (input, "%f %f %f %39c",
          &drvui->bond_size[drvui->nbond], &drvui->bond_min[drvui->nbond], 
          &drvui->bond_max[drvui->nbond], string);
      trim_string(string,40);
      if (strlen (string) == 0)
          strcpy (string, "Grey");
      strcpy(drvui->col_bond_p[drvui->nbond],string);
      strcpy (drvui->col_bond_v[drvui->nbond], drvui->col_bond_p[drvui->nbond]);
      check_color (drvui->col_bond_v[drvui->nbond]);
      drvui->bond_fn[drvui->nbond] = tmp_frame_no;
      if (intype == 11)
        drvui->bond_style[drvui->nbond] = 0;
      else
        drvui->bond_style[drvui->nbond] = 1; 
      drvui->nbond++;
      if (drvui->nbond >= MAX_BOND) {
          Error_Box("Too many bond types specified - increase MAX_BOND.");
      }
      break;
    case 12:                     /* box command */
      (void)sscanf (input, "%s %f %39c", string, &rad_cell, t_color);
      trim_string(t_color,40);
      strcpy(col_cell_p,t_color);
      if (strlen (col_cell_p) == 0)
        strcpy (col_cell_p, "Black");
      strcpy (col_cell_v, col_cell_p);
      check_color (col_cell_v);
      if (rad_cell == 0.0) {
        docell = 0;
        Labels = 0;
      }
      break;
    case 13:                     /* edges command */
      (void)sscanf (input, "%s %f %39c", string, &rad_edge, t_color);
      trim_string(t_color,40);
      strcpy(col_edge_p,t_color);
      if (rad_edge > 0.005)
        edges = 2;
      if (strlen (col_edge_p) == 0)
        strcpy (col_edge_p, "Black");
      strcpy (col_edge_v, col_edge_p);
      check_color (col_edge_v);
      break;
    case 14:                     /* view command */
//      if ((Options & V_OPT) == 0)         // only if no -v switch 
        (void)sscanf (input, "%s %f %f %f", string, &xrot, &yrot, &zrot);
      break;
    case 15:                     /* plane command */
      get_label (input, &drvui->plane_l[drvui->nplane][0], &drvui->plane_l[drvui->nplane][1]
         , &drvui->plane_l[drvui->nplane][2], &drvui->plane_l[drvui->nplane][3], 1);
      (void)sscanf (input, "%f %39c", &drvui->plane_size[drvui->nplane], string);
      trim_string(string,40);
      strcpy(drvui->plane_col_p[drvui->nplane],string);
      strcpy (drvui->plane_col_v[drvui->nplane], drvui->plane_col_p[drvui->nplane]);
      check_color (drvui->plane_col_v[drvui->nplane]);
      drvui->plane_fn[drvui->nplane] = tmp_frame_no;
      drvui->nplane++;
      if (drvui->nplane >= MAX_PLANE) {
          Error_Box("Too many plane types specified - increase MAX_PLANE.");
          return;
      }
      break;
    case 16:                     /* bij or Bij command */
      drvui->ell_type[n_ellips] = 2;
    case 17:                     /* betaij command */
      if (intype == 17)
        drvui->ell_type[n_ellips] = 0;
    case 18:                     /* uij or Uij command */
      if (intype == 18)
        drvui->ell_type[n_ellips] = 1;
      get_label (input, &drvui->ellips_l[n_ellips][0], &drvui->ellips_l[n_ellips][1]
         , &drvui->ellips_l[n_ellips][2], &drvui->ellips_l[n_ellips][3], 1);
      drvui->ellips_l[n_ellips][4] = 0;
      (void)sscanf (input, "%d %f %f %f %f %f %f %39c", &drvui->ellips_n[n_ellips],
          &drvui->ellips[n_ellips][0], &drvui->ellips[n_ellips][1],
          &drvui->ellips[n_ellips][2], &drvui->ellips[n_ellips][3], 
          &drvui->ellips[n_ellips][4], &drvui->ellips[n_ellips][5],
          string);                                       // read unique coefficients and color
      drvui->save_el_number[n_ellips] = drvui->ellips_n[n_ellips];
      trim_string(string,40);
      strcpy(drvui->ellips_col_p[n_ellips],string);
      drvui->ellips_col_p[n_ellips][39] = 0;
      if (!strlen (drvui->ellips_col_p[n_ellips]))
        strcpy (drvui->ellips_col_p[n_ellips], "Gray80");
      strcpy (drvui->ellips_col_v[n_ellips], drvui->ellips_col_p[n_ellips]);
      check_color (drvui->ellips_col_v[n_ellips]);
      n_ellips++;
      if (n_ellips >= MAX_EL) {
          Error_Box("Too many ellipsoid types specified - increase MAX_EL.");
          return;
      }
      break;
    case 19:                     /* bounds command */
//      if ((Options & B_OPT) == 0) {         // only do this if no -b switch 
        (void)sscanf (input, "%s %f %f %f", string, &boxlim[0], &boxlim[1],
            &boxlim[2]);
        boxflag = 1;
//      }
      break;
    case 20:                     /* pack command */
//      if ((Options & P_OPT) == 0) {         // only if no -p switch
        (void)sscanf (input, "%s %f %f %f %f %f %f", string, 
            &drvui->cryst_lim[tmp_frame_no][0], &drvui->cryst_lim[tmp_frame_no][3],
            &drvui->cryst_lim[tmp_frame_no][1], &drvui->cryst_lim[tmp_frame_no][4],
            &drvui->cryst_lim[tmp_frame_no][2], &drvui->cryst_lim[tmp_frame_no][5]);
        packflag = 1;
//      }
      break;
    case 21:                     /* magnification command */
//      if ((Options & M_OPT) == 0)         // only if no -m switch
        (void)sscanf (input, "%s %f", string, &Magnification);
      break;
    case 22:                     /* origin command */
//      if ((Options & O_OPT) == 0) {         // only if no -o switch
        (void)sscanf (input, "%s %f %f %f", string, &origin[0], &origin[1],
            &origin[2]);
//      }
      break;
    case 23:                     /* ellipcolor command */
      Token_Strip (input, 1);
      strcpy (el_color_tmp[n_el_tmp++], input);
      break;
    case 24:                     /* inline command */
    case 25:                     /* import command */
      in_line = 0;
      done = 0;
      if (strncmp (input, "inline", 6) == 0)
        in_line = 1;
      Token_Strip (input, 1);             /* strip command */
      if (strncmp (input, "gsas", 4) == 0) {
        import_gsas (input, in_line, Quick);
        done = 1;
      }
      if (strncmp (input, "cif", 3) == 0) {
        import_cif (input, in_line, Quick);
        done = 1;
      }
      if (strncmp (input, "schakal", 7) == 0) {
        import_schakal (input, in_line, Quick);
        done = 1;
      }
      if (strncmp (input, "shakal", 6) == 0) {
        import_schakal (input, in_line, Quick);
        done = 1;
      }
      if (strncmp (input, "shelx", 5) == 0) {
        import_shelx (input, in_line, Quick);
        done = 1;
      }
      if (strncmp (input, "fdat", 4) == 0) {
        import_fdat (input, in_line, Quick);
        done = 1;
      }
      if (strncmp (input, "csd", 3) == 0) {
        import_fdat (input, in_line, Quick);
        done = 1;
      }
      if (strncmp (input, "ccdf", 4) == 0) {
        import_fdat (input, in_line, Quick);
        done = 1;
      }
      if (strncmp (input, "wien2k", 6) == 0) {
        import_wien (input, in_line, Quick);
        done = 1;
      }
      if (done == 0) {
          char string[256];
          trim_string(input,100);
          sprintf (string,"Import/inline source type '%s' not recognized.", input);
          Error_Box(string);
          return;
      }
      break;
    case 26:                     /* orthographic */
      M_cameras = 0;
      break;
    case 27: {                   /* unit cell direction vectors */
      float v[3];
      Display_axes = 1;
      offset[0] = offset[1] = offset[2] = 0.0f;
      i = sscanf( input, "%s %f %f %f", string, &v[0], &v[1], &v[2]);
      if (i == 4)
          for (i=0; i<3; i++)
              offset[i] = v[i];
      }
      break;
    case 28:                     /* phong */
      (void)sscanf (input, "%s %f %f", string, &drvui->Phong_Value, &drvui->Phong_Size);
      break;
    case 29:                     /* molcomp */
      (void)sscanf (input, "%s %f", string, &drvui->mol_d);
      domolcomp = 1;
      break;
    case 30:                     /* cutout */
      El_Cutout = 1;
      (void)sscanf (input, "%s %39c", string, t_color);
      trim_string(t_color,40);
      strcpy(drvui->Cutout_color,t_color);
      break;
    case 31:                     /* depthcue - thickness of edges scaled by Z coordinate */
      (void)sscanf (input, "%s %f", string, &DepthCue);
      break;
    case 32:                     /* vrml1 - turn on VRML1 output */
      Vrml2 = 0;
      break;
    case 33:                     /* polylimit (formerly polyfudge) */
      (void)sscanf (input, "%s %f", string, &drvui->polylimit);
      if (!Quick)
        fprintf (drvui->flout,
             "Tolerance factor for non-planarity of polyhedra faces set to %f. This option should be used with care\n",
             drvui->polylimit);
      break;
    case 34:                     /* clip command */
      (void)sscanf (input, "%s %f %f %f %f %f %f", string, &drvui->clip_lim[tmp_frame_no][0],
          &drvui->clip_lim[tmp_frame_no][3], &drvui->clip_lim[tmp_frame_no][1], 
          &drvui->clip_lim[tmp_frame_no][4], &drvui->clip_lim[tmp_frame_no][2], 
          &drvui->clip_lim[tmp_frame_no][5]);
      clipflag = 1;
      break;
    case 35:                     /* list - limit for printed distance table */
      (void)sscanf (input, "%s %f", string, &printdist);
      if (!Quick)
        fprintf (drvui->flout,
            "Interatomic distances up to %6.3f Angstroms will be tabulated in the logfile\n",
            printdist);
      break;
    case 36:                     /* polyedge - enables individual coloring of polyhedral edges */
      get_label (input, &t_name[0], &t_name[1], &t_name[2], &t_name[3], 1);
      (void)sscanf (input, "%f %39c", &t_rad, t_color);
      trim_string(t_color,40);
      for (i = 1; i < drvui->npoly; i++) {         /* find this atom in poly list */
        if ((t_name[0] == drvui->poly_l[i][0]) &&
            (t_name[1] == drvui->poly_l[i][1]) &&
            (t_name[2] == drvui->poly_l[i][2]) && (t_name[3] == drvui->poly_l[i][3])) {
          edges = 1;
          drvui->p_rad_edge[i] = t_rad;
          strcpy (drvui->p_col_edge_p[i], t_color);
          strcpy (drvui->p_col_edge_v[i], drvui->p_col_edge_p[i]);
          check_color (drvui->p_col_edge_v[i]);
        }
      }
      break;
    case 37:                     /* frame command */
      tmp_frame_no++;
      break;
    case 38:                     /* axislines */
      (void)sscanf (input, "%s %f %39c", string, &Ellipaxis_width,t_color);
      trim_string(t_color,40);
      strcpy(Ellipaxis_color,t_color);
      if (strlen(Ellipaxis_color) == 0)
        strcpy (Ellipaxis_color, "Gray20");
      break;
    case 40:                     /* nolabels command */
      Labels = 0;                 /* set value false */
      break;
    case 41:                     /* lonepair command */
      get_label (input, &drvui->cone_l1[drvui->ncone][0], &drvui->cone_l1[drvui->ncone][1]
         , &drvui->cone_l1[drvui->ncone][2], &drvui->cone_l1[drvui->ncone][3], 1);
      (void)sscanf (input, "%d %f %f %f %39c",
          &drvui->numlonepairs[drvui->ncone], &drvui->cone_height[drvui->ncone],
          &drvui->cone_min[drvui->ncone], &drvui->cone_max[drvui->ncone], t_color);
      trim_string(t_color,40);
      strcpy(drvui->col_cone_p[drvui->ncone],t_color);
      if (strlen (drvui->col_cone_p[drvui->ncone]) == 0)
        strcpy (drvui->col_cone_p[drvui->ncone], "Grey");
      strcpy (drvui->col_cone_v[drvui->ncone], drvui->col_cone_p[drvui->ncone]);
      check_color (drvui->col_cone_v[drvui->ncone]);
      drvui->cone_fn[drvui->ncone] = tmp_frame_no;
      drvui->ncone++;
      if (drvui->ncone >= MAX_CONE) {
          Error_Box("Too many lp bond types specified - increase MAX_CONE.");
          return;
      }
      break;
    case 42:                     /* labeltext command */
      memset (drvui->label_label[drvui->nlabel], 0, 64);
      Token_Strip (input, 1);
      (void)sscanf (input, "%f %f %f %63c",
          &(drvui->label_x[drvui->nlabel][0]), &(drvui->label_x[drvui->nlabel][1]),
          &(drvui->label_x[drvui->nlabel][2]), drvui->label_label[drvui->nlabel]);
      trim_string(drvui->label_label[drvui->nlabel],64);
      drvui->label_fn[drvui->nlabel] = tmp_frame_no;
      intype = 0;
      if(!strcmp(drvui->label_label[drvui->nlabel],"a") && curframe != 1) break;
      if(!strcmp(drvui->label_label[drvui->nlabel],"b") && curframe != 1) break;
      if(!strcmp(drvui->label_label[drvui->nlabel],"c") && curframe != 1) break;
      if(!strcmp(drvui->label_label[drvui->nlabel],"o") && curframe != 1) break;
      intype = 42;
      drvui->nlabel++;
      if (drvui->nlabel >= MAX_LABELS) {
          Error_Box("Too many labels specified - increase MAX_LABELS.");
          return;
      }
      break;
    case 44:                     // background (color) command
      Token_Strip (input, 1);
      strcpy (drvui->col_bg_p, input);
      trim_string(drvui->col_bg_p,40);
      if (strlen(drvui->col_bg_p) < 2)strcpy(drvui->col_bg_p,"White");
      strcpy (drvui->col_bg_v, drvui->col_bg_p);
      check_color (drvui->col_bg_v);
      sscanf(drvui->col_bg_v,"%f %f %f",&drvui->glback[0],&drvui->glback[1],&drvui->glback[2]);
      break;
    case 45:
      (void)sscanf (input,"%*s %f %f %f %f %f %f %f %f %f %f %f %f %d",
                    &drvui->slab_con[0],&drvui->slab_con[1],&drvui->slab_con[2],
                    &drvui->slab_con[3],&drvui->slab_con[4],&drvui->slab_con[5],
                    &drvui->slab_off[0],&drvui->slab_off[1],&drvui->slab_off[2],
                    &drvui->slab_rot[0],&drvui->slab_rot[1],&drvui->slab_rot[2],
                    &slabmode);

      break;
    case 46:                     // mag_trans command
      Token_Strip (input, 1);
      (void) sscanf(input,"%f %f %f %f %f %f %f %f %f",&drvui->mag_matrix[0][0],
                    &drvui->mag_matrix[0][1],&drvui->mag_matrix[0][2],
                    &drvui->mag_matrix[1][0],&drvui->mag_matrix[1][1],
                    &drvui->mag_matrix[1][2],&drvui->mag_matrix[2][0],
                    &drvui->mag_matrix[2][1],&drvui->mag_matrix[2][2]);
      break;
    case 47:                     // special command
      if (!Quick) break;
      (void) sscanf (input, "%*7c %d %d", &Omit->omit1[Omit->nomits],
          &Omit->omit2[Omit->nomits]);
      Omit->nomits++;
      break;
    case 48:                    // noshadow command
      drvui->noshadow = 1;
      break;
    case 49:			// finish command
      (void) sscanf (input, "%*6c %f %f %f %f",&drvui->ambient,
            &drvui->diffuse, &drvui->specular, &drvui->roughness);	
      break;
    case 50:            // arrow command
      Token_Strip (input, 1);
      (void) sscanf(input,"%f %f %f %f %f %f %f %f %39c",&drvui->mag_xp[drvui->nmag][0],
                         &drvui->mag_xp[drvui->nmag][1],&drvui->mag_xp[drvui->nmag][2],
                         &drvui->mag_xc[drvui->nmag][0],&drvui->mag_xc[drvui->nmag][1],
                         &drvui->mag_xc[drvui->nmag][2],&drvui->arrow_length[drvui->nmag],
                         &drvui->arrow_diam[drvui->nmag],t_color);
      trim_string(t_color,40);
      strcpy(drvui->col_arrow_p[drvui->nmag],t_color);
      strcpy(drvui->col_arrow_v[drvui->nmag],drvui->col_arrow_p[drvui->nmag]);
      check_color(drvui->col_arrow_v[drvui->nmag]);
      drvui->arrow_fn[drvui->nmag] = tmp_frame_no;
      drvui->nmag++;
      if (drvui->nmag >= MAX_ARROW) {
          Error_Box("Too many arrow types specified - increase MAX_ARROW.");
          return;
      }
      break;
    case 52:            /* mapcontour command */
      numOfFourierContours++;
      if (numOfFourierContours > MAX_MAP_CONTOUR) {
          Error_Box("Too many contours specified - increase MAX_MAP_CONTOUR.");
          exit (0);
      }
      Token_Strip (input, 1);
      (void) sscanf(input,"%f %s %39c",&FourierContourLevel[numOfFourierContours],
          string,t_color);
      trim_string(t_color,40);
      strcpy(FourierContourColor[numOfFourierContours],t_color);
      if(strncmp(string, "mesh", 4) == 0)
          FourierContourSolid[numOfFourierContours] = 0;
      else
          FourierContourSolid[numOfFourierContours] = 1;
      break;
    case 53:              /* mapread command */
      Token_Strip (input, 1);
      if (!Quick) break;
      memset(filename,0,sizeof(filename));
      t_name[0]='\0';
      (void)sscanf(input, "%s %s %s", string, filename, t_name);
      trim_string(filename, 100);
      trim_string(t_name,100);

      if (strncmp(string, "grd", 3) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=1;
      } else if (strncmp(string, "stf", 3) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=2;
      } else if (strncmp(string, "w2k", 3) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=3;
      } else if (strncmp(string, "vsp", 4) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=4;
      } else if (strncmp(string, "flp", 4) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=5;
      } else if (strncmp(string, "fcf", 3) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=6;
          Map_Info.map_type = 0;
          if (!strcmp(t_name,"Fc")) Map_Info.map_type = 1;
          else if (!strcmp(t_name,"Fo-Fc")) Map_Info.map_type = 2;
          else if (!strcmp(t_name,"2Fo-Fc")) Map_Info.map_type = 3;
          else if (!strcmp(t_name,"Fo2")) Map_Info.map_type = 4;
      } else {
          Error_Box("mapread command does not specify a legal file type.");
      }
      break;
    case 54:              /* mapregion command */
      Token_Strip(input, 1);
      (void) sscanf(input, "%f %f %f %f %f %f",&xMin, &xMax, &yMin,
          &yMax, &zMin, &zMax);
	  MapRegionRead = 1;
      break;
    case 55:              /* mapcalclimits command */
      Token_Strip(input, 1);
      (void) sscanf(input, "%f %f %f %f %f %f",&Map_Info.xlim[0], &Map_Info.xlim[1],
          &Map_Info.ylim[0], &Map_Info.ylim[1], &Map_Info.zlim[0], &Map_Info.zlim[1]);
      break;
    case 56:              /* labelscale command */
      Token_Strip(input, 1);
      (void) sscanf(input, "%f", &drvui->label_scale);
      break;

    default:
      if (!Quick)
        fprintf (drvui->flout,
             "********** Next line does not start with a recognizable keyword - ignored\n");
      break;
    }
    if (!Quick && intype)
        fprintf (drvui->flout, "*         %s\n", input2);     // echo to listing file
  }
  if (n_ellips > 1)
    do_ellipsoids = 1;
  if (packflag != 0) {
    boxlim[0] = boxlim[1] = boxlim[2] = 500.0f;
  }
}                         /* end of read_inp */

/* ************************************************************** */
/* ************************************************************** */

void set_tf_status(void)
{
// set the TF_status flags for the atoms in the list

    int i, j;

    for (i=0; i<natom; i++) {
        if (drvui->atom_fn[i] != drvui->frame_no) continue;
        drvui->TF_status[i] = -1;
        for (j=1; j<n_ellips; j++) {
            if (check_atom_name(drvui->atom_l[i], drvui->ellips_l[j])) {
              if (drvui->ellips_n[j] == -1
                    || drvui->ellips_n[j] == drvui->sv_atom_n[i]) {
                if(drvui->ell_type[j] > 1000) {
                    drvui->TF_status[i] = 1;
                    j = n_ellips;
                    break;
                } else {
                    if(drvui->ell_type[j] > 0) {
                        drvui->TF_status[i] = 0;
                        j = n_ellips;
                        break;
                    }
                }
              }
            }
        }
        j = drvui->TF_status[i];
    }
}


/* ************************************************************** */
/* ************************************************************** */

void Token_Strip (char string[100], int no)

/*  string[100] - input output string */
/*  no - number of tokens to strip */
/* routine to strip tokens from the leading part of a string */
{
  int i, j;

  for (i = 0, j = 0; i < no; i++) {         /* loop through tokens to be removed */
    while (string[j++] != ' ')
      string[j - 1] = ' ';             /* characters to space */
    while (string[j] == ' ')
      j++;                     /* skip all spaces */
  }
  Blank_Strip (string);                 /* move stuff to left */
}

/* ************************************************************** */
/* ************************************************************** */

void trim_string(char string[], int len)
{
// trim string - remove trailing ^J (Windows) and spaces. 'len' is maximum length
    int i;
    string[len-1] = 0;
    if (strlen(string) == 0) {
        strcpy(string,"White");
        return;
    }
    for (i=0;i<(int)strlen(string);i++) {
        if (string[i] < ' ') {
            string[i] = 0;
            break;
        }
    }
    for (i=strlen(string);i > 0; --i) {
        if (string[i-1] <= ' ') {
            string[i-1] = 0;
        } else
            break;
    }
}

/* ************************************************************** */
/* ************************************************************** */

int Unique_Atom (void)

// routine to check if atom at position 'natom' is unique. Returns 1 if it
//   is. If the position is a duplicate, returns 0.
{
  int j, k, test;

  for (k = 0; k < natom - 1; k++) {         /* loop through atoms */
    if (drvui->atom_fn[k] != drvui->frame_no) continue;
    if (drvui->atom_n[k] == drvui->atom_n[natom]) {
      test = 0;
      for (j = 0; j < 4; j++) {
    if (drvui->atom_l[k][j] != drvui->atom_l[natom][j])
      test = 1;
      }
      if (!test) {                     /* name and number the same - check coordinates */
    expand_atom (natom);             /* find all of test atoms in unit cell */
    for (j = 0; j < ncell; j++) {
      if ((fabs (drvui->atom_xyz[k][1] - drvui->cell_xyz[j][1]) < 0.0001) &&
          (fabs (drvui->atom_xyz[k][2] - drvui->cell_xyz[j][2]) < 0.0001) &&
          (fabs (drvui->atom_xyz[k][0] - drvui->cell_xyz[j][0]) < 0.0001))
        return (0);                 /* duplicate */
    }
      }
    }
  }
  return (1);                     /* unique */
}

